/*
 * Launcher.java Created on August 6, 2003, 2:03 PM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.exec;

import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * Utility used to launch threads. It requires a Runnable object the type of thread you want to run in and the priority to run at.
 * It will create the Thread, SchedulingParameters, ReleaseParameters MemoryParameters MemoryArea and ProcessingGroupParameters. It
 * will the start and join to the thread.
 * 
 * @author briang
 * @version $Id: JavaThreadLauncher.java,v 1.1 2004/09/23 15:34:23 mbi Exp $
 */
public class JavaThreadLauncher implements ThreadLauncher {

    /**
     * The thread to launch/kill
     * 
     * @uml.property name="thread"
     */
    protected Thread thread = null;


    /**
     * The type of thread to use 'RealtimeThread' or 'NoHeapRealtimeThread'
     * 
     * @uml.property name="threadType"
     * @uml.associationEnd 
     * @uml.property name="threadType" multiplicity="(0 1)"
     */
    protected ThreadType threadType = ThreadType.JAVATHREAD;

    /**
     * The type of memory area to run in.
     * 
     * @uml.property name="memoryType"
     * @uml.associationEnd 
     * @uml.property name="memoryType" multiplicity="(0 1)"
     */
    protected MemoryType memoryType = MemoryType.HEAPMEMORY;

    /**
     * The object to run.
     * 
     * @uml.property name="runnable"
     */
    protected Runnable runnable = new DummyRunnable();

    /**
     * The priority to run the test thread at.
     * 
     * @uml.property name="priority"
     */
    protected int priority = 250;


    /**
     * The priority to run the test thread at.
     */
    protected long                  iDelay               = 0;

    /**
     * The priority to run the test thread at.
     * 
     * @uml.property name="period"
     */
    protected long period = 0;

    /**
     * The number of iterations to perform
     * 
     * @uml.property name="iterations"
     */
    protected int iterations = -1; // run


    // forever
    // default

    /**
     * Creates a new instance of Launcher
     */
    public JavaThreadLauncher() {
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param data
     *            initialize from data structure
     */
    public JavaThreadLauncher( ThreadLauncherData data ) {
        setData( data );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param runnable
     *            the object to run in the created thread. The deafult
     */
    public JavaThreadLauncher( Runnable runnable ) {
        this.runnable = runnable;
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     */
    public JavaThreadLauncher( Runnable runnable, int priority ) {
        this( runnable );
        this.priority = priority;
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     */
    public JavaThreadLauncher( Runnable runnable, int priority, long period ) {
        this( runnable, priority );
        setPeriod( period );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     * @param initialDelay
     *            the initialDelay in milliseconds.
     */
    public JavaThreadLauncher( Runnable runnable, int priority, long period, long initialDelay ) {
        this( runnable, priority, period );
        setInitialDelay( initialDelay );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param runnable
     *            Runnable the object to run in the created thread. The deafult Launcher.ThreadType is REALTIMETHREAD the deafault
     *            MemoryArea is Heap and the deafult priority is PriorityScheduler.MAX_PRIORITY.
     */
    public JavaThreadLauncher( ThreadType tt, Runnable runnable ) {
        this( runnable );
        setThreadType( tt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param runnable
     *            Runnable the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     */
    public JavaThreadLauncher( ThreadType tt, Runnable runnable, int priority ) {
        this( runnable, priority );
        setThreadType( tt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     */
    public JavaThreadLauncher( ThreadType tt, Runnable runnable, int priority, long period ) {
        this( runnable, priority, period );
        setThreadType( tt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     * @param initialDelay
     *            the initialDelay in milliseconds.
     */
    public JavaThreadLauncher( ThreadType tt, Runnable runnable, int priority, long period, long initialDelay ) {
        this( runnable, priority, period, initialDelay );
        setThreadType( tt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param mt
     *            the MemoryType in which to execute the created thread.
     * @param runnable
     *            the object to run in the created thread. The deafult Launcher.ThreadType is REALTIMETHREAD the deafault MemoryArea
     *            is Heap and the deafult priority is PriorityScheduler.MAX_PRIORITY.
     */
    public JavaThreadLauncher( ThreadType tt, MemoryType mt, Runnable runnable ) {
        this( tt, runnable );
        setMemoryType( mt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param mt
     *            the MemoryType in which to execute the created thread.
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     */
    public JavaThreadLauncher( ThreadType tt, MemoryType mt, Runnable runnable, int priority ) {
        this( tt, runnable, priority );
        setMemoryType( mt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param mt
     *            the MemoryType in which to execute the created thread.
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     */
    public JavaThreadLauncher( ThreadType tt, MemoryType mt, Runnable runnable, int priority, long period ) {
        this( tt, runnable, priority, period );
        setMemoryType( mt );
    }

    /**
     * Creates a new instance of Launcher
     * 
     * @param tt
     *            the ThreadType of the thread to create.
     * @param mt
     *            the MemoryType in which to execute the created thread.
     * @param runnable
     *            the object to run in the created thread.
     * @param priority
     *            the priority of the thread in question. the deafult priority is PriorityScheduler.MAX_PRIORITY.
     * @param period
     *            the period in milliseconds.
     * @param initialDelay
     *            the initialDelay in milliseconds.
     */
    public JavaThreadLauncher( ThreadType tt, MemoryType mt, Runnable runnable, int priority, long period, long initialDelay ) {
        this( tt, runnable, priority, period, initialDelay );
        setMemoryType( mt );
    }

    /**
     * Setters and getters for member fields
     * 
     * @uml.property name="runnable"
     */
    public void setRunnable(Runnable runnable) {
        this.runnable = runnable;
    }

    /**
     * 
     * @uml.property name="runnable"
     */
    public Runnable getRunnable() {
        return runnable;
    }

    /**
     * 
     * @uml.property name="iterations"
     */
    public void setIterations(int iterations) {
        this.iterations = iterations;
    }

    /**
     * 
     * @uml.property name="iterations"
     */
    public int getIterations() {
        return iterations;
    }


    /**
     * 
     * @uml.property name="memoryType"
     */
    public void setMemoryType(MemoryType mt) {
        memoryType = mt;
    }

    /**
     * 
     * @uml.property name="memoryType"
     */
    public MemoryType getMemoryType() {
        return memoryType;
    }

    /**
     * 
     * @uml.property name="thread"
     */
    public Thread getThread() {
        if (thread == null) {
            thread = create();
        }
        return thread;
    }


    /**
     * 
     * @uml.property name="threadType"
     */
    public void setThreadType(ThreadType tt) {
        threadType = tt;
    }

    /**
     * 
     * @uml.property name="threadType"
     */
    public ThreadType getThreadType() {
        return threadType;
    }

    /**
     * 
     * @uml.property name="priority"
     */
    public void setPriority(int p) {
        priority = p;
    }

    /**
     * 
     * @uml.property name="priority"
     */
    public int getPriority() {
        return priority;
    }

    /**
     * 
     * @uml.property name="period"
     */
    public void setPeriod(long p) {
        period = p;
        if (0 != period) {
            setThreadType(ThreadType.PERIODICJAVATHREAD);
        }
    }

    /**
     * 
     * @uml.property name="period"
     */
    public long getPeriod() {
        return period;
    }

    public void setInitialDelay( long d ) {
        iDelay = d;
        if ( 0 != iDelay ) {
            setThreadType( ThreadType.PERIODICJAVATHREAD );
        }
    }

    public long getInitialDelay() {
        return iDelay;
    }

    /**
     * start the thread being managed by this launcher. This method can be overloaded by a subclass to provide more extensive setup
     * for more complex thread types
     */
    protected void start() {
        if ( null != thread ) {
            thread.start();
        }
    }

    /**
     * Launches the test. Creates the Thread, SchedulingParameters, ReleaseParameters MemoryParameters MemoryArea and
     * ProcessingGroupParameters. It will the start he thread.
     * 
     * @return the thread that was launched.
     */
    public Thread launch() {
        if (thread == null) {
            create();
        }
        thread.start();
        return thread;
    }

    /**
     * Creates the Thread, SchedulingParameters, ReleaseParameters MemoryParameters MemoryArea and ProcessingGroupParameters. It
     * will the start the thread as a Deamom
     * 
     * @return the thread that was launched.
     */
    public Thread launchDaemon() {
        if (thread == null) {
            create();
        }
        thread.setDaemon( true );
        thread.start();
        return thread;
    }

    /**
     * Kill the launched thread appropriately for its type.
     */
    public void kill() {
        if ( threadType.equals( ThreadType.PERIODICJAVATHREAD ) ) {
            ((PeriodicJavaThread)thread).kill();
        }
        else if ( threadType.equals( ThreadType.JAVATHREAD ) ) {
            ((PeriodicJavaThread)thread).kill();
        }
        else {
            throw new RuntimeException( "Unknown threadType: " + threadType );
        }
    }

    /**
     * Creates the Thread, SchedulingParameters, ReleaseParameters MemoryParameters MemoryArea and ProcessingGroupParameters.
     * 
     * @return the created thread.
     * @throws RuntimeException
     *             if the thread type requested is not known.
     */
    protected Thread create() throws RuntimeException {
        /*
         * Warn if there is no logic to run
         */
        if ( runnable == null ) System.out.println( "WARNING: Runnable null" );

        // run initialization specific to test.
        if ( runnable instanceof TestInitializer ) {
            runnable.run();
            runnable = ((TestInitializer)runnable).getTestRunner();
        }

        if ( threadType.equals( ThreadType.PERIODICJAVATHREAD ) ) {
            thread = new PeriodicJavaThread( runnable, iDelay, period, iterations );
        }
        else if ( threadType.equals( ThreadType.JAVATHREAD ) ) {
            thread = new PeriodicJavaThread( runnable, iDelay, period, 1 );
        }
        else {
            throw new RuntimeException( "Unknown threadType: " + threadType );
        }
        return thread;
    }

    /**
     * @see suramadu.util.exec.ThreadLauncher#setData(suramadu.util.exec.ThreadLauncherData)
     */
    public void setData( ThreadLauncherData data ) {
        if ( data.getRunnable() == null ) {
            System.out.println( "WARNING: " + this + ".setData(): runnable is null from data" );
        }
        setRunnable( data.getRunnable() );
        setIterations( data.getIterations() );
        setMemoryType( data.getMemoryType() );
        setThreadType( data.getThreadType() );
        setPriority( data.getPriority() );
        setPeriod( data.getPeriod() );
        setPriority( data.getPriority() );
        setInitialDelay( data.getInitialDelay() );
    }

    /**
     * @return a string representing the contents of this class.
     */
    public String toString() {
        StringWriter s = new StringWriter();
        PrintWriter  p = new PrintWriter( s );
        p.println( "thread=" + threadType + ",memType=" + memoryType + ",priority=" + priority + ",idDelay=" + iDelay + ",period=" + period + ",");
        return s.toString();
    }

    /** Internal class used as an Enumeration type for defining the different ThreadTypes supported. */

    /** This exists as a place holder in case a runnable is not specified. */
    private static final class DummyRunnable implements Runnable {
        public void run() {
            System.out.println( "Running: DummyRunnable (no other Runnable specified)." );
        }
    }
}