/*
 * PeriodicJavaThread.java Created on August 12, 2003, 9:09 AM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.exec;

import java.util.Timer;
import java.util.TimerTask;

/**
 * Simulates the periodic capabilities of a RealtimeThread and a NoHeapRealtimeThread in the RTSJ. It allows both the period and the
 * priority to be set.
 * 
 * @author mbi
 */
public class PeriodicJavaThread extends Thread {
    /**
     * The PeriodicJavaThread uses the java.util.Timer to control its execution and timing.
     */
    protected static final Timer timer      = new Timer( true );

    /**
     * Holder for the PeriodicJavaThread.Task class that is created for each PeriodicJavaThread created. The PeriodicJavaThread.Task
     * class extends java.util.TimerTask so that it can be given to the java.util.Timer for execution.
     */
    protected final TimerTask    task       = new Kicker();

    /**
     * Set to false when starting a thread, set to true by the kill() method. This flag is checked at the beginning and end of every
     * period. If found to be true, then the thread's periodic execution is terminated.
     */
    protected boolean            killed     = false;

    /**
     * Supports an iteration counter.
     */
    protected int                iterations = -1;               // run forever default

    /**
     * The period in milliseconds
     * 
     * @uml.property name="period"
     */
    protected long period = 0;

    /**
     * The initial delay in milliseconds
     * 
     * @uml.property name="delay"
     */
    protected long delay = 0;


    /**
     * The actual code to be run periodically
     */
    protected Runnable           runner     = null;

    /** For quickly checking periodic or aperiodic nature of the thread */
    protected boolean            periodic   = false;

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable.
     * 
     * @param r
     *            The Runnable to execute periodically.
     */
    public PeriodicJavaThread( Runnable r ) {
        super();
        runner = r;
    }

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable and period.
     * 
     * @param r
     *            the Runnable to execute periodically.
     * @param period
     *            the period (specified in milliseconds) in which to run.
     */
    public PeriodicJavaThread( Runnable r, long period ) {
        this( r );
        setPeriod( period );
        setPeriodic();
    }

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable, initial delay, and period.
     * 
     * @param r
     *            the Runnable to execute periodically.
     * @param delay
     *            initial delay before the first period (specified in milliseconds).
     * @param period
     *            the period (specified in milliseconds) in which to run.
     */
    public PeriodicJavaThread( Runnable r, long delay, long period ) {
        this( r );
        setPeriod( period );
        setDelay( delay );
        setPeriodic();
    }

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable and iteration count.
     * 
     * @param r
     *            The Runnable to execute periodically.
     * @param iterations
     *            The number of times to repeat a periodic thread.
     */
    public PeriodicJavaThread( Runnable r, int iterations ) {
        this( r );
        this.iterations = iterations;
        setPeriodic();
    }

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable, period and iterations.
     * 
     * @param r
     *            the Runnable to execute periodically.
     * @param period
     *            the period (specified in milliseconds) in which to run.
     * @param iterations
     *            The number of times to repeat a periodic thread.
     */
    public PeriodicJavaThread( Runnable r, long period, int iterations ) {
        this( r, period );
        this.iterations = iterations;
        setPeriodic();
    }

    /**
     * Creates a new instance of PeriodicJavaThread with the specified Runnable, initial delay, and period and iterations.
     * 
     * @param r
     *            the Runnable to execute periodically.
     * @param delay
     *            initial delay before the first period (specified in milliseconds).
     * @param period
     *            the period (specified in milliseconds) in which to run.
     * @param iterations
     *            The number of times to repeat a periodic thread.
     */
    public PeriodicJavaThread( Runnable r, long delay, long period, int iterations ) {
        this( r, delay, period );
        this.iterations = iterations;
        setPeriodic();
    }

    /**
     * Allows after-construction of the period of the thread.
     * 
     * @param period
     *            the period (specified in milliseconds) in which to run.
     * 
     * @uml.property name="period"
     */
    public void setPeriod(long period) {
        this.period = period;
    }

    /**
     * Allows after-construction of the initial delay and period of the thread.
     * 
     * @param delay
     *            initial delay before the first period (specified in milliseconds).
     * 
     * @uml.property name="delay"
     */
    public void setDelay(long delay) {
        this.delay = delay;
    }

    /** Cancels this PeriodicJavaThread by calling its task's cancel() method. */
    public synchronized void kill() {
        killed = true;
    }

    /**
     * sets whether this threadi s periodic or aperiodic based on if the period is non-zero, and the iterations are greater than 1.
     */
    protected void setPeriodic() {
        periodic = (period > 0) && (iterations > 1);
    }

    /**
     * When called, this thread will block until its next period is triggered.
     * 
     * @see #fire
     */
    protected synchronized void waitForNextPeriod() {
        try {
            wait();
        }
        catch ( InterruptedException e ) {
        }
    }

    /**
     * Called by TimerTask to release the block in waitForNextPeriod().
     * 
     * @see #waitForNextPeriod
     */
    protected synchronized void fire() {
        notify();
    }

    /**
     * Overloads Thread.start(). Handles scheduling with Timer and Timertask.
     * 
     * @see java.lang.Thread#start
     */
    public void start() {
        super.start();
        if ( periodic ) {
            timer.scheduleAtFixedRate( task, delay, period );
        }
        else {
            if ( delay < 100 ) {
                delay = 100;
            }
            timer.schedule( task, delay );
        }
    }

    /** The run() method for Runnable */
    public void run() {
        try {
            while ( !killed && (iterations != 0) ) {
                if ( periodic ) waitForNextPeriod();
                runner.run();
                if ( iterations > 0 ) iterations--;
            }
        }
        catch ( Throwable t ) {
            t.printStackTrace();
        }
    }

    /**
     * Private class that extends java.util.TimerTask that acts as the trigger for the PeriodicJavaThread's periodicity.
     */
    private class Kicker extends TimerTask {
        /** The run() method for Runnable */
        public void run() {
            PeriodicJavaThread.this.fire();
        }
    }
}