/*
 * RTJavaThread.java Created on August 12, 2003, 9:09 AM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.exec;

import javax.realtime.MemoryArea;
import javax.realtime.MemoryParameters;
import javax.realtime.PeriodicParameters;
import javax.realtime.ProcessingGroupParameters;
import javax.realtime.RealtimeThread;
import javax.realtime.ReleaseParameters;
import javax.realtime.SchedulingParameters;

/**
 * Wraps the periodic capabilities of a RealtimeThread and a NoHeapRealtimeThread in the RTSJ. It allows both the period and the
 * priority to be set.
 * 
 * @author mbi
 */
public class RTJavaThread extends RealtimeThread {
    public static final int NUMBEROFERRORSALLOWED = 1000;

    /**
     * Set to false when starting a thread, set to true by the kill() method. This flag is checked at the beginning and end of every
     * period. If found to be true, then the thread's periodic execution is terminated.
     */
    protected boolean       killed                = false;

    /**
     * Supports an iteration counter.
     */
    protected int           iterations            = -1;   // run forever default
    protected int           initialIterations     = -1;

    /**
     * The actual code to be run periodically
     */
    protected Runnable      runner                = null;

    /** For quickly checking periodic or aperiodic nature of the thread */
    protected boolean       periodic              = false;

    /**
     * Keep reference to memory area
     * 
     * @uml.property name="memArea"
     * @uml.associationEnd 
     * @uml.property name="memArea" multiplicity="(0 1)"
     */
    protected MemoryArea memArea;

    /**
     * Creates a new instance of RTJavaThread with the specified Runnable.
     */
    public RTJavaThread() {
        super();
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling ) {
        super( scheduling );
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling, ReleaseParameters release ) {
        super( scheduling, release );
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling, ReleaseParameters release, MemoryParameters memory, MemoryArea area, ProcessingGroupParameters group, java.lang.Runnable logic ) {
        super( scheduling, release, memory, area, group, null );
        memArea = area;
        setPeriodic();
        runner = logic;
    }

    public RTJavaThread( int iterations ) {
        super();
        this.iterations = initialIterations = iterations;
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling, int iterations ) {
        super( scheduling );
        this.iterations = initialIterations = iterations;
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling, ReleaseParameters release, int iterations ) {
        super( scheduling, release );
        this.iterations = initialIterations = iterations;
        setPeriodic();
    }

    public RTJavaThread( SchedulingParameters scheduling, ReleaseParameters release, MemoryParameters memory, MemoryArea area, ProcessingGroupParameters group, java.lang.Runnable logic, int iterations ) {
        super( scheduling, release, memory, area, group, null );
        memArea = area;
        runner = logic;
        this.iterations = initialIterations = iterations;
        setPeriodic();
    }

    /**
     * sets whether this threadi s periodic or aperiodic based on the kind of ReleaseParamters it as. If ReleaseParameters is null,
     * then we assume aperiodic
     */
    protected void setPeriodic() {
        periodic = (getReleaseParameters() instanceof PeriodicParameters) && (iterations != 0);
    }

    /** Cancels this RTJavaThread by calling its task's cancel() method. */
    public void kill() {
        killed = true;
    }

    public void run() {
        boolean isNextPeriodValid = false;

        for ( int i = 0; i < NUMBEROFERRORSALLOWED; i++ ) {
            restart: {
                while ( !killed && (iterations != 0) ) {
                    if ( periodic ) {
                        isNextPeriodValid = RealtimeThread.waitForNextPeriod();
		                if ( !isNextPeriodValid ) {
		                    // If we reach this point we may have an error
		                    break restart;
		                }
                    }
	                
                    if ( memArea != RealtimeThread.getCurrentMemoryArea() ) {
                        memArea.enter( runner );
                    }
                    else {
                        runner.run();
                    }
                    if ( iterations > 0 ) iterations--;
                }

                //  If the thead has been commanded to stop, break from the
                // for loop.
                if ( killed || (iterations == 0) ) {
                    break;
                }
            }
	    	if ( !killed && (iterations != 0) && !isNextPeriodValid ) {
	    	    System.out.println( "RT WARNING: Invalid next period: iteration = " + (initialIterations - iterations) + ", error #" + i + " of " + NUMBEROFERRORSALLOWED );
	    	}
        } // end for (int i=0; i<numberOfErrors; i++ )
    }
}