/**
 * $Id: TestProperties.java,v 1.1 2004/09/23 15:34:23 mbi Exp $ Copyright 2004, by the California Institute of Technology. ALL
 * RIGHTS RESERVED. United States Government Sponsorship acknowledged. Any commercial use must be negotiated with the Office of
 * Technology Transfer at the California Institute of Technology. This software may be subject to U.S. export control laws and
 * regulations. By accepting this document, the user agrees to comply with all applicable U.S. export laws and regulations. User has
 * the responsibility to obtain export licenses, or other export authority as may be required before exporting such information to
 * foreign countries or providing access to foreign persons.
 */

package suramadu.util.exec;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.StringTokenizer;

import suramadu.util.stats.Counter;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * The TestProperties class supports the loading and maintenance of a properties file with naming convention of &lt;class
 * name&gt;.properties This object is mutable, so properties that are loaded are subject to change by the user
 * <p>
 * 
 * @author Mark B. Indictor
 * @version $Id: TestProperties.java,v 1.1 2004/09/23 15:34:23 mbi Exp $
 */

public abstract class TestProperties extends Properties {
    public static final String INPUT_PREFIX        = "input.";
    public static final String OUTPUT_PREFIX       = "output.";
    public static final String TIMESTAMP           = "timestamp";
    public static final String PLATFORM            = "platform";
    public static final String OSNAME              = "osname";
    public static final String OSVERSION           = "osversion";
    public static final String CLASSNAME           = "classname";
    public static final String HOSTNAME            = "hostname";
    public static final String CPS                 = "cps";
    public static final String PROPERTIESLOADED    = "propertiesfile";
    public static final String RAWDATAFILE         = "rawdatafile";
    public static final String FILEOUTPUTERROR     = "datafileerror";

    public static final int    DATAFILEHEADERWIDTH = 132;

    /** Constant: toString() printout column 1 width. */
    static final int           COL1_WIDTH          = 40;
    /** Constant: toString() printout column 2 width. */
    static final int           COL2_WIDTH          = 20;
    /** Constant: toString() printout column 3 width. */
    static final int           COL3_WIDTH          = 20;

    /**
     * This static method is used to get an array of properties objects for multiple runs of the same Suramadu test.
     * 
     * @param pClass The specific subclass of TestProperties being instantiated.
     * @param aClass The class of the actual test being run.
     * 
     * @return An array of TestProperties objects representing discreet property sets for multiple test runs.
     * @see suramadu.util.exec.TestProperties#TestProperties(Class)
     */
    public static Iterator factory( Class pClass, Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
        return factory( pClass, aClass, (Properties)null );
    }

    /**
     * This static method is used to get an array of properties objects for multiple runs of the same Suramadu test.
     * 
     * @param pClass The specific subclass of TestProperties being instantiated.
     * @param aClass The class of the actual test being run.
     * @param defaultProps
     * 
     * @return An array of TestProperties objects representing discreet property sets for multiple test runs.
     * @see suramadu.util.exec.TestProperties#TestProperties(Class, String[])
     */
    public static Iterator factory( Class pClass, Class aClass, String[] defaultProps ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
        return factory( pClass, aClass, arrayToProperties( aClass, defaultProps ) );
    }

    /**
     * This static method is used to get an array of properties objects for multiple runs of the same Suramadu test.
     * 
     * @param pClass The specific subclass of TestProperties being instantiated.
     * @param aClass The class of the actual test being run.
     * @param defaultProps
     * 
     * @return An array of TestProperties objects representing discreet property sets for multiple test runs.
     * @see suramadu.util.exec.TestProperties#TestProperties(Class, Properties)
     */
    public static Iterator factory( Class pClass, Class aClass, Properties defaultProps ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
        PropertiesFileName[] names = PropertiesFileName.factory( aClass );

        if ( (names == null) || (names.length == 0) ) {
            names = new PropertiesFileName[ 1 ];
            names[ 0 ] = new PropertiesFileName( aClass );
        }
        return new TestProperties.TestPropertiesIterator( pClass, aClass, defaultProps, names );
    }

    /**
     * 
     * TestPropertiesIterator
     * 
     * Specialized Iterator class for late loading of static TestProperties classes.
     * 
     * @author mbi
     * Aug 25, 2004
     */
    static class TestPropertiesIterator implements Iterator {
        Class                pClass;
        Class                aClass;
        Properties           defProps;
        PropertiesFileName[] names;
        int                  idx;
        
        public TestPropertiesIterator( Class pClass, Class aClass, Properties defProps, PropertiesFileName[] names ) {
            this.pClass    = pClass;
            this.aClass    = aClass;
            this.defProps  = defProps;
            this.names     = names;
            this.idx       = 0;
        }
        
        public boolean hasNext() {
            return (idx < names.length);
        }
        
        public Object next() {
            try {
	            TestProperties props;
                props = (TestProperties)pClass.newInstance();
	            props.loadProperties( names[ idx++ ], defProps );
	            return props;
            }
            catch ( Exception e ) {
                throw new NoSuchElementException( "TestPropertiesIterator: " + e );
            }
        }
        
        public void remove() {
            throw new UnsupportedOperationException( "Cannot remove from TestPropertiesIterator" );
        }
    }
    
    /**
     * Environmental information about properties
     */
    private PropertiesFileName pfn;

    /**
     * A boolean indicating whether the properties file was created on this run.
     */
    protected boolean            propertiesCreated;

    /**
     * A more human readable form of the test run's date and time.
     */
    protected String             dateTime;

    /**
     * A prefix for all properties for this instantiation of TestProperties. Typically, this header contains a timestamp, machine
     * id, architectural id, and platform id.
     */
    protected String             propertyPrefix;

    /**
     * A boolean value indicating whether output properties have been set yet. After output properties have been set, input
     * properties can no longer be set.
     */
    private boolean            outputPropsSet   = false;

    /**
     * Hashmap to keep track of the number of output files with the same property name. If one has already been created, then append
     * the count to it, otherwise, leave it pristine.
     * 
     * @uml.property name="filenameMap"
     * @uml.associationEnd @uml.property name="filenameMap" multiplicity="(0 1)" qualifier="key:java.lang.String
     *                     cnt:suramadu.util.stats.Counter"
     */
    private Map                filenameMap      = new HashMap();

    /**
     * Keeps track of how many raw data files were produced by this run.
     */
    private int                rawDataFileCount = 0;

    /**
     * Records the time that the TestProperties instance was created.rtyValidationException
     */
    protected Date               creationTime     = null;

    /**
     * No arg constructor. Use loadProperties(class, defaults) to instantiate class properties at a later time.
     */
    public TestProperties() {
        super();
        creationTime = new Date();
    }

    /**
     * Constructor that takes an a Class object. No default properties are specified.
     * 
     * @param aClass
     *            the name of the class whose properties file is to be loaded
     */
    public TestProperties( Class aClass ) throws PropertyValidationException {
        this( aClass, (Properties)null );
    }

    /**
     * Constructor that takes an a Class object and an array of Strings. This array must have an even length since key and value
     * pairs will be extracted from it.
     * 
     * @param aClass
     *            the name of the class whose properties file is to be loaded
     * @param defaultProps
     *            An array of Strings that will define the contents of the default properties for this instance. It's length must be
     *            even (name value pairs). The even numbered elements (zero based) are names, the odd numbered elements are values.
     * @see #TestProperties( Class aClass, Properties defaultProps )
     */
    public TestProperties( Class aClass, String[] defaultProps ) throws PropertyValidationException {
        this( aClass, arrayToProperties( aClass, defaultProps ) );
    }

    /**
     * Constructor that takes an a Class object and another TestProperties object to act as a source for default properties should
     * the property not exist in the properties file for this class.
     * 
     * @param aClass
     *            the name of the class whose properties file is to be loaded
     * @param defaultProps
     *            if provided (not null), will provide a set of default values for unspecified properties
     */
    public TestProperties( Class aClass, Properties defaultProps ) throws PropertyValidationException {
        this( new PropertiesFileName( aClass ), defaultProps );
    }

    /**
     * Constructor that takes an a Class object and another TestProperties object to act as a source for default properties should
     * the property not exist in the properties file for this class.
     * 
     * @param aPfn
     *            A PropertiesFileName class for the properties file to be read.
     * @param defaultProps
     *            if provided (not null), will provide a set of default values for unspecified properties
     */
    public TestProperties( PropertiesFileName aPfn, Properties defaultProps ) throws PropertyValidationException {
        this();
        initTestProperties( aPfn, defaultProps );
    }

    /**
     * Load class properties
     * 
     * @param aClass
     * @param defaultProps
     * @throws PropertyValidationException
     *             If a property is malformed
     */
    public void loadProperties( Class aClass, Properties defaultProps ) throws PropertyValidationException {
        initTestProperties( new PropertiesFileName( aClass ), defaultProps );
    }

    /**
     * Load class properties
     * 
     * @param aClass
     * @param defaultProps
     * @throws PropertyValidationException
     */
    public void loadProperties( Class aClass, String[] defaultProps ) throws PropertyValidationException {
        initTestProperties( new PropertiesFileName( aClass ), arrayToProperties( aClass, defaultProps ) );
    }

    /**
     * Load class properties
     * 
     * @param aPfn
     * @param defaultProps
     * @throws PropertyValidationException
     *             If a property is malformed
     */
    public void loadProperties( PropertiesFileName aPfn, Properties defaultProps ) throws PropertyValidationException {
        initTestProperties( aPfn, defaultProps );
    }

    /**
     * Saves argument values in static class members. 7/12/04: This should be an abstract class, but for the sake of not breaking
     * the build, I hav chosen to implement it as an empty regular method.
     * 
     * @throws NumberFormatException
     *             If a property is misformed.
     */
    protected abstract void loadArguments() throws PropertyValidationException;

    /**
     * Saves argument values in static class members. 7/12/04: This should be an abstract class, but for the sake of not breaking
     * the build, I hav chosen to implement it as an empty regular method.
     * 
     * @throws NumberFormatException
     *             If a property is misformed.
     */
    protected abstract void validateArguments() throws PropertyValidationException;

    /**
     * Get the name of the package containing for the current properties file.
     * 
     * @return the name of the package containing for the current properties file
     * @uml.property name="packageName"
     */
    public String getPackageName() {
        return pfn.packageName;
    }

    /**
     * Get the fully qualified path of the properties file acting as the source to this TestProperties instantiation.
     * 
     * @return the fully qualified path of the properties file
     */
    public String getPropertiesFilename() {
        return pfn.fullPropertiesFileName;
    }

    /**
     * Get the full package and class name of the class for which this properties object was created.
     * 
     * @return the full package and class name of the class for which this properties object was created.
     * @uml.property name="className"
     */
    public String getClassName() {
        return pfn.fullClassName;
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value, or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    protected int getIntProperty( String key ) throws PropertyValidationException {
        try {
            return Integer.parseInt( getProperty( key ).trim() );
        }
        catch ( NumberFormatException nfe ) {
            throw new PropertyValidationException( "Badly formatted number: " + key + "=\"" + getProperty( key ) + "\"" );
        }
        catch ( NullPointerException nfe ) {
            throw new PropertyValidationException( "Property not found: " + key + "=\"" + getProperty( key ) + "\"" );
        }
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    protected long getLongProperty( String key ) throws PropertyValidationException {
        try {
            return Long.parseLong( getProperty( key ).trim() );
        }
        catch ( NumberFormatException nfe ) {
            throw new PropertyValidationException( "Badly formatted number: " + key + "=\"" + getProperty( key ) + "\"" );
        }
        catch ( NullPointerException nfe ) {
            throw new PropertyValidationException( "Property not found: " + key + "=\"" + getProperty( key ) + "\"" );
        }
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    protected float getFloatProperty( String key ) throws PropertyValidationException {
        try {
            return Float.parseFloat( getProperty( key ).trim() );
        }
        catch ( NumberFormatException nfe ) {
            throw new PropertyValidationException( "Badly formatted number: " + key + "=\"" + getProperty( key ) + "\"" );
        }
        catch ( NullPointerException nfe ) {
            throw new PropertyValidationException( "Property not found: " + key + "=\"" + getProperty( key ) + "\"" );
        }
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    protected double getDoubleProperty( String key ) throws PropertyValidationException {
        try {
            return Double.parseDouble( getProperty( key ).trim() );
        }
        catch ( NumberFormatException nfe ) {
            throw new PropertyValidationException( "Badly formatted number: " + key + "=\"" + getProperty( key ) + "\"" );
        }
        catch ( NullPointerException nfe ) {
            throw new PropertyValidationException( "Property not found: " + key + "=\"" + getProperty( key ) + "\"" );
        }
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    protected boolean getBooleanProperty( String key ) throws PropertyValidationException {
        try {
            return Boolean.valueOf( getProperty( key ).trim() ).booleanValue();
        }
        catch ( Exception e ) {
            throw new PropertyValidationException( "Badly formatted boolean: " + key + "=\"" + getProperty( key ) + "\"" );
        }
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value, or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public String getOutputProperty( String key ) throws PropertyValidationException {
        return getProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value, or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public int getIntOutputProperty( String key ) throws PropertyValidationException {
        return getIntProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public long getLongOutputProperty( String key ) throws PropertyValidationException {
        return getLongProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public float getFloatOutputProperty( String key ) throws PropertyValidationException {
        return getFloatProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public double getDoubleOutputProperty( String key ) throws PropertyValidationException {
        return getDoubleProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public boolean getBooleanOutputProperty( String key ) throws PropertyValidationException {
        return getBooleanProperty( OUTPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value, or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public String getInputProperty( String key ) throws PropertyValidationException {
        return getProperty( INPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value, or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public int getIntInputProperty( String key ) throws PropertyValidationException {
        return getIntProperty( INPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public long getLongInputProperty( String key ) throws PropertyValidationException {
        return getLongProperty( INPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public float getFloatInputProperty( String key ) throws PropertyValidationException {
        return getFloatProperty( INPUT_PREFIX + key );
    }

    /**
     * The get*Property() methods are the normal methods used to retrieve a property's value. All property names are forced to
     * conform with the 'package1.package2.classname.propertyname' convention, but may be specified simply with 'propertyname'.
     * 
     * @param key
     *            the property name
     * @return value or 0 if the specified property does not exist.
     * @throws NumberFormatException
     *             If there is a parsing error.
     */
    public double getDoubleInputProperty( String key ) throws PropertyValidationException {
        return getDoubleProperty( INPUT_PREFIX + key );
    }

    /**
     * @param key
     * @return value of specified boolean input property
     */
    public boolean getBooleanInputProperty( String key ) throws PropertyValidationException {
        return getBooleanProperty( INPUT_PREFIX + key );
    }

    /**
     * Recursive procedures
     */
    public Enumeration propertyNames() {
        return propertyNames( false );
    }

    public Enumeration propertyNames( boolean shortNames ) {
        Hashtable list = new Hashtable();

        propertyNames( this, null, list, shortNames );

        return list.keys();
    }

    protected void propertyNames( Properties p, String keyPrefix, Hashtable list, boolean shortNames ) {
        if ( null == p ) return;

        for ( Enumeration e = p.keys(); e.hasMoreElements(); ) {
            String key = (String)e.nextElement();
            String longKey = (null == keyPrefix) ? key : keyPrefix + "." + key;
            try {
                propertyNames( (Properties)p.get( key ), longKey, list, shortNames );
            }
            catch ( ClassCastException ex ) {
                list.put( shortNames ? key : longKey, p.get( key ) );
            }
            catch ( NullPointerException ex ) {
            }
        }
    }

    protected Properties getSubProperties( String key ) {
        int lastDot = 0;
        int nextDot = 0;
        int length = key.length();
        Properties p = this;

        if ( (null != propertyPrefix) && !key.startsWith( propertyPrefix ) ) {
            return getSubProperties( propertyPrefix + key );
        }

        do {
            nextDot = key.indexOf( '.', lastDot );
            if ( nextDot < 0 ) nextDot = length;

            String subKey = key.substring( lastDot, nextDot );
            try {
                p = (Properties)p.get( subKey );
            }
            catch ( ClassCastException e ) {
                p = null;
                break;
            }
            lastDot = nextDot;
        }
        while ( ++lastDot < length );
        return p;
    }

    /**
     * Get the value for the specified property. As with its superclass java.util.Properties, it will recursively search through all
     * default property sets until it finds a value or runs out of default sets.
     * 
     * @param key
     *            the name of the property
     * @return the value of the property, or null if the specified property does not exist.
     * @see java.util.Properties
     */
    public String getProperty( String key ) {
        if ( (null != propertyPrefix) && key.startsWith( propertyPrefix ) ) {
            return getPropertyRecursive( key );
        }
        else {
            return getPropertyRecursive( propertyPrefix + key );
        }
    }

    /**
     * Overloaded from java.util.Properties. It is intercepted so that the package1.package2.classname.propertyname' property naming
     * constraints can be enforced.
     * 
     * @param key
     *            The property name whose value will be set.
     * @param value
     *            The value that will be set for the named property.
     * @return the previous value of this property.
     * @see java.util.Properties
     */
    public Object setProperty( String key, String value ) {
        if ( (null != propertyPrefix) && key.startsWith( propertyPrefix ) ) {
            return setPropertyRecursive( key, value );
        }
        else {
            return setPropertyRecursive( propertyPrefix + key, value );
        }
    }

    protected Object setProperty( String key, boolean value ) {
        return setProperty( key, (value ? "true" : "false") );
    }

    protected Object setProperty( String key, int value ) {
        return setProperty( key, Integer.toString( value ) );
    }

    protected Object setProperty( String key, long value ) {
        return setProperty( key, Long.toString( value ) );
    }

    protected Object setProperty( String key, float value ) {
        return setProperty( key, Float.toString( value ) );
    }

    protected Object setProperty( String key, double value ) {
        return setProperty( key, Double.toString( value ) );
    }

    /**
     * setting output properties
     */
    public Object setOutputProperty( String key, String value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, Number value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value.doubleValue() );
    }

    public Object setOutputProperty( String key, boolean value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, int value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, long value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, float value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, double value ) {
        outputPropsSet = true;
        return setProperty( OUTPUT_PREFIX + key, value );
    }

    public Object setOutputProperty( String key, Collection c ) {
        return setOutputProperty( key, c, true );
    }

    public Object setOutputProperty( String key, Collection c, boolean newFile ) {
        String fName = null;
        boolean fileCreated = false;

        if ( newFile ) {
            Counter cnt = (Counter)filenameMap.get( key );
            if ( null == cnt ) {
                cnt = new Counter( key );
                fName = pfn.propertiesFileName + "." + propertyPrefix + key;
            }
            else {
                cnt.postInc();
                fName = pfn.propertiesFileName + "." + propertyPrefix + key + cnt.getValue().toString();
            }
            filenameMap.put( key, cnt );

            System.out.println( "Output File Name: " + fName );

            try {
                PrintWriter out = new PrintWriter( new FileOutputStream( fName ) );
                int count = 0;

                out.println( pad( DATAFILEHEADERWIDTH, '#' ) );
                out.println( pad( "# Suramadu Test   : " + propertyPrefix.substring( 0, propertyPrefix.length() - 1 ), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Test Class      : " + getProperty( CLASSNAME ), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Test Date/Time  : " + dateTime, DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Platform        : " + getProperty( PLATFORM ) + " (" + getProperty( CPS ) + ")", DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Operating System: " + getProperty( OSNAME ).replace( '_', ' ' ) + " version: " + getProperty( OSVERSION ), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Hostname        : " + getProperty( HOSTNAME ), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( "# Property File   : " + getProperty( PROPERTIESLOADED ), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );

                // dump input properties to raw data output file.
                BufferedReader in = new BufferedReader( new StringReader( toHString( getSubProperties( "input" ), "input", 0 ) ) );
                String line = in.readLine();
                out.println( pad( "# Input Properties: " + ((line == null) ? "<none>" : line), DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                if ( null != line ) {
                    while ( (line = in.readLine()) != null ) {
                        out.println( pad( "#                 : " + line, DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                    }
                }

                out.println( pad( "# Output Property : \"" + key + "\"", DATAFILEHEADERWIDTH - 2, ' ' ) + " #" );
                out.println( pad( DATAFILEHEADERWIDTH, '#' ) );
                for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
                    out.println( i.next().toString() );
                }
                out.println();
                out.flush();
                out.close();
                setOutputProperty( RAWDATAFILE + (++rawDataFileCount), fName );
                fileCreated = true;
            }
            catch ( IOException e ) {
                e.printStackTrace();
                setOutputProperty( FILEOUTPUTERROR, "Error creating data file: " + fName + ": " + e.toString() );
            }
        }

        if ( !fileCreated ) {
            int count = 0;
            for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
                setOutputProperty( key + "." + count, i.next().toString() );
            }
        }
        return getProperty( key );
    }

    public Object setOutputProperty( String key, int[] values ) {
        return setOutputProperty( key, StatPack.toCollection( values ) );
    }

    public Object setOutputProperty( String key, int[] values, boolean newFile ) {
        return setOutputProperty( key, StatPack.toCollection( values ), newFile );
    }

    public Object setOutputProperty( String key, long[] values ) {
        return setOutputProperty( key, StatPack.toCollection( values ) );
    }

    public Object setOutputProperty( String key, long[] values, boolean newFile ) {
        return setOutputProperty( key, StatPack.toCollection( values ), newFile );
    }

    public Object setOutputProperty( String key, float[] values ) {
        return setOutputProperty( key, StatPack.toCollection( values ) );
    }

    public Object setOutputProperty( String key, float[] values, boolean newFile ) {
        return setOutputProperty( key, StatPack.toCollection( values ), newFile );
    }

    public Object setOutputProperty( String key, double[] values ) {
        return setOutputProperty( key, StatPack.toCollection( values ) );
    }

    public Object setOutputProperty( String key, double[] values, boolean newFile ) {
        return setOutputProperty( key, StatPack.toCollection( values ), newFile );
    }

    public Object setInputProperty( String key, String value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    /**
     * setting input properties
     */
    public Object setInputProperty( String key, Number value ) {
        outputPropsSet = true;
        return setProperty( INPUT_PREFIX + key, value.doubleValue() );
    }

    public Object setInputProperty( String key, boolean value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    public Object setInputProperty( String key, int value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    public Object setInputProperty( String key, long value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    public Object setInputProperty( String key, float value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    public Object setInputProperty( String key, double value ) throws PropertyValidationException {
        if ( outputPropsSet ) throw new PropertyValidationException( "Cannot set input properties after setting output properties." );
        return setProperty( INPUT_PREFIX + key, value );
    }

    /**
     * Get the value for the specified property. As with its superclass java.util.Properties, it will recursively search through all
     * default property sets until it finds a value or runs out of default sets.
     * 
     * @param key
     *            the name of the property
     * @return the value of the property, or null if the specified property does not exist.
     * @see java.util.Properties
     */
    public Object _get( Object key ) {
        if ( key.toString().startsWith( propertyPrefix ) ) {
            return get( key );
        }
        else {
            return get( propertyPrefix + key.toString() );
        }
    }

    /**
     * Overloaded from java.util.Properties. It is intercepted so that the package1.package2.classname.propertyname' property naming
     * constraints can be enforced.
     * 
     * @param key
     *            The property name whose value will be set.
     * @param value
     *            The value that will be set for the named property.
     * @return the previous value of this property.
     * @see java.util.Properties
     */
    public Object _put( Object key, Object value ) {
        if ( key.toString().startsWith( propertyPrefix ) ) {
            return put( key, value );
        }
        else {
            return put( propertyPrefix + key.toString(), value );
        }
    }

    /**
     * Overloads Hashtable.putAll(). Copies all of the mappings from the specified Map to this Hashtable These mappings will replace
     * any mappings that this Hashtable had for any of the keys currently in the specified Map.
     * 
     * @param t
     *            Mappings to be stored in this map.
     * @throws NullPointerException
     *             if the specified map is null.
     * @see java.util.Hashtable
     */
    public synchronized void putAll( Map t ) {
        Iterator i = t.entrySet().iterator();
        while ( i.hasNext() ) {
            Map.Entry e = (Map.Entry)i.next();
            _put( e.getKey(), e.getValue() );
        }
    }

    /**
     * Recursive procedures
     */

    /**
     * Get the value for the specified property. As with its superclass java.util.Properties, it will recursively search through all
     * default property sets until it finds a value or runs out of default sets.
     * 
     * @param key
     *            the name of the property
     * @return the value of the property, or null if the specified property does not exist.
     * @see java.util.Properties
     */
    protected String setPropertyRecursive( String key, String value ) {
        int lastDot = 0;
        int nextDot = 0;
        int length = key.length();
        Properties thisMap = this;
        String retVal = null;

        do {
            nextDot = key.indexOf( '.', lastDot );
            if ( nextDot < 0 ) nextDot = length;

            String subKey = key.substring( lastDot, nextDot );
            Properties nextMap = null;
            try {
                nextMap = (Properties)thisMap.get( subKey );
            }
            catch ( ClassCastException e ) {
            }

            // if we get here, then there was no previous entry
            if ( null == nextMap ) {
                if ( (nextDot + 1) < length ) {
                    nextMap = new Properties();
                    thisMap.put( subKey, nextMap );
                }
                else {
                    retVal = (String)thisMap.setProperty( subKey, value );
                    break;
                }
            }
            lastDot = nextDot;
            thisMap = nextMap;
        }
        while ( ++lastDot < length );
        return retVal;
    }

    /**
     * Get the value for the specified property. As with its superclass java.util.Properties, it will recursively search through all
     * default property sets until it finds a value or runs out of default sets.
     * 
     * @param key
     *            the name of the property
     * @return the value of the property, or null if the specified property does not exist.
     * @see java.util.Properties
     */
    protected String getPropertyRecursive( String key ) {
        int lastDot = 0;
        int nextDot = 0;
        int length = key.length();
        Properties p = this;
        String retVal = null;

        do {
            nextDot = key.indexOf( '.', lastDot );
            if ( nextDot < 0 ) nextDot = length;

            String subKey = key.substring( lastDot, nextDot );
            try {
                p = (Properties)p.get( subKey );
            }
            catch ( ClassCastException e ) {
                retVal = p.getProperty( subKey );
                break;
            }
            catch ( NullPointerException e ) {
                break;
            }
            lastDot = nextDot;
        }
        while ( ++lastDot < length );
        return retVal;
    }

    /**
     * Writes the output file
     */
    public void close() throws IOException {
        String name = pfn.propertiesFileName + "." + propertyPrefix + "SUMMARY";
        System.out.println( "Output File Name: " + name );
        PrintWriter out = new PrintWriter( new FileOutputStream( name ) );
        out.println( toString() );
        out.flush();
        out.close();
        System.out.println( "Done: Output ---> " + pfn.propertiesFileName + "." + propertyPrefix + "*" );
    }


    public void read( InputStream is ) throws IOException {
        BufferedReader reader = new BufferedReader( new InputStreamReader( is ) );

        /*
         * String line = null; while ((line = reader.readLine()) != null) { // Read in each property from file }
         */
        load( is );
    }

    /**
     * Displays a nicely formatted summary of test and its current working property list.
     * 
     * @return A String formatted for printing.
     */
    public String toString() {
        return toHString( this, null, -1 );
    }

    protected String toHString( Properties p, String key, int level ) {
        StringWriter sw = new StringWriter();
        PrintWriter  pw = new PrintWriter( sw );

        if ( null != key ) {
            pw.println( pad( level * 4, ' ' ) + key + "={" );
        }

        for ( Enumeration e = p.keys(); e.hasMoreElements(); ) {
            String subKey = (String)e.nextElement();
            String longKey = (null != key) ? key + "." + subKey : subKey;
            try {
                Properties subP = (Properties)p.get( subKey );
                pw.println( toHString( subP, subKey, level + 1 ) );
            }
            catch ( ClassCastException ex ) {
                pw.println( pad( (level + 1) * 4, ' ' ) + subKey + "=\"" + p.getProperty( subKey ) + "\"" );
            }
            catch ( NullPointerException ex ) {
                pw.println( pad( (level + 1) * 4, ' ' ) + subKey + "=\"" + p.getProperty( subKey ) + "\"" );
            }
        }

        if ( null != key ) {
            pw.print( pad( level * 4, ' ' ) + "}" );
        }

        return sw.toString();
    }

    /**
     * This is a helper method that translates an array of Strings into a TestProperties object. The length of the array is enforced
     * to be even, and a message is displayed on standard error and the defaults are not created if this is not the case.
     * 
     * @param aClass
     *            the name of the class whose properties file is to be loaded
     * @param props
     *            An array of Strings that will define the contents of the default properties for this instance. It's length must be
     *            even (name value pairs). The even numbered elements (zero based) are names, the odd numbered elements are values.
     * @return a new instance of TestProperties containing the default properties specified in the props argument.
     */
    public static Properties arrayToProperties( Class aClass, String[] props ) throws PropertyValidationException {
        Properties defaultProps = null;

        if ( null != props ) {
            int length = props.length;
            if ( ((length / 2) * 2) == length ) { // check for an even number of parameters in array
                defaultProps = new Properties();
                for ( int i = 0; i < length; i += 2 ) {
                    defaultProps.setProperty( INPUT_PREFIX + props[ i ], props[ i + 1 ] );
                }
            }
            else {
                throw new PropertyValidationException( "WARNING: Property array has an odd number of elements. " + "Default values will be ignored." );
            }
        }
        return defaultProps;
    }

    /**
     * Used by constructors to initialize local member fields and load properties from the file system. If properties file does not
     * exist, no exception is thrown.
     * 
     * @param aClass
     *            the Class object of the main test class. This is used to generate the appropriate file and package paths, and
     *            properties file name.
     */
    private void initTestProperties( PropertiesFileName pfn, Properties defaultProps ) throws PropertyValidationException {
        this.pfn = pfn;
        try {
            InputStream is = null;

            propertyPrefix = new SimpleDateFormat( "SSssmmHHddMMyyyy" ).format( creationTime ) + ".";
            dateTime = new SimpleDateFormat( "MM/dd/yyyy @ HH:mm:ss.SS" ).format( creationTime );

            // initialize defaults
            if ( null != defaultProps ) {
                for ( Enumeration e = defaultProps.propertyNames(); e.hasMoreElements(); ) {
                    String key = (String)e.nextElement();
                    setProperty( key, defaultProps.getProperty( key ) );
                }
            }

            // initialize identification properties.
            setProperty( PLATFORM, System.getProperty( "os.arch" ).replace( ' ', '_' ) );
            setProperty( OSNAME, System.getProperty( "os.name" ).replace( ' ', '_' ) );
            setProperty( OSVERSION, System.getProperty( "os.version" ).replace( ' ', '_' ) );
            setProperty( CLASSNAME, pfn.fullClassName );
            if ( 0L != Time.cps ) {
                setProperty( CPS, (Time.cps > 1e9) ? Double.toString( ((double)Time.cps) / 1e9 ) + " Ghz" : Double.toString( ((double)Time.cps) / 1e6 ) + " Mhz" );
            }
            else {
                setProperty( CPS, "*** Unable to determine clock frequency of CPU ***" );
            }
            /*
            try {
                setProperty( HOSTNAME, InetAddress.getLocalHost().getHostName() );
            }
            catch ( UnknownHostException e ) {
                setProperty( HOSTNAME, "UnknownHost" );
            }
*/
            try {
                is = openPropertiesFile( pfn.fullPropertiesFileName, defaultProps );
                if ( null == is ) {
                    setProperty( PROPERTIESLOADED, pfn.shortPropertiesFileName + ": ERROR" );
                    return;
                }

                Properties p = new Properties();
                p.load( is );
                for ( Enumeration e = p.propertyNames(); e.hasMoreElements(); ) {
                    String key = (String)e.nextElement();
                    setProperty( INPUT_PREFIX + key, p.getProperty( key ) );
                }
                setProperty( PROPERTIESLOADED, pfn.shortPropertiesFileName + (propertiesCreated ? ": CREATED" : ": OK") );
            }
            catch ( IOException e ) {
                setProperty( PROPERTIESLOADED, pfn.shortPropertiesFileName + ": ERROR" );
                throw new PropertyValidationException( e );
            }
        }
        finally {
            /**
             * call back up into overriding subclass for any late initializations.
             */
            loadArguments();
            validateArguments();
        }
    }

    private InputStream openPropertiesFile( String pFileName, Properties defaultProps ) throws IOException {
        InputStream is = null;
        propertiesCreated = false;

        /*
         * try and open existing file if it exists
         */
        try {
            is = new FileInputStream( pFileName );
        }
        catch ( Exception e ) {
        }

        /*
         * File does not exist. Try and create it from default properties
         */
        if ( null == is ) {
            /*
             * create a new file and all required directories in the path.
             * 
             * create path to file if it doesn't already exist
             */
            new File( pFileName.substring( 0, pFileName.lastIndexOf( PropertiesFileName.fileSeparatorString ) ) ).mkdirs();
            File f = new File( pFileName );
            f.createNewFile(); // create file

            /**
             * strip off input prefix from properties before writing out (if they exist)
             */
            Properties outProps = new Properties();
            for ( Enumeration e = defaultProps.propertyNames(); e.hasMoreElements(); ) {
                String fullProp = (String)e.nextElement();
                String prop = fullProp.startsWith( INPUT_PREFIX ) ? fullProp.substring( INPUT_PREFIX.length() ) : fullProp;
                outProps.setProperty( prop, defaultProps.getProperty( fullProp ) );
            }

            /**
             * the file name has been assembled. Write out the properties.
             */
            FileOutputStream os = new FileOutputStream( f );
            outProps.store( os, "Default Propertie Auto-Generated by: " + pfn.fullClassName );
            os.close();
            propertiesCreated = true;
            System.out.println( "WARNING: Properties File Created:" );

            /**
             * Now, re-open file for input.
             */
            is = new FileInputStream( pFileName );
        }
        return is;
    }

    /**
     * Creates a new String object that contains the 'count' number of 'c' characters.
     * 
     * @param count
     *            the number of characters (total) to output. The return value is guaranteed not to be longer than this value.
     * @param c
     *            the character to use to pad
     * @return the resulting padded string.
     * @see String pad( String s, int count, char c )
     */
    private String pad( int count, char c ) {
        return pad( null, count, c );
    }

    /**
     * Creates a new String object that contains original value of 's', padded to 'count' characters with spaces. The return value
     * is guaranteed not to be longer than the value of 'count'. If the input is longer, then it will be truncated.
     * 
     * @param s
     *            the string to pad
     * @param count
     *            the number of characters (total) to output. The return value is guaranteed not to be longer than this value. If
     *            the input is longer, then it will be truncated.
     * @return the resulting padded string.
     * @see String pad( String s, int count, char c )
     */
    private String pad( String s, int count ) {
        return pad( s, count, ' ' );
    }

    /**
     * Creates a new String object that contains original value of 's', padded to 'count' characters with 'c'. The return value is
     * guaranteed not to be longer than the value of 'count'. If the input is longer, then it will be truncated.
     * 
     * @param s
     *            the string to pad
     * @param count
     *            the number of characters (total) to output. The return value is guaranteed not to be longer than this value. If
     *            the input is longer, then it will be truncated.
     * @param c
     *            the character to use to pad
     * @return the resulting padded string.
     * @see String pad( String s, int count, char c )
     */
    private String pad( String s, int count, char c ) {
        if ( (null != s) && (s.length() >= count) ) return s.substring( 0, count );

        StringBuffer sb = new StringBuffer();
        if ( (null != s) && (s.length() > 0) ) {
            sb.append( s );
        }

        for ( int i = sb.length(); i < count; i++ ) {
            sb.append( c );
        }
        return sb.toString();
    }
}

class PropertiesFileName {
    private static final String PROPERTIES_DIR = "properties";
    private static final String PROPERTIES     = ".properties";
    private static final int    PROPERTIES_LEN = PROPERTIES.length();
    
    /**
     * Local copy of the underlying OS's file separator character for constructing fully qualified file names.
     */
    public static String fileSeparatorString;

    /**
     * Local copy of the underlying OS's file separator character for constructing fully qualified file names.
     */
    public static char   fileSeparatorChar;
    
    /**
     * Local copy of the underlying OS's file separator character for separating multiple search paths.
     */
    public static String pathSeparatorString;

    /**
     * Local copy of the underlying OS's file separator character for separating multiple search paths.
     */
    public static char   pathSeparatorChar;
    
    /**
     * The value of the "user.home" System property.
     */
    public static String userDir;

    /**
     * The value of the "user.dir" System property.
     */
    public static String pwDir;

    /**
     * The fully qualified, file system name of directory in which the class is located.
     */
    public static String classPath;
    
    /**
     * Properties file path as specified on command line
     */
    public static String specifiedPropertiesFilePath = null;
    
    /**
     * Static initialization
     */
    static {
        String       javaClassPath   = System.getProperty( "java.class.path" );
        StringBuffer systemClassPath = new StringBuffer();
        
        userDir                      = System.getProperty( "user.home" );
        pwDir                        = System.getProperty( "user.dir" );
        specifiedPropertiesFilePath  = System.getProperty( "suramadu.properties.path" );
        fileSeparatorString          = System.getProperty( "file.separator" );
        fileSeparatorChar            = fileSeparatorString.charAt( 0 );
        pathSeparatorString          = System.getProperty( "path.separator" );
        pathSeparatorChar            = pathSeparatorString.charAt( 0 );

        if ( specifiedPropertiesFilePath != null ) {
    	    systemClassPath.append( specifiedPropertiesFilePath );
            systemClassPath.append( pathSeparatorChar );
        }
        systemClassPath.append( PROPERTIES_DIR );
        if ( javaClassPath != null ) {
            systemClassPath.append( pathSeparatorChar );
            systemClassPath.append( javaClassPath );
        }
        classPath = systemClassPath.toString();
    }

    /**
     * This is the fully qualified package path of the class for which the properties are being retrieved. This is derived from
     * className.
     * 
     * @uml.property name="packageName"
     */
    public String packageName;

    /**
     * The fully qualified (reletive to the class-loader's class path) name of the properties file to load
     * (not including the .properties extension).
     * This is derived from the className.
     */
    public String propertiesFileName;

    /**
     * The fully qualified (reletive to the class-loader's class path) name of the properties file to load
     * (including the .properties extension).
     * This is derived from the className.
     */
    public String fullPropertiesFileName;

    /**
     * The fully qualified (reletive to the class-loader's class path) path of the properties file to load.
     * This is derived from the className.
     */
    public String propertiesFilePath;

    /**
     * The fully qualified (reletive to the class-loader's class path) of the properties file to load. This is derived from the
     * className.
     */
    public String shortPropertiesFileName;

    /**
     * The fully qualified, dot-separated name of the test class requesting its properties be read. The class is passed as a
     * constructor argument.
     * 
     * @uml.property name="className"
     */
    public String fullClassName;
    
    /**
     * The class name without package qualification
     */
    public String className;

    /**
     * The fully qualified, file system name of class file for class that is requesting its properties be read.
     */
    public String classFileName;
    
    /**
     * File modifier for multiple properties file support.
     */
    public String fileMod;
    
    /**
     * 
     * @param aClass
     */
    static public PropertiesFileName[] factory( Class aClass ) {
        final PropertiesFileName pfn  = new PropertiesFileName( aClass );
        File f = new File( pfn.propertiesFilePath );
        File[] files = f.listFiles( new FileFilter() {
            public boolean accept( File pathName ) {
                try {
                    return pathName.getName().startsWith( pfn.className ) && pathName.getName().endsWith( PROPERTIES );
                }
                catch ( Exception e ) {
                    return false;
                }
            }
        });
        
        PropertiesFileName[] names = null;
        if ( (files == null) || (files.length == 0 ) ) {
            names = new PropertiesFileName[ 1 ];
            names[ 0 ] = new PropertiesFileName( aClass );
        }
        else {
	        names = new PropertiesFileName[ files.length ];
	        for ( int i = 0; i < names.length; i++ ) {
	            names[ i ] = new PropertiesFileName( aClass, files[ i ].getAbsolutePath() );
	        }
        }
        return names;
    }

    /**
     * Constructor
     * 
     * @param aClass
     */
    public PropertiesFileName( Class aClass ) {
        this( aClass, null );
    }
    
    /**
     * Constructor
     * 
     * @param aClass
     */
    public PropertiesFileName( Class aClass, String mod ) {
        createFromClass( aClass, mod );
    }
    
    /**
     * Creates a new properties file from properties passed in as default properties This method sets up several global data
     * members.
     * 
     * @param aClass
     *            the class for which to determine the property file name.
     * @return a string representing the fully qualified path of the properties file.
     */
    private void createFromClass( Class aClass ) {
        createFromClass( aClass, null );
    }
    
    private void createFromClass( Class aClass, String mod ) {
        boolean found = false;
        
        fullClassName           = aClass.getName();
        classFileName           = fullClassName.replace( '.', fileSeparatorChar );
    	className               = fullClassName.substring( fullClassName.lastIndexOf( '.' ) + 1 );
        packageName             = fullClassName.substring( 0, fullClassName.lastIndexOf( '.' ) );
        fullPropertiesFileName  = fullClassName.replace( '.', fileSeparatorString.charAt( 0 ) ) + PROPERTIES;
        
        if ( null == specifiedPropertiesFilePath ) {
	        for ( StringTokenizer t = new StringTokenizer( classPath, pathSeparatorString ); t.hasMoreTokens(); ) {
	            StringBuffer tok = new StringBuffer( t.nextToken() );
	            int len = tok.length();
	            if ( (len > 0) && (tok.charAt( len - 1 ) != fileSeparatorChar) ) {
	                tok.append( fileSeparatorChar );
	            }
	            String classPath = tok.toString();
	
	            tok.append( classFileName );
	            tok.append( ".class" );
	            String classFileName = tok.toString();
	
	            /**
	             * Try and open the class file for reading to determine if it exists.
	             */
	            try {
	                new FileReader( classFileName ).close();
	            }
	            catch ( Exception e ) {
	                //                System.out.println( " error opening: " + classFileName );
	                /*
	                 * File does not exist or is unreadable -- try the next one.
	                 */
	                continue;
	            }
	
	            /**
	             * if we get here, then the file exists.
	             */
	            tok.setLength( 0 );
	            tok.append( classPath );
	            tok.append( fullPropertiesFileName );
	            fullPropertiesFileName = tok.toString();
	            found = true;
	            break;
	        }
        }

        /*
         * Check to see if class file was found. If not, we're probably executing from a .jar file
         * and need to find a different place for the properties file.
         * That place is ./properties/<package directory>/<classname>
         */
        if ( found ) {
	        propertiesFilePath     = fullPropertiesFileName.substring( 0, fullPropertiesFileName.lastIndexOf( fileSeparatorChar ) );
	        propertiesFileName     = fullPropertiesFileName.substring( propertiesFilePath.length() + 1, fullPropertiesFileName.lastIndexOf( "." ) );
        }
        else {
            if (specifiedPropertiesFilePath == null) {
                propertiesFilePath = pwDir + fileSeparatorString + PROPERTIES_DIR + fileSeparatorString + fullPropertiesFileName.substring( 0, fullPropertiesFileName.lastIndexOf( fileSeparatorChar ) );
            }
            else {
                propertiesFilePath = specifiedPropertiesFilePath + fileSeparatorString + fullPropertiesFileName.substring( 0, fullPropertiesFileName.lastIndexOf( fileSeparatorChar ) );
            }
	        propertiesFileName     = fullPropertiesFileName.substring( fullPropertiesFileName.lastIndexOf( fileSeparatorChar ) + 1, fullPropertiesFileName.length() - PROPERTIES_LEN );
	        fullPropertiesFileName = propertiesFilePath + fileSeparatorString + propertiesFileName + PROPERTIES;
        }

        /*
         * Extract any properties file modifier for multiple file support
         */
        if ( (mod != null) && (mod.length() > 0) ) {
            int start = mod.indexOf( className ) + className.length();
            int end   = mod.length() - PROPERTIES_LEN;
            fileMod = mod.substring( start, end );
            propertiesFileName = className + fileMod;
	        fullPropertiesFileName = propertiesFilePath + fileSeparatorString + propertiesFileName + PROPERTIES;
        }

        /*
         * replace prefix with '~' if possible.
         */
        if ( fullPropertiesFileName.startsWith( userDir ) ) {
            shortPropertiesFileName = "~" + fullPropertiesFileName.substring( userDir.length() );
        }
        else {
            shortPropertiesFileName = fullPropertiesFileName;
        }
    }

    public String toString() {
        StringWriter s   = new StringWriter();
        PrintWriter  out = new PrintWriter( s );
        out.println( "fileSeparatorString     = " + fileSeparatorString );
        out.println( "fileSeparatorChar       = " + fileSeparatorChar );
        out.println( "userDir                 = " + userDir );
        out.println( "pwDir                   = " + pwDir );
        out.println( "classPath               = " + classPath );
        out.println( "packageName             = " + packageName );
        out.println( "className               = " + className );
        out.println( "fullClassName           = " + fullClassName );
        out.println( "classFileName           = " + classFileName );
        out.println( "propertiesFilePath      = " + propertiesFilePath );
        out.println( "fileMod                 = " + fileMod );
        out.println( "propertiesFileName      = " + propertiesFileName );
        out.println( "shortPropertiesFileName = " + shortPropertiesFileName );
        out.print( "fullPropertiesFileName  = " + fullPropertiesFileName );
        return s.toString();
    }
}