/*
 * Counter.java
 *
 * Created on August 15, 2003, 6:55 AM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.stats;

import java.util.Map;

/**
 * A counter object suitable for storing in a Map
 * 
 * @author mbi
 */
public class Counter implements Map.Entry, Comparable {
    /** The key used when putting in a Map. */
    private Object key;
    /** The value of the counter itself */
    private int    value;

    /**
     * Creates a new instance of Counter with the specified key
     * 
     * @param key
     *            the key to use to put in a Map.
     */
    public Counter( Object key ) {
        this( key, 0 );
    }

    /**
     * Creates a new instance of Counter with the specified key, and with the initial value of the
     * counter specified.
     * 
     * @param key
     *            the key to use to put in a Map.
     * @param initialValue
     *            the initial value of the counter.
     */
    public Counter( Object key, int initialValue ) {
        this.key = key;
        this.value = initialValue;
    }

    /**
     * Return the current value of the counter, and then increment it.
     * 
     * @return the value of the counter before incrementing it.
     */
    public int postInc() {
        return value++;
    }

    /**
     * Return the current value of the counter, and then decrement it.
     * 
     * @return the value of the counter before decrementing it.
     */
    public int postDec() {
        return value--;
    }

    /**
     * Return the value of the counter after incrementing it.
     * 
     * @return the incremented value of the counter.
     */
    public int preInc() {
        return ++value;
    }

    /**
     * Return the value of the counter after incrementing it.
     * 
     * @return the decremented value of the counter.
     */
    public int preDec() {
        return --value;
    }

    /** Clear the counter to zero. */
    public void clear() {
        value = 0;
    }

    /**
     * Gets the key that is used to put this Object in a Map.
     * 
     * @return the Object used as a key when putting this Object in a Map.
     * 
     * @uml.property name="key"
     */
    public final Object getKey() {
        return key;
    }

    /**
     * Gets the current value of the counter as an Integer. This is a method for implementing the
     * Map.Entry interface.
     * 
     * @return an Integer Object representing the current value of the counter.
     */
    public Object getValue() {
        return new Integer( value );
    }

    /**
     * Sets the current value of the counter as Number. This is a method for implementing the
     * Map.Entry interface.
     * 
     * @param value
     *            the Number Object to use to set the value of the counter.
     * @return an Integer Object representing the current value of the counter.
     */
    public Object setValue( Object value ) {
        this.value = ((Number)value).intValue();
        return getValue();
    }

    /**
     * Returns the current value of the counter as an int.
     * 
     * @return the current value of the counter as an int.
     */
    public int getValueAsInt() {
        return value;
    }

    /**
     * Returns the hash code value for this map entry. The hash code of a map entry e is defined to
     * be: <BR>
     * <CODE>(e.getKey()==null ? 0 : e.getKey().hashCode()) ^ (e.getValue()==null ? 0 :
     * e.getValue().hashCode())</CODE><BR>
     * This ensures that e1.equals(e2) implies that e1.hashCode()==e2.hashCode() for any two Entries
     * e1 and e2, as required by the general contract of Object.hashCode.
     * 
     * @return Returns the hash code value for this map entry. The hash code of a map entry e is
     *         defined to be: <BR>
     *         <CODE>(e.getKey()==null ? 0 : e.getKey().hashCode()) ^ (e.getValue()==null ? 0 :
     *         e.getValue().hashCode())</CODE><BR>
     *         This ensures that e1.equals(e2) implies that e1.hashCode()==e2.hashCode() for any two
     *         Entries e1 and e2, as required by the general contract of Object.hashCode.
     * 
     * 
     * @see java.lang.Object#hashCode()
     */
    public int hashCode() {
        return (getKey() == null ? 0 : getKey().hashCode()) ^ (getValue() == null ? 0 : getValue().hashCode());
    }

    /**
     * Compares the specified object with this entry for equality. Returns true if the given object
     * is also a map entry and the two entries represent the same mapping. More formally, two
     * entries e1 and e2 represent the same mapping if <BR>
     * <CODE>(e1.getKey()==null ? e2.getKey()==null : e1.getKey().equals(e2.getKey())) &&
     * (e1.getValue()==null ? e2.getValue()==null : e1.getValue().equals(e2.getValue()))</CODE>
     * <BR>
     * This ensures that the equals method works properly across different implementations of the
     * Map.Entry interface.
     * 
     * @param o
     *            object to be compared for equality with this map entry.
     * @return object to be compared for equality with this map entry, otherwise false.
     */
    public boolean equals( Object o ) {
        return this.getClass().equals( o.getClass() ) && (this.value == ((Counter)o).value);
    }

    /**
     * Comparaable Interface Implementation
     * 
     * @param o
     *            The Object that we will be comparing.
     * @return see java.lang.Comparable
     * @throws ClassCastException
     *             if the Object being compared is not of the same type.
     * @see java.lang.Comparable
     */
    public int compareTo( Object o ) {
        if ( null == o ) {
            throw new NullPointerException( "Null passed to Counter..compareTo()" );
        } else if ( this.getClass().equals( o.getClass() ) ) {
            return this.value - ((Counter)o).value;
        } else {
            throw new ClassCastException( "Object of type" + o.getClass().getName() + " passed to Counter..compareTo()" );
        }
    }
}