/*
 * HistogramBin.java
 *
 * Created on August 18, 2003, 5:36 PM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.stats;

/**
 * This class represents a single bin in a histogram. It contains the complete definition of the
 * criteria for being a member of this bin. By calling this classes isMember() method, it can be
 * immediately determined whether a particular value is a member of this bin.
 * 
 * @see suramadu.util.stats.Range
 * @author mbi
 */
public class HistogramBin implements Comparable, Range {
    /**
     * The low bound of the range represented by this HistogramBin. This is the value that is
     * <B>less than or equal to </B> all values that are enclosed by this range.
     */
    private Comparable loBound;

    /**
     * The high bound of the range represented by this HistogramBin. This is the value that is
     * <B>grater than </B> all values that are enclosed by this range.
     */
    private Comparable hiBound;

    /**
     * A counter to tally the number of values that fall within this HistogramBin's range.
     */
    private long       count;

    /**
     * Creates a new instance of HistogramBin with the specified low and high bounds. This
     * constructor enforces that the loBond < hiBound.
     * 
     * @param loBound
     *            The low bound of the range represented by this HistogramBin. This is the value
     *            that is <B>less than or equal to </B> all values that are enclosed by this range.
     * @param hiBound
     *            The high bound of the range represented by this HistogramBin. This is the value
     *            that is <B>grater than </B> all values that are enclosed by this range.
     * @see suramadu.util.stats.Range
     */
    public HistogramBin( Comparable loBound, Comparable hiBound ) {
        if ( loBound.compareTo( hiBound ) > 0 ) {
            throw new RuntimeException( "Error creating HistogramBin: loBound > hiBound" );
        }
        this.loBound = loBound;
        this.hiBound = hiBound;
        this.count = 0;
    }

    /**
     * Adds the specified Number object to this range if it truly falls within it. If added, the
     * HistogramBin's counter is incremented.
     * 
     * @param n
     *            the Number object to add
     * @return true if added and the count incremented, false if the Number object is not in the
     *         HistogramBin's range.
     */
    public boolean add( Number n ) {
        if ( contains( n ) ) {
            count++;
            return true;
        } else {
            return false;
        }
    }

    /**
     * Gets the low bound of the Range described by this HistorgramBin
     * 
     * @return he low bound of the range represented by this HistogramBin. This is the value that is
     *         <B>less than or equal to </B> all values that are enclosed by this range.
     * 
     * @uml.property name="loBound"
     */
    public Comparable getLoBound() {
        return loBound;
    }

    /**
     * Gets the high bound of the Range described by this HistogramBin.
     * 
     * @return The high bound of the range represented by this HistogramBin. This is the value that
     *         is <B>grater than </B> all values that are enclosed by this range.
     * 
     * @uml.property name="hiBound"
     */
    public Comparable getHiBound() {
        return hiBound;
    }


    /**
     * A helper method that will merge this HistogramBin with another HistogramBin iff the ranges
     * are adjescent.
     * 
     * @param range
     *            the Range object to merge with this range.
     * @return a new Object that implements the Range interface which encompasses both the current
     *         Range, and the specified Range. If the two ranges are not adjascent, then
     *         &lt;null&gt; is returned.
     * @see suramadu.util.stats.Range
     */
    public Range merge( Range range ) {
        HistogramBin retVal = null;

        if ( hiBound.compareTo( range.getLoBound() ) == 0 ) { // consecutive, with this first
            retVal = new HistogramBin( loBound, range.getHiBound() );
        } else if ( range.getHiBound().compareTo( loBound ) == 0 ) { // consecutive with range first
            retVal = new HistogramBin( range.getLoBound(), hiBound );
        }

        if ( (null != retVal) && (range instanceof HistogramBin) ) {
            retVal.count = count + ((HistogramBin)range).count;
        }
        return retVal;
    }

    /**
     * Gets the current count for this HistogramBin.
     * 
     * @return The courrent count for this HistogramBin.
     * 
     * @uml.property name="count"
     */
    public long getCount() {
        return count;
    }

    /**
     * Determines if a given number falls within the Range described by this HistogramBin. Falling
     * within range is defined as being greater than or equal to the low bound of the range, and
     * less than the high bound of the range. The only exception is in the special case where the
     * low bound is equal to the high bound, and the test number is equal to both of them.
     * 
     * @param n
     *            The Number object to test for inclusion in this HistogramBin's Range.
     * @return true if it is within this HistogramBin's range, false if not.
     * @see suramadu.util.stats.Range
     */
    public boolean contains( Number n ) {
        boolean contained = (loBound.compareTo( new Double( n.doubleValue() ) ) <= 0) && (hiBound.compareTo( new Double( n.doubleValue() ) ) > 0);
        if ( !contained ) {
            contained = (loBound.compareTo( hiBound ) == 0);
        }
        return contained;
    }

    /**
     * Determines if the specified HistogramBin overlaps this HistogramBin.
     * 
     * @param bin
     *            The HistogramBin to test.
     * @return true if the specified HistogramBin overlaps, false if not.
     * @see suramadu.util.stats.HistogramBin
     */
    public boolean overlaps( HistogramBin bin ) {
        if ( loBound.compareTo( bin.loBound ) < 0 ) {
            return (hiBound.compareTo( bin.loBound ) > 0);
        } else {
            return (loBound.compareTo( bin.hiBound ) < 0);
        }
    }

    /**
     * Comparaable Interface Implementation
     * 
     * @param o
     *            The Object that we will be comparing.
     * @return see java.lang.Comparable
     * @see java.lang.Comparable
     */
    public int compareTo( Object o ) {
        if ( null == o ) {
            throw new NullPointerException( "Null passed to HistogramBin.compareTo()" );
        } else if ( this.getClass().equals( o.getClass() ) ) {
            return loBound.compareTo( ((HistogramBin)o).getLoBound() );
        } else {
            throw new ClassCastException( "Object of type" + o.getClass().getName() + " passed to Counter..compareTo()" );
        }
    }

    /**
     * Creates a String representation fo this HistogramBin indicating its low bound, high bound,
     * and current counter.
     * 
     * @return a String representation fo this HistogramBin indicating its low bound, high bound,
     *         and current counter.
     */
    public String toString() {
        return loBound + " <= X < " + hiBound + ": " + count;
    }
}