/*
 * HistogramBinSpec.java
 *
 * Created on August 18, 2003, 5:43 PM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.stats;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

/**
 * This class is a container for HistogramBin classes. Each HistogramBin defines the specification
 * for a single bin in a Histogram. It is the duty of the HistogramBinSpec to enforce consistency
 * across all HistogramBins added to it, making sure that there are not overlaps, etc.
 * 
 * @author mbi
 */
public class HistogramBinSpec {

    /**
     * An ordered list of adjascent HistogramBins comprising this HistogramBinSpec
     * 
     * @uml.property name="bins"
     * @uml.associationEnd 
     * @uml.property name="bins" multiplicity="(0 -1)" elementType="suramadu.util.stats.HistogramBin"
     */
    private List bins;

    /**
     * An ordered list HistogramBins (created as needed), sorted in descending frequency order.
     * I.e., the HistogramBin with the largest count is first, the one with the smallest count is
     * last.
     * 
     * @uml.property name="freqSortedList"
     * @uml.associationEnd 
     * @uml.property name="freqSortedList" multiplicity="(0 -1)" elementType="suramadu.util.stats.HistogramBin"
     */
    private List freqSortedList;

    /**
     * Creates a new instance of HistogramBinSpec. This initializes bins to a new ArrayList, and
     * nulls out freqSortedList.
     */
    public HistogramBinSpec() {
        bins = new ArrayList();
        freqSortedList = null;
    }

    /**
     * Creates a new HistogramBinSpec using the specified array of HistogramBins. The consistency of
     * the provided bins is checked as the HistogramBinSpec is being created.
     * 
     * @param bins
     *            An array of HistogramBins from which to construct this HistogramBinSpec.
     * @see suramadu.util.stats.HistogramBin
     */
    public HistogramBinSpec( HistogramBin[] bins ) {
        this();
        for ( int i = 0; i < bins.length; i++ ) {
            addBin( bins[ i ] );
        }
        setHiLoBounds();
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( int loBound, int hiBound, int numBins ) {
        this( new Integer( loBound ), new Integer( hiBound ), new Integer( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( short loBound, short hiBound, short numBins ) {
        this( new Short( loBound ), new Short( hiBound ), new Short( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( byte loBound, byte hiBound, byte numBins ) {
        this( new Byte( loBound ), new Byte( hiBound ), new Byte( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( long loBound, long hiBound, long numBins ) {
        this( new Long( loBound ), new Long( hiBound ), new Long( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( float loBound, float hiBound, float numBins ) {
        this( new Float( loBound ), new Float( hiBound ), new Float( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     */
    public HistogramBinSpec( double loBound, double hiBound, double numBins ) {
        this( new Double( loBound ), new Double( hiBound ), new Double( numBins ) );
    }

    /**
     * Creates a new HistogramBinSpec using the specified loBound, hiBound, and numBins. The units
     * are arbitrary.
     * 
     * @param loBound
     *            The lowest bound of the HistogramBinSpec
     * @param hiBound
     *            The highest bound of the HistogramBinSpec
     * @param numBins
     *            The number of bins to create in this HistogramBinSpec (minus 2 for the bottom bin
     *            and the top bin).
     * @see suramadu.util.stats.HistogramBin
     */
    public HistogramBinSpec( Number loBound, Number hiBound, Number numBins ) {
        this();
        double loBoundVal = loBound.doubleValue();
        double hiBoundVal = hiBound.doubleValue();
        double numBinsVal = numBins.doubleValue();
        double binSizeVal = (hiBoundVal - loBoundVal) / numBinsVal;

        do {
            addBin( new HistogramBin( new Double( loBoundVal ), new Double( loBoundVal + binSizeVal ) ) );
            loBoundVal += binSizeVal;
        } while ( loBoundVal < hiBoundVal );
        setHiLoBounds();
    }

    /**
     * Creates a new HistogramBinSpec using an array of values organized as pairs of bounds, the
     * even numbered elements being low bounds, and the odd numbered elements being high bounds
     * (zero based). Bin consistency is checked at construction time.
     * 
     * @param limits
     *            an array of bounds, alternating between low bound and high bound for consecutive
     *            and adjascent HistogramBins.
     */
    public HistogramBinSpec( Comparable[] limits ) {
        this();
        int length = limits.length;
        if ( ((length / 2) * 2) == length ) { // check for an even number of parameters in array
            for ( int i = 0; i < length; i += 2 ) {
                if ( !addBin( limits[ i ], limits[ i + 1 ] ) ) {
                    throw new RuntimeException( "Error setting limits in HistogramBinSpec: " + limits[ i ] + ":" + limits[ i + 1 ] );
                }
            }
            setHiLoBounds();
        } else {
            throw new RuntimeException( "The limits array specified in the HistogramBinSpec constructor had an odd number of elements." );
        }
    }

    /**
     * This method adds a single bin to the beginning and end of the histogram that represents all
     * values not specifically contained in the created HistogramBinSpec. The bottom bin ranges from
     * -Infinity to the low bound of the first bin, and the top bin ranges from the high bound of
     * the last bin to +Infinity.
     * 
     * @see suramadu.util.stats.HistogramBin
     */
    private void setHiLoBounds() {
        if ( 0 == bins.size() ) {
            throw new RuntimeException( "HistogramBinSpec.setHiLoBounds: HistogramBinSpec is empty." );
        }
        Comparable minBound = new Double( Double.NEGATIVE_INFINITY );
        Comparable maxBound = new Double( Double.POSITIVE_INFINITY );
        bins.add( 0, new HistogramBin( minBound, ((HistogramBin)bins.get( 0 )).getLoBound() ) );
        addBin( new HistogramBin( ((HistogramBin)bins.get( bins.size() - 1 )).getHiBound(), maxBound ) );
    }

    /**
     * Adds a bin to the HistogramBinspec. Bins are checked for overlapping, but are not checked for
     * continuously adjascent.
     * 
     * @param loBound
     *            the lower boundary of the HistogramBin to be created.
     * @param hiBound
     *            the upper boundary of the HistogramBin to be created.
     * @return true if successful, false if the new HistogramBin overlaps any existing HistogramBin
     *         in the HistogramBinSpec.
     * @see suramadu.util.stats.HistogramBin
     * @see suramadu.util.stats.Range
     */
    public boolean addBin( Comparable loBound, Comparable hiBound ) {
        return addBin( new HistogramBin( loBound, hiBound ) );
    }

    /**
     * Adds a bin to the HistogramBinspec. Bins are checked for overlapping, but are not checked for
     * continuously adjascent.
     * 
     * @param bin
     *            the HistogramBin to add to this HistogramBinSpec
     * @return true if successful, false if the new HistogramBin overlaps any existing HistogramBin
     *         in the HistogramBinSpec.
     * @see suramadu.util.stats.HistogramBin
     * @see suramadu.util.stats.Range
     */
    public boolean addBin( HistogramBin bin ) {
        if ( !overlaps( bin ) ) {
            bins.add( bin );
            return true;
        } else {
            return false;
        }
    }

    /**
     * Find the HistogramBin in this HistogramBinSpec in which the specified Number object resides.
     * 
     * @param n
     *            A Number object for which to find the HistogramBin whose range encompasses it.
     * @return the HistogramBin object that describes the range that would include the value of the
     *         specified Number
     * @see suramadu.util.stats.HistogramBin
     */
    public HistogramBin getBin( Number n ) {
        for ( Iterator i = bins.iterator(); i.hasNext(); ) {
            HistogramBin testBin = (HistogramBin)i.next();
            if ( testBin.contains( n ) ) {
                return testBin;
            }
        }
        return null;
    }

    /**
     * Get the HistogramBin whose count has the greatest value.
     * 
     * @return the HistogramBin that contains the largest count in the current HistogramBinSpec.
     * @see suramadu.util.stats.HistogramBin
     */
    public HistogramBin getModeBin() {
        HistogramBin retVal = null;

        createFrequencySortedList();
        for ( Iterator i = freqSortedList.iterator(); i.hasNext(); ) {
            HistogramBin bin = (HistogramBin)i.next();
            if ( null == retVal ) {
                retVal = bin;
            } else {
                if ( retVal.getCount() == bin.getCount() ) {
                    retVal = (HistogramBin)retVal.merge( bin );
                } else {
                    break;
                }
            }
        }
        return retVal;
    }

    /**
     * Sorts the bins list in descending frequency order
     * 
     * @return a new ArrayList of HistogramBin objects sorted in descending frequency order.
     * @see suramadu.util.stats.HistogramBin
     */
    protected List createFrequencySortedList() {
        if ( null == freqSortedList ) {
            freqSortedList = new ArrayList( bins );
            Collections.sort( freqSortedList, new Comparator() {
                public int compare( Object o1, Object o2 ) {
                    if ( (o1 instanceof HistogramBin) && (o2 instanceof HistogramBin) ) {
                        return (int)(((HistogramBin)o2).getCount() - ((HistogramBin)o1).getCount());
                    } else {
                        throw new ClassCastException( "Illegal class passed to Comparator in HistogramBinSpec.createFrequencySortedSet()" );
                    }
                }
            } );
        }
        return freqSortedList;
    }

    /**
     * Determines if the specified HistogramBin overlaps any HistogramBin already existing in this
     * HistogramBinSpec.
     * 
     * @param bin
     *            The HistogramBin to test.
     * @return true if the specified HistogramBin overlaps, false if not.
     * @see suramadu.util.stats.HistogramBin
     */
    protected boolean overlaps( HistogramBin bin ) {
        for ( Iterator i = bins.iterator(); i.hasNext(); ) {
            HistogramBin testBin = (HistogramBin)i.next();
            if ( testBin.overlaps( bin ) ) return true;
        }
        return false;
    }

    /**
     * Get an iterator to the HistogramBin list in ascending range order
     * 
     * @return an iterator to the HistogramBin list in ascending range order
     * @see suramadu.util.stats.HistogramBin
     */
    public Iterator iterator() {
        return bins.iterator();
    }

    /**
     * Create a descriptive String that indicates the ranges and counts of all HistogramBins in this
     * HistogramBinSpec
     * 
     * @return a descriptive String that indicates the ranges and counts of all HistogramBins in
     *         this HistogramBinSpec
     * @see suramadu.util.stats.HistogramBin
     */
    public String toString() {
        StringBuffer s = new StringBuffer();
        int j = 0;

        for ( Iterator i = bins.iterator(); i.hasNext(); j++ ) {
            if ( j > 0 ) s.append( ' ' );
            s.append( i.next() );
        }
        return s.toString();
    }
}