/**
 * StatPack.java Created on August 8, 2003, 3:30 PM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.stats;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Static helper class to compute various statistical values of collections and arrays.
 * 
 * @author mbi
 */
public class StatPack {
    /**
     * Given an array of ints of length n, returns an array of length n-1 whos elements consist of the difference between each
     * consecutive element of the specified array.
     * 
     * @param intArray
     *            array of ints to be differenced
     * @return an array of ints one element short of the original array that consists of the differences between each consecutive
     *         element of the origianl array.
     * @see java.util.Collection
     */
    public static int[] diff( int[] intArray ) {
        return diff( intArray, 0 );
    }
    
    public static int[] diff( int[] intArray, int offset ) {
        int length = intArray.length - offset;
        int[] diff = new int[ length - 1 ];
        for ( int i = 1; i < length; i++ ) {
            diff[ i - 1 ] = intArray[ i + offset ] - intArray[ i + offset - 1 ];
        }
        return diff;
    }

    /**
     * Given an array of longs of length n, returns an array of length n-1 whos elements consist of the difference between each
     * consecutive element of the specified array.
     * 
     * @param longArray
     *            array of longs to be differenced
     * @return an array of longs one element short of the original array that consists of the differences between each consecutive
     *         element of the origianl array.
     * @see java.util.Collection
     */
    public static long[] diff( long[] longArray ) {
        return diff( longArray, 0 );
    }
    
    public static long[] diff( long[] longArray, int offset ) {
        int length = longArray.length - offset;
        long[] diff = new long[ length - 1 ];
        for ( int i = 1; i < length; i++ ) {
            diff[ i - 1 ] = longArray[ i + offset ] - longArray[ i + offset - 1 ];
        }
        return diff;
    }

    /**
     * Given an array of floats of length n, returns an array of length n-1 whos elements consist of the difference between each
     * consecutive element of the specified array.
     * 
     * @param floatArray
     *            array of floats to be differenced
     * @return an array of floats one element short of the original array that consists of the differences between each consecutive
     *         element of the origianl array.
     * @see java.util.Collection
     */
    public static float[] diff( float[] floatArray ) {
        return diff( floatArray, 0 );
    }
    
    public static float[] diff( float[] floatArray, int offset ) {
        int length = floatArray.length - offset;
        float[] diff = new float[ length - 1 ];
        for ( int i = 1; i < length; i++ ) {
            diff[ i - 1 ] = floatArray[ i + offset ] - floatArray[ i + offset - 1 ];
        }
        return diff;
    }

    /**
     * Given an array of doubles of length n, returns an array of length n-1 whos elements consist of the difference between each
     * consecutive element of the specified array.
     * 
     * @param doubleArray
     *            array of doubles to be differenced
     * @return an array of doubles one element short of the original array that consists of the differences between each consecutive
     *         element of the origianl array.
     * @see java.util.Collection
     */
    public static double[] diff( double[] doubleArray ) {
        return diff( doubleArray, 0 );
    }
    
    public static double[] diff( double[] doubleArray, int offset ) {
        int length = doubleArray.length - offset;
        double[] diff = new double[ length - 1 ];
        for ( int i = 1; i < length; i++ ) {
            diff[ i - 1 ] = doubleArray[ i + offset ] - doubleArray[ i + offset - 1 ];
        }
        return diff;
    }

    /**
     * Utility method for converting array of int to a Collection.
     * 
     * @param intArray
     *            an array of ints to be converted to a Collection
     * @return a collection of Integer objects representing the origianl array of ints.
     * @see java.util.Collection
     */
    public static Collection toCollection( int[] intArray ) {
        if ( null == intArray ) return new ArrayList();

        List l = new ArrayList( intArray.length );
        for ( int i = 0; i < intArray.length; i++ ) {
            l.add( new Integer( intArray[ i ] ) );
        }
        return l;
    }

    /**
     * Utility method for converting array of long to a Collection.
     * 
     * @param longArray
     *            an array of ints to be converted to a Collection
     * @return a collection of Long objects representing the origianl array of longs.
     * @see java.util.Collection
     */
    public static Collection toCollection( long[] longArray ) {
        if ( null == longArray ) return new ArrayList();

        List l = new ArrayList( longArray.length );
        for ( int i = 0; i < longArray.length; i++ ) {
            l.add( new Long( longArray[ i ] ) );
        }
        return l;
    }

    /**
     * Utility method for converting array of float to a Collection.
     * 
     * @param floatArray
     *            an array of floats to be converted to a Collection
     * @return a collection of Float objects representing the origianl array of floats.
     * @see java.util.Collection
     */
    public static Collection toCollection( float[] floatArray ) {
        if ( null == floatArray ) return new ArrayList();

        List l = new ArrayList( floatArray.length );
        for ( int i = 0; i < floatArray.length; i++ ) {
            l.add( new Float( floatArray[ i ] ) );
        }
        return l;
    }

    /**
     * Utility method for converting array of double to a Collection.
     * 
     * @param doubleArray
     *            an array of doubles to be converted to a Collection
     * @return a collection of Doubles objects representing the origianl array of doubles.
     * @see java.util.Collection
     */
    public static Collection toCollection( double[] doubleArray ) {
        if ( null == doubleArray ) return new ArrayList();

        List l = new ArrayList( doubleArray.length );
        for ( int i = 0; i < doubleArray.length; i++ ) {
            l.add( new Double( doubleArray[ i ] ) );
        }
        return l;
    }

    /**
     * Utility method for converting Collection Number objects to an array of ints.
     * 
     * @param c
     *            a Collection of Number objects
     * @return an array of ints representing the origianl Collection of Numbers.
     * @see java.util.Collection
     */
    public static int[] toIntArray( Collection c ) {
        int count = 0;
        int[] array = new int[ c.size() ];

        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            array[ count ] = ((Number)i.next()).intValue();
        }
        return array;
    }

    /**
     * Utility method for converting Collection Number objects to an array of longs.
     * 
     * @param c
     *            a Collection of Number objects
     * @return an array of longs representing the origianl Collection of Numbers.
     * @see java.util.Collection
     */
    public static long[] toLongArray( Collection c ) {
        int count = 0;
        long[] array = new long[ c.size() ];

        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            array[ count ] = ((Number)i.next()).longValue();
        }
        return array;
    }

    /**
     * Utility method for converting Collection Number objects to an array of floats.
     * 
     * @param c
     *            a Collection of Number objects
     * @return an array of floats representing the origianl Collection of Numbers.
     * @see java.util.Collection
     */
    public static float[] toFloatArray( Collection c ) {
        int count = 0;
        float[] array = new float[ c.size() ];

        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            array[ count ] = ((Number)i.next()).floatValue();
        }
        return array;
    }

    /**
     * Utility method for converting Collection Number objects to an array of doubles.
     * 
     * @param c
     *            a Collection of Number objects
     * @return an array of doubles representing the origianl Collection of Numbers.
     */
    public static double[] toDoubleArray( Collection c ) {
        int count = 0;
        double[] array = new double[ c.size() ];

        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            array[ count ] = ((Number)i.next()).doubleValue();
        }
        return array;
    }

    /**
     * Calculate the statistical mode of an array of ints. The mode is the range of numbers that occurs most frequently in a list.
     * Bin size defaults to 1.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return a Range representing the statisticl mode of the specified array of ints.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( int[] intArray ) {
        return mode( intArray, 0 );
    }

    /**
     * Calculate the statistical mode of an array of ints. The mode is the range of numbers that occurs most frequently in a list.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @param numBins
     *            the number of bins to create for the Histogram to calculate the mode.
     * @return a Range representing the statisticl mode of the specified array of ints.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( int[] intArray, int numBins ) {
        return mode( toCollection( intArray ), new Integer( numBins ) );
    }

    /**
     * Calculate the statistical mode of an array of longs. The mode is the range of numbers that occurs most frequently in a list.
     * Bin size defaults to 1.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return a Range representing the statisticl mode of the specified array of longs.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( long[] longArray ) {
        return mode( longArray, 0 );
    }

    /**
     * Calculate the statistical mode of an array of longs. The mode is the range of numbers that occurs most frequently in a list.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @param numBins
     *            the number of bins to create for the Histogram to calculate the mode.
     * @return a Range representing the statisticl mode of the specified array of longs.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( long[] longArray, long numBins ) {
        return mode( toCollection( longArray ), new Long( numBins ) );
    }

    /**
     * Calculate the statistical mode of an array of floats. The mode is the range of numbers that occurs most frequently in a list.
     * Bin size defaults to 1.
     * 
     * @param floatArray
     *            an array of floats to on which to calculate the statistical mode.
     * @return a Range representing the statisticl mode of the specified array of floats.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( float[] floatArray ) {
        return mode( toCollection( floatArray ), new Float( 0 ) );
    }

    /**
     * Calculate the statistical mode of an array of floats. The mode is the range of numbers that occurs most frequently in a list.
     * 
     * @param floatArray
     *            an array of floats to on which to calculate the statistical mode.
     * @param numBins
     *            the number of bins to create for the Histogram to calculate the mode.
     * @return a Range representing the statisticl mode of the specified array of floats.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( float[] floatArray, float numBins ) {
        return mode( floatArray, 1.0f );
    }

    /**
     * Calculate the statistical mode of an array of doubles. The mode is the range of numbers that occurs most frequently in a
     * list. Bin size defaults to 1.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return a Range representing the statisticl mode of the specified array of doubles.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( double[] doubleArray ) {
        return mode( doubleArray, 0.0 );
    }

    /**
     * Calculate the statistical mode of an array of doubles. The mode is the range of numbers that occurs most frequently in a
     * list.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @param numBins
     *            the number of bins to create for the Histogram to calculate the mode.
     * @return a Range representing the statisticl mode of the specified array of doubles.
     * @see suramadu.util.stats.Range
     */
    public static Range mode( double[] doubleArray, double numBins ) {
        return mode( toCollection( doubleArray ), new Double( numBins ) );
    }

    /**
     * Calculate the statistical mode of a Collection of Numbers. The mode is the range of numbers that occurs most frequently in a
     * list.
     * 
     * @param c
     *            a collection of Comparables on which to calculate the statistical mode.
     * @param numBins
     *            the number of bins to create for the Histogram to calculate the mode.
     * @return a Range representing the statisticl mode of the specified array of ints.
     * @see suramadu.util.stats.Range
     * @see java.util.Collection
     * @see java.lang.Comparable
     * @see java.lang.Number
     */
    public static Range mode( Collection c, Number numBins ) {
        double min = min( c );
        double max = max( c );
        double nbins = numBins.doubleValue();
        if ( 0.0 == nbins ) {
            nbins = 1000.0;
        }
        return new Histogram( min, max, nbins, c ).getMode();
    }

    /**
     * Calculate the statistical variance of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical variance.
     */
    public static double variance( int[] intArray ) {
        return variance( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical variance of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical variance.
     */
    public static double variance( long[] longArray ) {
        return variance( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical variance of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical variance.
     */
    public static double variance( float[] floatArray ) {
        return variance( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical variance of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical variance.
     */
    public static double variance( double[] doubleArray ) {
        return variance( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical variance of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical variance.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double variance( Collection c ) {
        double mean = mean( c );

        long count = 0;
        double sum = 0.0;
        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            double deviation = ((Number)i.next()).doubleValue() - mean;
            sum += (deviation * deviation);
        }
        if ( count > 1 ) {
            return sum / (count - 1);
        }
        else {
            return Double.NaN;
        }
    }

    /**
     * Calculate the statistical standard deviation of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical standard deviation.
     */
    public static double stddev( int[] intArray ) {
        return stddev( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical standard deviation of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical standard deviation.
     */
    public static double stddev( long[] longArray ) {
        return stddev( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical standard deviation of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical standard deviation.
     */
    public static double stddev( float[] floatArray ) {
        return stddev( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical standard deviation of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical standard deviation.
     */
    public static double stddev( double[] doubleArray ) {
        return stddev( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical standard deviation of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical standard deviation.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double stddev( Collection c ) {
        return Math.sqrt( variance( c ) );
    }

    /**
     * Calculate the statistical mean of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical mean.
     */
    public static double mean( int[] intArray ) {
        return mean( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical mean of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical mean.
     */
    public static double mean( long[] longArray ) {
        return mean( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical mean of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical mean.
     */
    public static double mean( float[] floatArray ) {
        return mean( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical mean of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical mean.
     */
    public static double mean( double[] doubleArray ) {
        return mean( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical mean of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical mean.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double mean( Collection c ) {
        long count = 0;
        double sum = 0.0;
        for ( Iterator i = c.iterator(); i.hasNext(); count++ ) {
            sum += ((Number)i.next()).doubleValue();
        }
        if ( count > 0 ) {
            return sum / count;
        }
        else {
            return Double.NaN;
        }
    }

    /**
     * Calculate the statistical median of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical median.
     */
    public static double median( int[] intArray ) {
        return median( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical median of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical median.
     */
    public static double median( long[] longArray ) {
        return median( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical median of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical median.
     */
    public static double median( float[] floatArray ) {
        return median( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical median of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical median.
     */
    public static double median( double[] doubleArray ) {
        return median( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical median of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical median.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double median( Collection c ) {
        List l = new ArrayList( c );
        Collections.sort( l );
        return (l.size() > 1) ? ((Number)l.get( (l.size() / 2) + 1 )).doubleValue() : ((Number)l.get( 0 )).doubleValue();
    }

    /**
     * Calculate the statistical minimum of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical minimum.
     */
    public static double min( int[] intArray ) {
        return min( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical minimum of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical minimum.
     */
    public static double min( long[] longArray ) {
        return min( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical minimum of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical minimum.
     */
    public static double min( float[] floatArray ) {
        return min( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical minimum of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical minimum.
     */
    public static double min( double[] doubleArray ) {
        return max( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical minimum of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical minimum.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double min( Collection c ) {
        return ((Number)Collections.min( c )).doubleValue();
    }

    /**
     * Calculate the statistical maximum of an array of ints.
     * 
     * @param intArray
     *            an array of ints to on which to calculate the statistical mode.
     * @return the value of the statistical maximum.
     */
    public static double max( int[] intArray ) {
        return max( toCollection( intArray ) );
    }

    /**
     * Calculate the statistical maximum of an array of longs.
     * 
     * @param longArray
     *            an array of longs to on which to calculate the statistical mode.
     * @return the value of the statistical maximum.
     */
    public static double max( long[] longArray ) {
        return max( toCollection( longArray ) );
    }

    /**
     * Calculate the statistical maximum of an array of floats.
     * 
     * @param floatArray
     *            an array of float to on which to calculate the statistical mode.
     * @return the value of the statistical maximum.
     */
    public static double max( float[] floatArray ) {
        return max( toCollection( floatArray ) );
    }

    /**
     * Calculate the statistical maximum of an array of doubles.
     * 
     * @param doubleArray
     *            an array of doubles to on which to calculate the statistical mode.
     * @return the value of the statistical maximum.
     */
    public static double max( double[] doubleArray ) {
        return max( toCollection( doubleArray ) );
    }

    /**
     * Calculate the statistical maximum of Collection of Numbers.
     * 
     * @param c
     *            a Collection of Number objects over which to calculate the statistical varience
     * @return the value of the statistical maximum.
     * @see java.util.Collection
     * @see java.lang.Number
     */
    public static double max( Collection c ) {
        return ((Number)Collections.max( c )).doubleValue();
    }
}