/**
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 *---------------------------------------------------------------------------------
 * Determines clockspeed in GHz using clock_gettime, which is supposed
 * to have nanosecond resolution. Exits when a constant value has been
 * reached for a specified number of iterations. Includes standard 
 * deviation calculations and argument passing through the command line.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "estimate_GHz.h"

int main(int argc, char** argv) {
    int          pid, priority, i;
    FILE         *stream;
    unsigned     _bound = 3, _loops = 1, _tolerance = 1, _tolerance_retries = 5;
    const char*  _output = "output.txt";

    unsigned long long tot_avg = 0;

    // parse commands
    for ( i = 1; i < argc; i++ ) {
        if ( strcmp(argv[i], "-b") == 0 ) {
            _bound = atoi( argv[ ++i ] );
        }
        else if ( strcmp(argv[i], "-l") == 0 ) {
            _loops = atoi( argv[ ++i ] );
        }
        else if ( strcmp(argv[i], "-t") == 0 ) {
            _tolerance = atoi( argv[ ++i ] );
        }
        else if ( strcmp(argv[i], "-r") == 0 ) {
            _tolerance_retries = atoi( argv[ ++i ] );
        }
        else if ( strcmp( argv[ i ], "-o") == 0 ) {
            _output = argv[++i];
        }
        else {
            printf( "ClockspeedGHz: Unrecognized option: '%s'\n", argv[i] );
            printf( "ClockspeedGHz: Default values: bound=%u; loops=%u; tolerance=%u; tolerance_retries=%u; output=%s\n", _bound, _loops, _tolerance, _tolerance_retries, _output );
            printf( "ClockspeedGHz: Usage: ClockspeedGHz [-b bound] [-l loops] [-t tolerance] [-r tolerance_retries] [-o output]\n" );
            return( 0 );
        }
    }
   
    printf( "\n\n\n\n>> CPU Frequency Analyzer\n" );
    printf( ">> Using clock_gettime() - NANOSECOND resolution\n" );

#ifndef PRIO_MAX // this lets me compile this on Solaris.
#define PRIO_MAX 2
#endif // PRIO_MAX

#ifndef PRIO_PROCESS // this lets me compile on a Mac
#define PRIO_PROCESS 0
#endif // PRIO_PROCESS

#ifndef DARWIN
    setpriority( PRIO_PROCESS, 0 ,PRIO_MAX ) ; // Ensure highest priority
#endif // DARWIN

    pid = (int)getpid( );

#ifndef DARWIN
    priority = getpriority( PRIO_PROCESS, 0 );
#else
    priority = -1;
#endif // DARWIN

    printf( ">> PID=%d; PRIORITY=%d; BOUND=%d; LOOPS=%d\n>> TOLERANCE=%u; TOLERANCE RETRIES=%u\n>> Begin...\n\n", pid, priority, _bound, _loops, _tolerance, _tolerance_retries);

    if ( (stream = fopen(_output, "w")) == NULL ) {
        printf( "File error\n" );
        return 0;
    }
    fprintf( stream, "Clock Tests\n\n");
    fprintf( stream, "Bound = %d; Loop = %d; Tolerance = %u; Tolerance Retries = %u\n\n", _bound, _loops, _tolerance, _tolerance_retries );
    fclose(  stream );

    tot_avg = estimate_GHz(_bound, _loops, _tolerance, _tolerance_retries, _output);

    if ( (stream = fopen(_output, "a")) == NULL ) {
        printf("File error\n");
    }    
    fprintf( stream, "Clock = %llu\n", tot_avg / _loops );
    fclose(  stream );
    
    return 0;
}

