/**
 * Time Utility - Used to convert and get times/cycles.
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.util.time;

import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Properties;

/**
 * This class is a utility class for measuring and calculating very high-resolution time based on th enumber of machine cycles
 * between events being measured. It contains two native methods, getCycles(), and getTime() for retrieving low level time
 * measurements from the underlying operating system. (See method descriptions for details of their functionality).
 */
public class Time {
    /** String for indicating unit type: minutes */
    public static final String TIME_MIN            = "min";
    /** String for indicating unit type: seconds */
    public static final String TIME_S              = "s";
    /** String for indicating unit type: milliseconds */
    public static final String TIME_MS             = "ms";
    /** String for indicating unit type: microseconds */
    public static final String TIME_US             = "us";     // µi
    /** String for indicating unit type: nanoseconds */
    public static final String TIME_NS             = "ns";
    /** Native library name */
    public static final String NATIVE_LIBRARY_NAME = "nanoRes";

    /**
     * This variable contains the clock speed of the microprocessor on which this class is running measured in cycles per second
     * (cps). It is retrieved at static initialization of the class as a property from a file created by the ClockspeedGHz utility.
     * If this file is not found, an attempt is made to run the ClockspeedGHz utility to generate it before proceeding. This can be
     * very time consuming.
     */
    public static final long   cps;

    // lib created by RTPresto
    static {
        long tempCPS = 0L;
        Properties p = new Properties();
        InputStream is = null;
        String timeProperties = "SuramaduTime.properties";
        boolean libraryLoaded = false;

        // load the native library...
        try {
            System.loadLibrary( NATIVE_LIBRARY_NAME );
            libraryLoaded = true;
        }
        catch ( UnsatisfiedLinkError e ) {
            System.out.println( "SYSTEM ERROR: Cannot load native library \"" + NATIVE_LIBRARY_NAME
                    + "\". All timing operations will fail!" );
            System.out.println( "              Library Path = " + System.getProperty( "java.library.path" ) );
            System.out.println( "              Exception = " + e );
        }

        try {
            is = new FileInputStream( timeProperties );
            if ( null == is ) {
                throw new NullPointerException();
            }
            else {
                p.load( is );
                tempCPS = Long.parseLong( p.getProperty( "cps" ) );
                is.close();
            }
        }
        catch ( Exception e ) {
        }

        // Check to see if cps has been set through properties file read.
        if ( 0L == tempCPS ) {
            // if not, check to see if native library is loaded before attempting to run
            // calculations.
            if ( libraryLoaded ) {
                System.out.println( "Running ClockspeedGHz utility to calibrate time calculations..." );
                tempCPS = getCPS( 3, 1, 1, 5, timeProperties );
                System.out.println( "Done." );

                if ( tempCPS > 1e9 ) {
                    System.out.println( "Timebase Initialized: Clock is " + ((double)tempCPS) / 1e9 + " Ghz." );
                }
                else {
                    System.out.println( "Timebase Initialized: Clock is " + ((double)tempCPS) / 1e6 + " Mhz." );
                }
            }
            else {
                System.out.println( "SYSTEM ERROR: Could not determine CPU clock speed!" );
            }
        }
        cps = tempCPS;
    }

    /**
     * Get the current high resolution cycle count using inline assembly similar to 'rdtsc'.
     * 
     * @return The current microprocessor cycle count
     */
    public static native long getCycles();

    /**
     * Calculate the clock frequency of the processor on which it is executing.
     * 
     * @return The number of cycles per second (CPS) at which the processor runs.
     */
    public static native long getCPS( int bounds, int loops, int tolerace, int toleranceRetries, String output );

    /**
     * Get the current high resolution time from the clock. The first array element is the seconds value, and the second array
     * element is the nanoseconds value. (This data structure is consistent with Linux's formattig of tv_struct).
     * 
     * @param timeArray
     *            Writes the first two elements to the specified array of type long, the first element indicates the number of
     *            milliseconds, the second indicates the number of nanoseconds. This time format is analogous to the format used in
     *            Linux to define time, as specified in the OS data structure, tv_struct.
     */
    public static native void getTime( long[] timeArray );

    /**
     * This method throws a RuntimeException if the cps value has never been initialized, meaning that the class has no knowledge of
     * the current CPU speed. It is still possible and meaningful to call the versions of the methods herein that take as their
     * arguments the cps value to use for real-time calculation, but any such method that relies on the internally calculated value
     * will fail with this exception.
     */
    static private void checkCPS() {
        if ( 0L == cps )
                throw new RuntimeException(
                        "System clock's speed has not been calculated. Cannot call methods that depend on its value." );
    }

    /**
     * Given an array of ints of length n where each element represents an absolute time measurement, calculate the average period
     * over the length of the array.
     * 
     * @param intArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in intArray
     * @return average period or -1 if parameters are inappropriate
     * @see java.util.Collection
     */
    public static long avgPeriod( int[] intArray, int offset ) {
        int length = intArray.length;
        if ( (length <= 2) || (offset >= length - 1) ) return -1;

        return ((long)intArray[ length - 1 ] - intArray[ offset ]) / ((length - offset) - 1);
    }

    /**
     * Given an array of ints of length n where each element represents an absolute time measurement, calculate the average period
     * over the length of the array.
     * 
     * @param longArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in longArray
     * @return average period or -1 if parameters are inappropriate
     * @see java.util.Collection
     */
    public static long avgPeriod( long[] longArray, int offset ) {
        int length = longArray.length;
        if ( (length <= 2) || (offset >= length - 1) ) return -1;

        return (longArray[ length - 1 ] - longArray[ offset ]) / ((length - offset) - 1);
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, and a period that represents
     * a periodic time in the same units as the time measurements in the array, convert the values in the array to represent
     * deviation from a calculated average period.
     * 
     * @param intArray
     *            array of times to be converted
     * @return converted array.
     * @see java.util.Collection
     */
    public static int[] jitter( int[] intArray ) {
        return jitter( intArray, 0 );
    }

    /**
     * Given an array of ints of length n where each element represents an absolute time measurement, and a period that represents a
     * periodic time in the same units as the time measurements in the array, convert the values in the array to represent deviation
     * from the specified period.
     * 
     * @param period
     *            period from which to perform jitter calculations
     * @param intArray
     *            array of times to be converted
     * @return converted array.
     * @see java.util.Collection
     */
    public static int[] jitter( long period, int[] intArray ) {
        return jitter( period, intArray, 0 );
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, and a period that represents
     * a periodic time in the same units as the time measurements in the array, convert the values in the array to represent
     * deviation from a calculated average period.
     * 
     * @param longArray
     *            array of times to be converted
     * @return converted array.
     * @see java.util.Collection
     */
    public static long[] jitter( long[] longArray ) {
        return jitter( longArray, 0 );
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, and a period that represents
     * a periodic time in the same units as the time measurements in the array, convert the values in the array to represent
     * deviation from the specified period.
     * 
     * @param period
     *            period from which to perform jitter calculations
     * @param longArray
     *            array of times to be converted
     * @return converted array.
     * @see java.util.Collection
     */
    public static long[] jitter( long period, long[] longArray ) {
        return jitter( period, longArray, 0 );
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, convert the values in the
     * array to represent deviation from a calculated average period.
     * 
     * @param intArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in longArray
     * @return converted array.
     * @see java.util.Collection
     */
    public static int[] jitter( int[] intArray, int offset ) {
        return jitter( avgPeriod( intArray, offset ), intArray, offset );
    }

    /**
     * Given an array of ints of length n where each element represents an absolute time measurement, and a period that represents a
     * periodic time in the same units as the time measurements in the array, convert the values in the array to represent deviation
     * from the specified period.
     * 
     * @param period
     *            period from which to perform jitter calculations
     * @param intArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in intArray
     * @return converted array.
     * @see java.util.Collection
     */
    public static int[] jitter( long period, int[] intArray, int offset ) {
        int startTime = intArray[ offset ];
        int jitterArray[] = new int[ intArray.length - offset ];
        for ( int i = offset; i < intArray.length; i++ ) {
            jitterArray[ i - offset ] = (int)(intArray[ i ] - (startTime + ((i - offset) * period)));
        }
        return jitterArray;
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, convert the values in the
     * array to represent deviation from a calculated average period.
     * 
     * @param longArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in longArray
     * @return converted array.
     * @see java.util.Collection
     */
    public static long[] jitter( long[] longArray, int offset ) {
        return jitter( avgPeriod( longArray, offset ), longArray, offset );
    }

    /**
     * Given an array of longs of length n where each element represents an absolute time measurement, and a period that represents
     * a periodic time in the same units as the time measurements in the array, convert the values in the array to represent
     * deviation from the specified period.
     * 
     * @param period
     *            period from which to perform jitter calculations
     * @param longArray
     *            array of times to be converted
     * @param offset
     *            begin calculations from offset position in longArray
     * @return converted array.
     * @see java.util.Collection
     */
    public static long[] jitter( long period, long[] longArray, int offset ) {
        long startTime = longArray[ offset ];
        long jitterArray[] = new long[ longArray.length - offset ];
        for ( int i = offset; i < longArray.length; i++ ) {
            jitterArray[ i - offset ] = longArray[ i ] - (startTime + ((i - offset) * period));
        }
        return jitterArray;
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of doubles representing the converted values.
     */
    public static double[] minFromCycles( long[] cycles ) {
        return minFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of doubles representing the converted values.
     */
    public static double[] minFromCycles( long[] cycles, long frequency ) {
        checkCPS();

        double[] mins = new double[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = minFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of longs representing the converted values.
     */
    public static long[] minFromCyclesL( long[] cycles ) {
        return minFromCyclesL( cycles, cps );
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of longs representing the converted values.
     */
    public static long[] minFromCyclesL( long[] cycles, long frequency ) {
        checkCPS();

        long[] mins = new long[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = (long)minFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return the numbe of minutes represented by the specified number of machine cycles.
     */
    public static double minFromCycles( long cycles ) {
        checkCPS();
        return minFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of minutes represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of minutes represented by the specified number of machine cycles.
     */
    public static double minFromCycles( long cycles, long frequency ) {
        return (double)cycles / (double)frequency / (double)60;
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of doubles representing the converted values.
     */
    public static double[] sFromCycles( long[] cycles ) {
        return sFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of doubles representing the converted values.
     */
    public static double[] sFromCycles( long[] cycles, long frequency ) {
        checkCPS();

        double[] mins = new double[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = sFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of longs representing the converted values.
     */
    public static long[] sFromCyclesL( long[] cycles ) {
        return sFromCyclesL( cycles, cps );
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of longs representing the converted values.
     */
    public static long[] sFromCyclesL( long[] cycles, long frequency ) {
        checkCPS();

        long[] mins = new long[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = (long)sFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return the numbe of seconds represented by the specified number of machine cycles.
     */
    public static double sFromCycles( long cycles ) {
        checkCPS();
        return sFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of seconds represented by the specified number of machine cycles based on its intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of seconds represented by the specified number of machine cycles.
     */
    public static double sFromCycles( long cycles, long frequency ) {
        return (double)cycles / (double)frequency * (double)1.0;
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of doubles representing the converted values.
     */
    public static double[] msFromCycles( long[] cycles ) {
        return msFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of doubles representing the converted values.
     */
    public static double[] msFromCycles( long[] cycles, long frequency ) {
        checkCPS();

        double[] mins = new double[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = msFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of longs representing the converted values.
     */
    public static long[] msFromCyclesL( long[] cycles ) {
        return msFromCyclesL( cycles, cps );
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of longs representing the converted values.
     */
    public static long[] msFromCyclesL( long[] cycles, long frequency ) {
        checkCPS();

        long[] mins = new long[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = (long)msFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return the numbe of milliseconds represented by the specified number of machine cycles.
     */
    public static double msFromCycles( long cycles ) {
        checkCPS();
        return msFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of milliseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of milliseconds represented by the specified number of machine cycles.
     */
    public static double msFromCycles( long cycles, long frequency ) {
        return (double)cycles / (double)frequency * (double)1000.0;
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of doubles representing the converted values.
     */
    public static double[] usFromCycles( long[] cycles ) {
        return usFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of doubles representing the converted values.
     */
    public static double[] usFromCycles( long[] cycles, long frequency ) {
        checkCPS();

        double[] mins = new double[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = usFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of longs representing the converted values.
     */
    public static long[] usFromCyclesL( long[] cycles ) {
        return usFromCyclesL( cycles, cps );
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of longs representing the converted values.
     */
    public static long[] usFromCyclesL( long[] cycles, long frequency ) {
        checkCPS();

        long[] mins = new long[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = (long)usFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return the numbe of microseconds represented by the specified number of machine cycles.
     */
    public static double usFromCycles( long cycles ) {
        checkCPS();
        return usFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of microseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of microseconds represented by the specified number of machine cycles.
     */
    public static double usFromCycles( long cycles, long frequency ) {
        return (double)cycles / (double)frequency * (double)1000000.0;
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of doubles representing the converted values.
     */
    public static double[] nsFromCycles( long[] cycles ) {
        return nsFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of doubles representing the converted values.
     */
    public static double[] nsFromCycles( long[] cycles, long frequency ) {
        checkCPS();

        double[] mins = new double[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = nsFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return an array of longs representing the converted values.
     */
    public static long[] nsFromCyclesL( long[] cycles ) {
        return nsFromCyclesL( cycles, cps );
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (frequency) for an array of values.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return an array of longs representing the converted values.
     */
    public static long[] nsFromCyclesL( long[] cycles, long frequency ) {
        checkCPS();

        long[] mins = new long[ cycles.length ];
        for ( int i = 0; i < cycles.length; i++ ) {
            mins[ i ] = (long)nsFromCycles( cycles[ i ], frequency );
        }
        return mins;
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return the numbe of nanoseconds represented by the specified number of machine cycles.
     */
    public static double nsFromCycles( long cycles ) {
        checkCPS();
        return nsFromCycles( cycles, cps );
    }

    /**
     * Calculates the number of nanoseconds represented by the specified number of machine cycles based on its intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of nanoseconds represented by the specified number of machine cycles.
     */
    public static double nsFromCycles( long cycles, long frequency ) {
        return (double)cycles / (double)frequency * (double)1000000000.0;
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of minutes based on the intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param minutes
     *            The number of minutes to use for the conversion.
     * @return the numbe of machine cycles represented by the specified number of minutes.
     */
    public static long cyclesFromMin( long minutes ) {
        checkCPS();
        return cyclesFromMin( minutes, cps );
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of minutes based on the specified CPU speed
     * timebase (frequency).
     * 
     * @param minutes
     *            The number of minutes to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of machine cycles represented by the specified number of minutes.
     */
    public static long cyclesFromMin( long minutes, long frequency ) {
        return cyclesFromS( minutes, frequency ) * 60L;
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of seconds based on the intrinsically calculated
     * CPU speed timebase (cps).
     * 
     * @param seconds
     *            The number of seconds to use for the conversion.
     * @return the numbe of machine cycles represented by the specified number of seconds.
     */
    public static long cyclesFromS( long seconds ) {
        checkCPS();
        return cyclesFromS( seconds, cps );
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of seconds based on the specified CPU speed
     * timebase (frequency).
     * 
     * @param seconds
     *            The number of seconds to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of machine cycles represented by the specified number of seconds.
     */
    public static long cyclesFromS( long seconds, long frequency ) {
        return seconds * frequency;
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of milliseconds based on the intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param ms
     *            The number of milliseconds to use for the conversion.
     * @return the numbe of machine cycles represented by the specified number of milliseconds.
     */
    public static long cyclesFromMs( long ms ) {
        checkCPS();
        return cyclesFromMs( ms, cps );
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of milliseconds based on the specified CPU speed
     * timebase (frequency).
     * 
     * @param ms
     *            The number of milliseconds to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of machine cycles represented by the specified number of milliseconds.
     */
    public static long cyclesFromMs( long ms, long frequency ) {
        return cyclesFromS( ms, frequency ) / 1000L;
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of microseconds based on the intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param us
     *            The number of microseconds to use for the conversion.
     * @return the numbe of machine cycles represented by the specified number of microseconds.
     */
    public static long cyclesFromUs( long us ) {
        checkCPS();
        return cyclesFromUs( us, cps );
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of microseconds based on the specified CPU speed
     * timebase (frequency).
     * 
     * @param us
     *            The number of microseconds to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of machine cycles represented by the specified number of microseconds.
     */
    public static long cyclesFromUs( long us, long frequency ) {
        return cyclesFromS( us, frequency ) / 1000000;
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of nanoseconds based on the intrinsically
     * calculated CPU speed timebase (cps).
     * 
     * @param ns
     *            The number of nanoseconds to use for the conversion.
     * @return the numbe of machine cycles represented by the specified number of nanoseconds.
     */
    public static long cyclesFromNs( long ns ) {
        checkCPS();
        return cyclesFromNs( ns, cps );
    }

    /**
     * Calculates the number of machine cycles represented by the specified number of nanoseconds based on the specified CPU speed
     * timebase (frequency).
     * 
     * @param ns
     *            The number of nanoseconds to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return the numbe of machine cycles represented by the specified number of nanoseconds.
     */
    public static long cyclesFromNs( long ns, long frequency ) {
        return cyclesFromS( ns, frequency ) / 1000000000;
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return String containing real-time value and its scaled units designation.
     */
    public static String timeFromCycles( long cycles ) {
        checkCPS();
        return timeFromCycles( cycles, cps );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return String containing real-time value and its scaled units designation.
     */
    public static String timeFromCycles( long cycles, long frequency ) {
        String[] result = splitTimeFromCycles( cycles, frequency, -1 );
        return result[ 0 ] + " " + result[ 1 ];
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param scale
     *            the number of decimal places to which to round the result.
     * @return String containing real-time value and its scaled units designation.
     */
    public static String timeFromCycles( long cycles, int scale ) {
        checkCPS();
        return timeFromCycles( cycles, cps, scale );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @param scale
     *            the number of decimal places to which to round the result.
     * @return String containing real-time value and its scaled units designation.
     */
    public static String timeFromCycles( long cycles, long frequency, int scale ) {
        String[] result = splitTimeFromCycles( cycles, frequency, scale );
        return result[ 0 ] + " " + result[ 1 ];
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown. It separates the result into two Strings returned in a two element array of Strings, the zeroth element being the
     * value, the first element being a mnemonic describing the units being represented by the value.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitTimeFromCycles( long cycles ) {
        checkCPS();
        return splitTimeFromCycles( cycles, cps );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the
     * specified time base for its calculations. It separates the result into two Strings returned in a two element array of
     * Strings, the zeroth element being the value, the first element being a mnemonic describing the units being represented by the
     * value.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitTimeFromCycles( long cycles, long frequency ) {
        return splitTimeFromCycles( cycles, frequency, -1 );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the Time
     * classes intrinsic cps timebase for its calculations. If this timebase has not been initialized, a RuntimeExcpetion will be
     * thrown. It separates the result into two Strings returned in a two element array of Strings, the zeroth element being the
     * value, the first element being a mnemonic describing the units being represented by the value.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param scale
     *            the number of decimal places to which to round the result.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitTimeFromCycles( long cycles, int scale ) {
        checkCPS();
        return splitTimeFromCycles( cycles, cps, scale );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the
     * specified time base for its calculations. It separates the result into two Strings returned in a two element array of
     * Strings, the zeroth element being the value, the first element being a mnemonic describing the units being represented by the
     * value.
     * 
     * @param cycles
     *            The number of machine cycles to use for the conversion.
     * @param frequency
     *            The number of machine cycles in a real-time second (cps). Overrides the intrinsic cps value measured at startup.
     * @param scale
     *            the number of decimal places to which to round the result.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitTimeFromCycles( long cycles, long frequency, int scale ) {
        double time;
        String time_format;

        time = minFromCycles( cycles, frequency );
        if ( (long)Math.abs( time ) > 0 ) {
            time_format = TIME_MIN;
        }
        else {
            time = sFromCycles( cycles, frequency );
            if ( (long)Math.abs( time ) > 0 ) {
                time_format = TIME_S;
            }
            else {
                time = msFromCycles( cycles, frequency );
                if ( (long)Math.abs( time ) > 0 ) {
                    time_format = TIME_MS;
                }
                else {
                    time = usFromCycles( cycles, frequency );
                    if ( (long)Math.abs( time ) > 0 ) {
                        time_format = TIME_US;
                    }
                    else {
                        time = nsFromCycles( cycles, frequency );
                        time_format = TIME_NS;
                    }
                }
            }
        }

        String[] result = new String[ 2 ];

        if ( scale < 0 ) {
            // don't change scale
            result[ 0 ] = "" + time;
        }
        else {
            // round to scale decimal places
            result[ 0 ] = roundToDecimals( time, scale );
        }
        result[ 1 ] = time_format;

        return result;
    }

    /**
     * Helper method that rounds a double value to a particular decimal scale
     * 
     * @param time
     *            The value to scale
     * @param scale
     *            The number of decimal places to round to.
     * @return A String representing the formatted version of the specified value at the specified scale.
     */
    public static String roundToDecimals( double time, int scale ) {
        double factor = Math.pow( 10, scale );
        long time_long_int = Math.round( time * factor );

        // convert to string for manipulation
        String[] time_string = new String[ 2 ];
        time_string[ 0 ] = "" + time_long_int;
        int index_decimal = Math.max( 0, time_string[ 0 ].length() - scale );
        time_string[ 1 ] = time_string[ 0 ].substring( index_decimal );
        time_string[ 0 ] = time_string[ 0 ].substring( 0, index_decimal );

        // check if integer part is empty
        if ( time_string[ 0 ].equals( "" ) ) time_string[ 0 ] = "0";

        // check if decimal part is empty
        if ( time_string[ 1 ].equals( "" ) )
            return time_string[ 0 ];
        else
            // return original_int . scale decimal places
            return time_string[ 0 ] + "." + time_string[ 1 ];
    }

    /**
     * Formats to a String the specified time value in milliseconds. Resulting String output contains a scaled value along with the
     * mnemonic for the units represented by the value.
     * 
     * @param ms_time
     *            The number of milliseconds to format.
     * @return A string containing the formatted output.
     */
    public static String formatTime( double ms_time ) {
        String[] result = splitFormatTime( ms_time, -1 );
        return result[ 0 ] + " " + result[ 1 ];
    }

    /**
     * Formats to a String the specified time value in milliseconds. Resulting String output contains a scaled value along with the
     * mnemonic for the units represented by the value.
     * 
     * @param ms_time
     *            The number of milliseconds to format.
     * @param scale
     *            The number of decimal places to round to.
     * @return A string containing the formatted output.
     */
    public static String formatTime( double ms_time, int scale ) {
        String[] result = splitFormatTime( ms_time, scale );
        return result[ 0 ] + " " + result[ 1 ];
    }

    /**
     * Formats a String that represents the real time value represented by the number of milliseconds specified. It separates the
     * result into two Strings returned in a two element array of Strings, the zeroth element being the value, the first element
     * being a mnemonic describing the units being represented by the value.
     * 
     * @param ms_time
     *            The number of milliseconds to format.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitFormatTime( double ms_time ) {
        return splitFormatTime( ms_time, -1 );
    }

    /**
     * Formats a String that represents the real time value represented by the number of machine cycles passed in. It uses the
     * specified time base for its calculations. It separates the result into two Strings returned in a two element array of
     * Strings, the zeroth element being the value, the first element being a mnemonic describing the units being represented by the
     * value.
     * 
     * @param ms_time
     *            The number of milliseconds to format.
     * @param scale
     *            The number of decimal places to round to.
     * @return Two element array of type String containing a formatted floating- point real-time value in its zeroth element, and a
     *         mnemonic describing the units being represented by the value in its first element.
     */
    public static String[] splitFormatTime( double ms_time, int scale ) {
        double time;
        String time_format;

        time = ms_time / 1000.0 / 60.0;
        if ( (long)Math.abs( time ) > 0 ) {
            time_format = TIME_MIN;
        }
        else {
            time = ms_time / 1000.0;
            if ( (long)Math.abs( time ) > 0 ) {
                time_format = TIME_S;
            }
            else {
                time = ms_time;
                if ( (long)Math.abs( time ) > 0 ) {
                    time_format = TIME_MS;
                }
                else {
                    time = ms_time * 1000.0;
                    if ( (long)Math.abs( time ) > 0 ) {
                        time_format = TIME_US;
                    }
                    else {
                        time = ms_time * 1000000.0;
                        time_format = TIME_NS;
                    }
                }
            }
        }

        String[] result = new String[ 2 ];

        if ( scale < 0 ) {
            // don't change scale
            result[ 0 ] = "" + time;
        }
        else {
            // round to scale decimal places
            result[ 0 ] = roundToDecimals( time, scale );
        }
        result[ 1 ] = time_format;

        return result;
    }
}