/*
 * Created on Apr 15, 2004
 *
 * Waster.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.util.time;

import java.text.DecimalFormat;

/**
 * @author mbi Apr 15, 2004
 * 
 * $Id: Waster.java,v 1.1 2004/09/23 15:34:23 mbi Exp $
 * 
 * Waster
 * 
 * This class wastes time in a very accurate way. When instantiated, it calibrates itself such that
 * a very accurate amount of time can be "wasted".
 */
public class Waster {
    public static final int           ITERATIONS          = 8000;
    public static final int           INCREMENT           = 1000;
    public static final int           THRESHOLD           = 5000;
    public static final int           MAX_ITERATIONS      = 100000;
    public static final int           OVERHEAD_ITERATIONS = 1000000;

    private static int                _minimumIterations;
    private static long               _shortAverageCycles;
    private static long               _longAverageCycles;
    private static double             _sFudgeFactor;
    private static double             _msFudgeFactor;
    private static double             _usFudgeFactor;
    private static double             _nsFudgeFactor;

    public final static DecimalFormat fmt                 = (DecimalFormat)DecimalFormat.getInstance();

    static {
        fmt.applyLocalizedPattern( "##0%" );
        calibrate();
    }

    public final static int           minimumIterations   = _minimumIterations;
    public final static long          shortAverageCycles  = _shortAverageCycles;
    public final static long          longAverageCycles   = _longAverageCycles;
    public final static double        sFudgeFactor        = _sFudgeFactor;
    public final static double        msFudgeFactor       = _msFudgeFactor;
    public final static double        usFudgeFactor       = _usFudgeFactor;
    public final static double        nsFudgeFactor       = _nsFudgeFactor;

    private static final void calibrate() {
        long shortTimes[] = new long[ THRESHOLD ];
        int longTimeCount = (MAX_ITERATIONS - (THRESHOLD + INCREMENT)) / INCREMENT;
        long longTimes[] = new long[ longTimeCount ];
        long cycles, shortTotal, longTotal, shortAvg, longAvg;
        long actualCycles, totalCycles, iterations;
        long start, stop;
        int index, minimumCycles;
        double sFudge, msFudge, usFudge, nsFudge;

        // First, calculate the overhead for calling getCycles()
        long startCycles = Time.getCycles();
        long endCycles = 0;
        for ( long i = 0; i < OVERHEAD_ITERATIONS; i++ ) {
            endCycles = Time.getCycles();
        }
        long getCyclesOverhead = (endCycles - startCycles) / OVERHEAD_ITERATIONS;
        System.out.println( "Time.getCycles() overhead = " + getCyclesOverhead + " cycles (" + Time.usFromCycles( getCyclesOverhead ) + "us)" );

        // Cycle through short times and calculate number of cycles per loop iteration.
        index = 0;
        shortTotal = 0;
        for ( long i = 1; i < THRESHOLD; i++ ) {
            shortTotal += cycles = shortTimes[ index++ ] = loop( i );
        }

        // Cycle through long times and calculate number of cycles per loop iteration.
        index = 0;
        longTotal = 0;
        for ( long i = THRESHOLD + INCREMENT; i < MAX_ITERATIONS; i += INCREMENT ) {
            longTotal += cycles = longTimes[ index++ ] = loop( i );
        }

        // Calculate averages
        shortAvg = shortTotal / THRESHOLD;
        longAvg = longTotal / longTimeCount;

        // Search for point at which cycles/loop stablizes
        for ( index = 0; index < shortTimes.length; index++ ) {
            if ( shortTimes[ index ] <= longAvg ) break;
        }
        minimumCycles = index;

        // Check to be sure that we have a point at which it stablizes. Fatal error if not.
        if ( index == shortTimes.length ) {
            throw new RuntimeException( "Waster: calibration failed: THRESHOLD set too low." );
        }

        // Calculate fudge factor for nanoseconds
        totalCycles = Time.cyclesFromNs( 100 );
        iterations = totalCycles / shortAvg;
        start = Time.getCycles();
        for ( long i = 0; i < iterations; i++ );
        stop = Time.getCycles();
        actualCycles = stop - start;
        nsFudge = (double)totalCycles / (double)actualCycles;
        System.out.println( "100ns: iterations = " + iterations + ", totalCycles = " + totalCycles + ", actualCycles = " + actualCycles + ", nsFudge = " + nsFudge );

        // Calculate fudge factor for microseconds
        totalCycles = Time.cyclesFromUs( 100 );
        iterations = totalCycles / shortAvg;
        start = Time.getCycles();
        for ( long i = 0; i < iterations; i++ );
        stop = Time.getCycles();
        actualCycles = stop - start;
        usFudge = (double)totalCycles / (double)actualCycles;
        System.out.println( "100us: iterations = " + iterations + ", totalCycles = " + totalCycles + ", actualCycles = " + actualCycles + ", usFudge = " + usFudge );

        // Calculate fudge factor for milliseconds
        totalCycles = Time.cyclesFromMs( 100 );
        iterations = totalCycles / longAvg;
        start = Time.getCycles();
        for ( long i = 0; i < iterations; i++ );
        stop = Time.getCycles();
        actualCycles = stop - start;
        msFudge = (double)totalCycles / (double)actualCycles;
        System.out.println( "100ms: iterations = " + iterations + ", totalCycles = " + totalCycles + ", actualCycles = " + actualCycles + ", msFudge = " + msFudge );

        // Calculate fudge factor for seconds
        totalCycles = Time.cyclesFromS( 5 );
        iterations = totalCycles / longAvg;
        start = Time.getCycles();
        for ( long i = 0; i < iterations; i++ );
        stop = Time.getCycles();
        actualCycles = stop - start;
        sFudge = (double)totalCycles / (double)actualCycles;
        System.out.println( "  5s : iterations = " + iterations + ", totalCycles = " + totalCycles + ", actualCycles = " + actualCycles + ", sFudge = " + sFudge );

        // Populate final variables for later use.
        _minimumIterations = minimumCycles;
        _shortAverageCycles = shortAvg;
        _longAverageCycles = longAvg;
        _sFudgeFactor = sFudge;
        _msFudgeFactor = msFudge;
        _usFudgeFactor = usFudge;
        _nsFudgeFactor = nsFudge;
    }

    private static final long loop( long iterations ) {
        long start = Time.getCycles();
        for ( long i = 0; i < iterations; i++ );
        long stop = Time.getCycles();
        return (stop - start) / iterations;
    }

    /**
     * Dummy main just to verify that the class has initialized properly
     */
    public static final void main( String args[] ) {
        System.out.println( "Short Time Average ( <  " + THRESHOLD + " ) = " + _shortAverageCycles );
        System.out.println( "Long  Time Average ( >= " + THRESHOLD + " ) = " + _longAverageCycles );
        System.out.println( "Timing Stablizes at: " + minimumIterations + " iterations." );

        long start, stop;

        nsTest( 500 );
        usTest( 500 );
        msTest( 500 );
        sTest( 5 );
    }

    /**
     * Time accuracy testing methods
     * 
     * @param ns
     */
    private static final void nsTest( long ns ) {
        long start, stop;
        double error;

        System.out.print( "Testing " + ns + " ns..." );
        start = Time.getCycles();
        nsWaste( ns );
        stop = Time.getCycles();
        error = Math.abs( 1 - ((double)Time.cyclesFromNs( ns ) / (double)(stop - start)) );
        System.out.println( Time.nsFromCycles( stop - start ) + " ns (error = " + fmt.format( error ) + ")" );
    }

    private static final void usTest( long us ) {
        long start, stop;
        double error;

        System.out.print( "Testing " + us + " us..." );
        start = Time.getCycles();
        usWaste( us );
        stop = Time.getCycles();
        error = Math.abs( 1 - ((double)Time.cyclesFromUs( us ) / (double)(stop - start)) );
        System.out.println( Time.usFromCycles( stop - start ) + " us (error = " + fmt.format( error ) + ")" );
    }

    private static final void msTest( long ms ) {
        long start, stop;
        double error;

        System.out.print( "Testing " + ms + " ms..." );
        start = Time.getCycles();
        msWaste( ms );
        stop = Time.getCycles();
        error = Math.abs( 1 - ((double)Time.cyclesFromMs( ms ) / (double)(stop - start)) );
        System.out.println( Time.msFromCycles( stop - start ) + " ms (error = " + fmt.format( error ) + ")" );
    }

    private static final void sTest( long s ) {
        long start, stop;
        double error;

        System.out.print( "Testing " + s + " s..." );
        start = Time.getCycles();
        sWaste( s );
        stop = Time.getCycles();
        error = Math.abs( 1 - ((double)Time.cyclesFromS( s ) / (double)(stop - start)) );
        System.out.println( Time.sFromCycles( stop - start ) + " s (error = " + fmt.format( error ) + ")" );
    }

    /**
     * Time wasting methods
     * 
     * @param ns
     */
    public static final void nsWaste( long ns ) {
        long totalCycles = Time.cyclesFromNs( ns );
        totalCycles *= nsFudgeFactor;
        long iterations = totalCycles / longAverageCycles;
        for ( long i = 0; i < iterations; i++ );
    }

    public static final void usWaste( long us ) {
        long totalCycles = Time.cyclesFromUs( us );
        totalCycles *= usFudgeFactor;
        long iterations = totalCycles / longAverageCycles;
        for ( long i = 0; i < iterations; i++ );
    }

    public static final void msWaste( long ms ) {
        long totalCycles = Time.cyclesFromMs( ms );
        totalCycles *= msFudgeFactor;
        long iterations = totalCycles / longAverageCycles;
        for ( long i = 0; i < iterations; i++ );
    }

    public static final void sWaste( long s ) {
        long totalCycles = Time.cyclesFromS( s );
        totalCycles *= sFudgeFactor;
        long iterations = totalCycles / longAverageCycles;
        for ( long i = 0; i < iterations; i++ );
    }
}