/*
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

#ifndef surmamadu_util_time_estimate_GHz_h
#define surmamadu_util_time_estimate_GHz_h

/**
 * Determines clockspeed in GHz using clock_gettime, which is supposed
 * to have nanosecond resolution. Exits when a constant value has been
 * reached for a specified number of iterations. Includes standard 
 * deviation calculations and argument passing through the command line.
 */

#include <unistd.h>                     // for usleep function
#include <stdio.h>
#include <time.h>                       // For high resolution timer
#ifndef VXWORKS
#ifndef DARWIN
#include <sys/resource.h>               // For priority defines
#endif // DARWIN
#endif // VXWORKS
#include <math.h>                       // For calculating square root for the standard deviation
#include <string.h>                     // For parsing command arguments
#include <sys/time.h>

#include "getTimeRegister.h"            // machine independent macros/inlines for reading the processor cycle register.

#define BILLION  1000000000
#define MILLION 1000000

#define CYCLES 4                        // Number of cycles to determine frequency to 4 significant digits

unsigned long long estimate_GHz(unsigned _bound,
				unsigned _loops, 
                                unsigned _tolerance,
                                unsigned _tolerance_retries,
				const char* const _output) {
    unsigned long long int tot_avg = 0;     // Keep track of all averages

    int                     i, done, repeat;
    unsigned int            counter, const_counter, failed_attempts, tolerance;
#ifdef DARWIN
    struct timeval          tvstart, tvstop;
    struct timeval          sleep_time;
    struct timeval          sleep_time2;
#else // !DARWIN
    struct timespec         tvstart, tvstop;
    struct timespec         sleep_time;
    struct timespec         sleep_time2;
#endif // DARWIN
    long                    diff;
    unsigned long long int  cycles[2];
    unsigned long long int  seconds, nanoseconds, microseconds, total;
    unsigned long long int  cur_frequency, prev_frequency, tot_frequency, avg_frequency;
    unsigned long long int  const_frequency, tot_sqr_frequency, std_deviation;
   
    FILE                    *stream;
    int                     testloop = 0;
 
    for ( testloop = 1; testloop <= _loops; testloop++ ) {
        printf("Test %d of %d\n", testloop, _loops);
        done                = 0;
        cur_frequency       = 0;
        prev_frequency      = 0;
        tot_frequency       = 0;
        avg_frequency       = 0;
        const_frequency     = 0;
        tot_sqr_frequency   = 0;
        std_deviation       = 0;
        counter             = 0;
        repeat              = 0;
        const_counter       = 0;
        failed_attempts     = 0;
        tolerance           = _tolerance;

        if ( 0 == _tolerance_retries ) {
            _tolerance_retries = 5;
        }
        if ( 0 == tolerance ) {
            tolerance = 1;
        }
 
        if ( (stream = fopen( _output, "a" )) == NULL ) {
            printf("File error: %s\n", _output);
            return 0;
        }
        
    	if ( testloop == 1 ) {        // 100 ms
            sleep_time.tv_sec   = 0;
#ifdef DARWIN
            sleep_time.tv_usec  = 10000;
#else   // !DARWIN
            sleep_time.tv_nsec  = 100000000;
#endif  // DARWIN	    
            fprintf(stream, "# Test %d #\n", testloop);
            fprintf(stream, "# Sleep_time_%d = 100 ms\n", testloop);
        }
        else if ( testloop == 2)  {   // 1 sec
            sleep_time.tv_sec   = 1;
#ifdef DARWIN
            sleep_time.tv_usec  = 0;
#else   // !DARWIN
            sleep_time.tv_nsec  = 0;
#endif  // DARWIN	    
 
            fprintf(stream, "# Test %d #\n", testloop);
            fprintf(stream, "# Sleep_time_%d = 1 s\n", testloop);
        }
        else {                      // 10 sec
            sleep_time.tv_sec   = 10;
#ifdef DARWIN
            sleep_time.tv_usec  = 0;
#else   // !DARWIN
            sleep_time.tv_nsec  = 0;
#endif  // DARWIN	    
 
            fprintf(stream, "# Test %d #\n", testloop);
            fprintf(stream, "# Sleep_time_%d = 10 s\n", testloop);
        }

        fclose(stream);
	
	while( !done ) {
           //printf("In inner while loop, counter: %d, const_counter %d\n",
            //           counter, const_counter);
            // Get this function into cached memory
#ifdef  DARWIN
            gettimeofday( &tvstart, NULL );
#else   // !DARWIN
            clock_gettime( CLOCK_REALTIME, &tvstart );
#endif  // DARWIN

            // Measure, sleep, then measure again
            // 0x0f,0x31 is same as doing 'rdtsc' instruction
            // The sleep value is not part of the final calculation
            cycles[ 0 ] = getCycles();

#ifdef  DARWIN
            gettimeofday( &tvstart, NULL );
            usleep( (sleep_time.tv_sec * 1000000) + sleep_time.tv_usec );
#else   // !DARWIN
            clock_gettime( CLOCK_REALTIME, &tvstart );
            nanosleep( &sleep_time, NULL );
#endif  // DARWIN

            cycles[ 1 ] = getCycles();

#ifdef  DARWIN
            gettimeofday( &tvstop, NULL );
#else   // !DARWIN
            clock_gettime( CLOCK_REALTIME, &tvstop );
#endif  // DARWIN
            //cycles[ 1 ] = getCycles();

            seconds     = tvstop.tv_sec - tvstart.tv_sec - 1;
#ifdef  DARWIN
            microseconds = (MILLION - tvstart.tv_usec) + tvstop.tv_usec;
            if ( microseconds > MILLION ) {
                seconds++;
                microseconds -= MILLION;
            }
            total = (seconds * MILLION) + microseconds;
            cur_frequency = (unsigned long long int)(((cycles[ 1 ] - cycles[ 0 ]) * MILLION) / total);

#else   // !DARWIN
            nanoseconds  = (BILLION  - tvstart.tv_nsec) + tvstop.tv_nsec;
            if ( nanoseconds > BILLION ) {
                seconds++;
                nanoseconds -= BILLION;
            }
            total = (seconds * BILLION) + nanoseconds;
            cur_frequency = (unsigned long long int)(((cycles[ 1 ] - cycles[ 0 ]) * BILLION) / total);
#endif  // DARWIN

            // look for a constant frequency to 4 significant digits
            if ( const_counter < CYCLES ) {
                if ( const_frequency == (unsigned long long int)(cur_frequency / 1000000) * 1000000) {
                    ++const_counter;
                }
                else {
                    const_counter   = 0;
                    const_frequency = (unsigned long long int)(cur_frequency / 1000000) * 1000000;
                }
            }

            // start keeping track of changes in frequency in the lower 6 digits
            else {
                // subtract off the constant frequency to keep the numbers smaller
                tot_frequency      += cur_frequency - const_frequency;
                avg_frequency       = (unsigned long long int)(tot_frequency / (unsigned long long int)(++counter));
                tot_sqr_frequency  += (unsigned long long int)pow((float)(cur_frequency - const_frequency), 2);
        
                // See if value is repeating
                //printf("Checking done cond: avg: %lu", avg_frequency );
                //printf(", prev: %lu  \n", prev_frequency);

                diff = ( avg_frequency > prev_frequency ) ?
                       (long)( avg_frequency - prev_frequency ) :
                       (long)( prev_frequency - avg_frequency );
               
                //if ( avg_frequency == prev_frequency ) {
                if ( diff < tolerance ) {
                    repeat++;
                }
                else {
                    repeat = 0;
                    failed_attempts++;
                    if ( failed_attempts > _tolerance_retries ) {
                        tolerance <<= 1;
                        failed_attempts = 0;
                    }
                }
                //printf("repeat = %d...\n", repeat);

                // Exit if repetitions are sufficient
                if ( repeat == _bound ) {
                    if ( (stream = fopen(_output, "a")) == NULL ) {
                        printf("File error: %s\n", _output);
                        return 0;
                    }

                    std_deviation = (unsigned long long int)(0.5f + sqrt((float)tot_sqr_frequency / (float)counter
                                                                         - pow((float)avg_frequency, 2)));

                    fprintf(stream, "Tolerance = %u\n", tolerance);
                    fprintf(stream, "Average_%d = %llu\n", testloop, avg_frequency + const_frequency);
                    fprintf(stream, "SD_%d = %llu\n", testloop, std_deviation);
                    fprintf(stream, "Iterations_%d = %d\n\n", testloop, counter);
                    fclose(stream);
                    done = 1;
		    tot_avg += avg_frequency + const_frequency;
                }
                //printf("Assigning prev: avg: %lu,", avg_frequency);
                //printf(" prev: %lu\n", prev_frequency);
                prev_frequency = avg_frequency;
            }

#ifdef VXWORKS
	    sleep_time2.tv_sec = (time_t)((10000 + ((counter*7)%7000))/1000.0);
	    sleep_time2.tv_nsec = 10000 + ((counter*7)%7000);
            nanosleep(&sleep_time2, 0);
#else // VXWORKS
            // Sleep different amounts of time avoid potentially being caught in
            // same section of dispatch loop. Sleep ranges from 10ms to 16.993ms
            usleep( 10000 + ((counter*7)%7000) );
#endif // VXWORKS
        }
    }

    if ( (stream = fopen(_output, "a")) == NULL ) {
        printf("File error: %s\n", _output);
        return 0;
    }

    // write final to file.
    fprintf(stream, "cps = %llu\n\n", tot_avg / _loops );
    fclose(stream);
    return tot_avg;
}


#endif // file tag
