/*
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

#ifndef surmamadu_util_time_estimate_GHz_h
#define surmamadu_util_time_estimate_GHz_h

/**
 * Determines clockspeed in GHz using clock_gettime, which is supposed
 * to have nanosecond resolution. Exits when a constant value has been
 * reached for a specified number of iterations. Includes standard 
 * deviation calculations and argument passing through the command line.
 */

#include <unistd.h>                     // for usleep function
#include <fstream>
#include <iostream>
#include <time.h>                       // For high resolution timer
#ifndef VXWORKS
#include <sys/resource.h>               // For priority defines
#endif // VXWORKS
#include <cmath>                        // For calculating square root for the standard deviation
//#include <string.h>                     // For parsing command arguments

#include "getTimeRegister.h"            // machine independent macros/inlines for reading the processor cycle register.

namespace suramadu {
    namespace util {
	namespace time {

#ifdef ghs
	    // Greenhills can't print long longs!  DUH!
	    std::ostream& operator<<(std::ostream& os, unsigned long long ull) {
		char buffer[100];
		buffer[99] = 0;
		char* c = buffer+98;
		
		unsigned long long tmp;
		do {
		    tmp = ull / 10;
		    tmp *= 10;
		    tmp = ull - tmp;
		    *c-- = '0' + (char)tmp;
		    ull /= 10;
		}
		while (ull != 0);
		return (os << (c+1));
	    }
#endif // ghs

#define SURAMADU_UTIL_TIME_CLOCK CLOCK_REALTIME
//#define SURAMADU_UTIL_TIME_CLOCK CLOCK_HIGHRES

	    typedef unsigned long long ulonglong;

	    const ulonglong BILLION = 1000000000ULL;

	    /**  Used to reject absurdly large estimates of the clock speed, and in this
	     *  way protect the sum from overflowing.  On ggtb1, this cutoff results in fewer
	     *  than 10% of the estimates being capped off. */
	    const ulonglong MAX_ACCEPTED_FREQ_EST = (20 * BILLION);

	    ulonglong estimate_GHz(unsigned iterations, 
				   const char* const outFile) {
		ulonglong freqAverage = 0;     // Keep track of all averages

		//std::cout << 0x8000000000000000ULL << std::endl;
		struct timespec         res;
		int resStat = clock_getres(SURAMADU_UTIL_TIME_CLOCK, &res);
		std::cout << "Getres: " << res.tv_sec << "s " << res.tv_nsec << "ns" << std::endl;

#ifdef VXWORKS
		int sysClockRate = sysClkRateGet();
		std::cout << "sysClkRateGet: " << sysClockRate << " ticks/sec" << std::endl;
#endif // VXWORKS

		struct timespec         tvstart, tvstop;
		struct timespec         sleep_time;
		int                     numRejects = 0;
		ulonglong  cycles[2];
		ulonglong  *freqMeasurements = new ulonglong[iterations];

		std::ofstream           outStream;

		outStream.open(outFile, std::ios::out|std::ios::trunc);
		if ( ! outStream.is_open()) {
		    std::cerr << "File error: " << outFile << std::endl;
		    return 0;
		}

		std::cout << "Estimated time for estimation: "
			  << (int)((iterations * 3)/3600) << " hours, "
			  << (int)(((iterations * 3) - ((iterations * 3)/3600)*3600) / 60) + 1 << " min." << std::endl;

		for ( int testloop = 0; testloop < iterations; testloop++ ) {
 
		    sleep_time.tv_sec   = testloop % 5;
		    sleep_time.tv_nsec  = (ulonglong)((double)BILLION*(0.999 - (double)testloop/(double)iterations));
	    
		    outStream << "# Sleep_time_" << testloop+1 << " = " << sleep_time.tv_sec << "s"
			      << sleep_time.tv_nsec << "ns" << std::endl;

	
		    //  Just to make sure this function is cached before measuring.
		    clock_gettime( SURAMADU_UTIL_TIME_CLOCK, &tvstart );

		    // Measure, sleep, then measure again
		    cycles[ 0 ] = getCycles();
		    clock_gettime( SURAMADU_UTIL_TIME_CLOCK, &tvstart );
		    nanosleep( &sleep_time, NULL );
		    cycles[ 1 ] = getCycles();
		    clock_gettime( SURAMADU_UTIL_TIME_CLOCK, &tvstop );

		    //std::cout << " gettime stop(s,ns)/start: (" << tvstop.tv_sec << "," << tvstop.tv_nsec << ") ("
		    //      << tvstart.tv_sec << "," << tvstart.tv_nsec << ')' << std::endl;
		    ulonglong seconds     = tvstop.tv_sec - tvstart.tv_sec;
		    
		    ulonglong nanoseconds = seconds * BILLION;
		    nanoseconds += tvstop.tv_nsec;
		    nanoseconds -= tvstart.tv_nsec;

		    ulonglong diff = (cycles[1] - cycles[0]);

		    //std::cout << "cycle count: " << cycles[1] << ", " << cycles[0]
		    //      << ",  diff: " << diff << ", ns: " << nanoseconds << std::endl;

		    if (diff == 0ull) {
			freqMeasurements[testloop] = BILLION;
			outStream << "Estimate_guessed_" << testloop+1 << "=true # guessed because of 0 cycle diff." << std::endl;
		    }
		    else if (nanoseconds == 0ull) {
			freqMeasurements[testloop] = BILLION;
			outStream << "Estimate_guessed_" << testloop+1 << "=true # guessed because of 0 time diff." << std::endl;
		    }
		    else {
			freqMeasurements[testloop] = (ulonglong)((diff * BILLION) / nanoseconds);
		    }

		    if ( MAX_ACCEPTED_FREQ_EST < freqMeasurements[testloop]) {
			freqMeasurements[testloop] = MAX_ACCEPTED_FREQ_EST;
			++numRejects;
		    }
		    outStream << "Estimate_" << testloop+1 << " = "
			      << freqMeasurements[testloop] << std::endl;
		    freqAverage += freqMeasurements[testloop];

		    if ( 0 == (testloop+1)%10) { outStream.flush(); }
		}


		freqAverage /= iterations;
		double mean = static_cast<double>(freqAverage);
		double stddev = 0;
		for (int i = 0; i < iterations; ++i) {
		    stddev += (  freqMeasurements[i] * freqMeasurements[i]  - mean);
		}
		if ( iterations > 1) {
		    stddev /= (iterations - 1);
		}

		if (0 < numRejects) {
		    outStream << "Num estimates rejected because too large: " << numRejects << std::endl;
		}

		outStream << "stddev_frequency = " << stddev << std::endl;
		//  Greenhills doesn't know have an operator<< that takes long long...
		outStream << "cps = " << static_cast<unsigned long>(freqAverage) << std::endl;
		outStream.close();

		return freqAverage;
	    }
	}
    }
}


#endif // file tag
