/*
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED.
 *  United States Government Sponsorship acknowledged. Any commercial use must be
 *  negotiated with the Office of Technology Transfer at the California Institute
 *  of Technology.
 *
 *  This software may be subject to U.S. export control laws and regulations.
 *  By accepting this document, the user agrees to comply with all applicable U.S.
 *  export laws and regulations.  User has the responsibility to obtain export
 *  licenses, or other export authority as may be required before exporting such
 *  information to foreign countries or providing access to foreign persons.
 */
                                                                                                        
/**
 * Determines clockspeed in GHz using clock_gettime, which is supposed
 * to have nanosecond resolution. Exits when a constant value has been
 * reached for a specified number of iterations. Includes standard 
 * deviation calculations and argument passing through the command line.
 */
#include <unistd.h>                     // for usleep function
#include <stdio.h>
#include <time.h>                       // For high resolution timer
#include <sys/time.h>
#ifndef VXWORKS
#ifndef DARWIN
#include <sys/resource.h>               // For priority defines
#endif // DARWIN
#endif // VXWORKS
#include <math.h>                       // For calculating square root for the standard deviation
#include <string.h>                     // For parsing command arguments

#include "getTimeRegister.h"            // machine independent macros/inlines for reading the processor cycle register.

#define BILLION 1000000000
#define MILLION 1000000
#define TESTLENGTH 5

#define CYCLES 4                        // Number of cycles to determine frequency to 4 significant digits

int main( int argc, char **argv ) {
    unsigned long long cycles1, cycles2, cycleDiff, total, seconds, useconds, nseconds, i;

#ifdef  DARWIN
    struct timeval  tvstart, tvstop, sleep_time;
    sleep_time.tv_usec = 0;
#else   // !DARWIN    
    struct timespec tvstart, tvstop, sleep_time;
    sleep_time.tv_nsec = 0;
#endif

    sleep_time.tv_sec  = TESTLENGTH;

#ifdef DARWIN
    gettimeofday( &tvstart, NULL );
#else
    clock_gettime( CLOCK_REALTIME, &tvstart );
#endif // DARWIN

    cycles1 = getCycles();
    usleep( TESTLENGTH * 1000000 );
    cycles2 = getCycles();

#ifdef DARWIN
    gettimeofday( &tvstop, NULL );
#else
    clock_gettime( CLOCK_REALTIME, &tvstop );
#endif // DARWIN

    seconds   = tvstop.tv_sec  - tvstart.tv_sec - 1;

#ifdef  DARWIN
    useconds  = (BILLION - tvstart.tv_usec) + tvstop.tv_usec;
    if ( useconds > MILLION ) {
        seconds++;
        useconds -= MILLION;
    }
    cycleDiff = cycles2 - cycles1;

    printf( "cycles1 = %llu, cycles2 = %llu\n", cycles1, cycles2 );
    printf( "seconds=%llu, useconds=%llu, cycleDiff=%llu\n", seconds, useconds, cycleDiff / TESTLENGTH );
#else   // !DARWIN
    nseconds  = (BILLION - tvstart.tv_nsec) + tvstop.tv_nsec;
    if ( nseconds > BILLION ) {
        seconds++;
        nseconds -= BILLION;
    }

    printf( "cycles1 = %llu, cycles2 = %llu\n", cycles1, cycles2 );
    printf( "seconds=%llu, nseconds=%llu, cycleDiff=%llu\n", seconds, nseconds, cycleDiff / TESTLENGTH );
#endif

    cycles1 = getCycles();
    for ( i = 0; i < BILLION; i++ );
    cycles2 = getCycles();
    cycleDiff = cycles2 - cycles1;
    printf( "cycles1 = %llu, cycles2 = %llu, difference = %llu\n", cycles1, cycles2, cycleDiff );
    printf( "cycles per iteration = %f\n", (double)cycleDiff / (double)BILLION );
    printf( "iterations per cycle = %f\n", (double)BILLION   / (double)cycleDiff );
}


