/*
 * Created on Jun 8, 2004
 *
 * AsynchronousEvent.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

import java.util.Iterator;
import javax.realtime.AsyncEvent;
import javax.realtime.AsyncEventHandler;
import javax.realtime.BoundAsyncEventHandler;
import javax.realtime.ImmortalMemory;
import javax.realtime.MemoryArea;
import javax.realtime.MemoryParameters;
import javax.realtime.PriorityParameters;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.Range;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;
import suramadu.util.time.Waster;

/**
 * <p>
 * <B><I>PURPOSE: </I> </B> Measure the latency between the firing of an asynchronous event and
 * the activation of its handler in the presence of other thread activity. Both bound and unbound
 * event handlers may be used.
 * <p>
 * <B><I>DESIGN: </I> </B> The primary thread takes a timestamp then fires an asynchronous event.
 * The handler takes a timestamp when it is activated. A global index, maintained by the primary
 * thread, is used to keep the trigger and handle times correctly paired. Data produced is the 
 * difference between the trigger and handler times. Memory, thread type, handler type, 
 * priorities and the amount of time the primary thread continues to operate
 * are all configurable.
 * <p>
 * A variable number of background threads can be run. Each background thread produces a
 * random amount of garbage (an array of integers with a random length). 
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> 
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"100"</td>
 * <td>Number of firing/handling tries</td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"200"</td>
 * <td>Priority of the primary (firing) thread</td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"100"</td>
 * <td>Period (ms) of the primary thread</td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td>Initial delay (ms) of the primary thread</td>
 * </tr>
 * <tr>
 * <td>"threadtype"</td>
 * <td>"RT"</td>
 * <td>Primary thread type (regular java, RT or NHRT)</td>
 * </tr>
 * <tr>
 * <td>"memtype"</td>
 * <td>"HEAP"</td>
 * <td>Memory type for the primary thread</td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"100"</td>
 * <td>Background thread priority</td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"100"</td>
 * <td>Period (ms) of the background threads</td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td>Initial delay (ms) for the background threads</td>
 * </tr>
 * <tr>
 * <td>"backgroundthreads"</td>
 * <td>"0"</td>
 * <td>Number of background threads</td>
 * </tr>
 * <tr>
 * <td>"handlertype"</td>
 * <td>"BOUND"</td>
 * <td>Type of handler. Other possible value is 'UNBOUND'</td>
 * </tr>
 * <tr>
 * <td>"handlerpriority"</td>
 * <td>"250"</td>
 * <td>Priority of the handler</td>
 * </tr>
 * <tr>
 * <td>"handlernonheap"</td>
 * <td>"TRUE"</td>
 * <td>If handler should have properties of an NHRT. Other possible value is 'FALSE'. If
 * <i>handlertype</i> is UNBOUND this value should be FALSE.</td>
 * </tr>
 * <tr>
 * <td>"waste"</td>
 * <td>"0"</td>
 * <td>Amount of time (ns) the primary thread should continue to run after firing the event.</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> Two data files:
 * suramadu.tests.latency.AsynchronousEvent.(timestamp).latency and
 * suramadu.tests.latency.AsynchronousEvent.(timestamp).SUMMARY
 * <p>
 * <B><I>EXAMPLES: </I> </B>
 * <p>
 * <B><I>ORIGINS: </I> </B>
 * <p>
 * 
 * @author Joseph Hutcherson
 * 
 * @version $Revision: 1.1 $
 */
public class AsynchronousEvent implements Runnable {

	private static final String BOUND   = "BOUND";
	private static final String UNBOUND = "UNBOUND";
	
	public static class TestProps extends TestProperties {
	    public static final String[] DEFAULT_PROPERTIES = {
    		"memtype",         "HEAP", 
			"threadtype",      "RT", 
			"iterations",      "100", 
			"priority",        "200", 
			"period",          "100", 
			"idelay",          "750", 
			"bgthreads",       "0", 
			"bgpriority",      "100", 
			"bgperiod",        "100", 
			"bgidelay",        "400", 
			"histbins",        "50",
			"handlertype",     BOUND,
			"handlerpriority", "250",
			"handlernonheap",  "FALSE",
			"waste",           "0",
	    };

	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void loadArguments() throws PropertyValidationException {
	        threadType = getInputProperty( "threadtype" );
	        memType = getInputProperty( "memtype" );
	        iterations = getIntInputProperty( "iterations" );
	        priority = getIntInputProperty( "priority" );
	        period = getIntInputProperty( "period" );
	        idelay = getIntInputProperty( "idelay" );
	        bgThreads = getIntInputProperty( "bgthreads" );
	        bgPriority = getIntInputProperty( "bgpriority" );
	        bgPeriod = getIntInputProperty( "bgperiod" );
	        bgIdelay = getIntInputProperty( "bgidelay" );
	        histBins = getIntInputProperty( "histbins" );
	        handlerType = getInputProperty( "handlertype" );
	        handlerPriority = getIntInputProperty( "handlerpriority" );
	        handlerNonHeap = getBooleanInputProperty( "handlernonheap" );
	        waste = getIntInputProperty( "waste" );
	    }

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void validateArguments() throws PropertyValidationException {
	        boolean ok = true;

	        if ( 0 > iterations ) {
	            ok = false;
	            System.out.println( "iterations property must be >= 0. Using default: " + iterations );
	        }
	        if ( 0 > priority ) {
	            ok = false;
	            System.out.println( "priority property must be >= 0. Using default: " + priority );
	        }
	        if ( 0 > iterations ) {
	            ok = false;
	            System.out.println( "period property must be >= 0. Using default: " + period );
	        }
	        if ( 0 > idelay ) {
	            ok = false;
	            System.out.println( "idelay property must be >= 0. Using default: " + idelay );
	        }
	        if ( 0 > bgThreads ) {
	            ok = false;
	            System.out.println( "bgthreads property must be >= 0. Using default: " + bgThreads );
	        }
	        if ( 0 > bgPriority ) {
	            ok = false;
	            System.out.println( "bgpriority property must be >= 0. Using default: " + bgPriority );
	        }
	        if ( 0 > bgPeriod ) {
	            ok = false;
	            System.out.println( "bgperiod property must be >= 0. Using default: " + bgPeriod );
	        }
	        if ( 0 > bgIdelay ) {
	            ok = false;
	            System.out.println( "bgidelay property must be >= 0. Using default: " + bgIdelay );
	        }
	        if ( 0 > handlerPriority ) {
	            ok = false;
	            System.out.println( "handlerpriority property must be >= 0. Using default: " + handlerPriority );
	        }
	        if ( 0 > waste ) {
	            ok = false;
	            System.out.println( "waste property must be >= 0. Using default: " + waste );
	        }
	        if ( !handlerType.equals(BOUND) && !handlerType.equals(UNBOUND) ) {
	            ok = false;
	            System.out.println( "handlertype property must either '" + BOUND + "' or '" +
	            		UNBOUND + "'. Using default: " + handlerType );
	        }
	        if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
	    }

	    /**
	     * actual properties used by this test run. Filled in from default and specified properties at class instantiation.
	     */
	    /** retrieved from default or specified properties */
	    public static String         threadType;
	
	    /** retrieved from default or specified properties */
	    public static String         memType;
	
	    /** retrieved from default or specified properties. */
	    public static int            iterations;
	
	    /** retrieved from default or specified properties. */
	    public static int            priority;
	
	    /** retrieved from default or specified properties. */
	    public static long           period;
	
	    /** retrieved from default or specified properties. */
	    public static long           idelay;
	
	    /** retrieved from default or specified properties. */
	    public static int            bgPriority;
	
	    /** retrieved from default or specified properties. */
	    public static long           bgPeriod;
	
	    /** retrieved from default or specified properties. */
	    public static long           bgIdelay;
	
	    /** retrieved from default or specified properties. */
	    public static int            bgThreads;
	
	    /** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
	    public static int            histBins;
	    
	    /** type of handler (bound or unbound) to use */
	    public static String 	     handlerType;

	    /** priority of handler (bound or unbound) to use */
	    public static int 			 handlerPriority;

	    /** value of nonheap flag to unbound handlers */
	    public static boolean 		 handlerNonHeap;

	    /** time (ns) to waste in the driver thread to see effects on handler pre-emption */
	    public static int 			 waste;
	}

	/**
	 * Local data for test: generated or otherwise calculated. Not from test properties.
	 */
	/** Test properties for this run */
    public TestProps 	  testProps;
    
    /** Thread type to use. */
    public ThreadType     tt;

    /** Memory type to use */
    public MemoryType     mt;

    /** array for collecting timestamps during test */
    public static long			triggerTime[];
    public static long			handlerStartTime[];

    /** iteration counter */
    public  static int           counter = -1;
    
    /** The event and handler */
	public static AsyncEvent event;
	public static AsyncEventHandler handler;

	/** Flag for the background threads */
	boolean keepRunning = true;
	
	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public AsynchronousEvent() throws Exception {
        this( new TestProps( AsynchronousEvent.class ) );
    }

    public AsynchronousEvent( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
        counter   = -1;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( AsynchronousEvent.class, TestProps.class );
    }

    public void run() {
        try {
            /**
             * Since we need iterations + 1 number of measurements to get iterations number of timings, increment iterations before
             * using it.
             */
        	TestProps.iterations++;

			/** Calibrate waster here, as it is a lengthy process */
			Waster.nsWaste(0);
        	
            ThreadLauncher[] bgThreadArray = null;
            if ( 0 < TestProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    double randomDelay = Math.random() * TestProps.bgIdelay;
                    bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( ThreadType.PERIODICJAVATHREAD, MemoryType.HEAPMEMORY, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                    bgThreadArray[ i ].launch();
                }
            }

            int duration = (int) (((float) TestProps.iterations * TestProps.period) / 1000);
            System.out.println("This will take about " + duration + " seconds to run");            
            
            /** Can vary thread types used */
            tt = ThreadType.getThreadType( TestProps.threadType );

            /** Can vary memory types used */
            mt = MemoryType.getMemoryType( TestProps.memType );

            ThreadLauncher launcher = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( tt, mt, new TestInit(), TestProps.priority, TestProps.period, TestProps.idelay ) );
            launcher.setIterations( TestProps.iterations );
            Thread t = launcher.launch();

            // wait for the test to complete (based on a calculation of how long it will take)
            t.join();

            // Kill off background threads.
            keepRunning = false;
            if ( null != bgThreadArray ) {
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    bgThreadArray[ i ].kill();
                }
            }

            // Latency (time from trigger to handler)
            long[] cyclesLate = new long[TestProps.iterations + 1];

            for (int i = 0; i < cyclesLate.length; i++ ) {
                cyclesLate[i] = handlerStartTime[i] - triggerTime[i];
            }

            // Convert to ns from cycles
            long[] nsLate = Time.nsFromCyclesL(cyclesLate);

            testProps.setOutputProperty( "latency", nsLate );
            testProps.setOutputProperty( "latencymax", StatPack.max( nsLate ) );
            testProps.setOutputProperty( "latencymin", StatPack.min( nsLate ) );
            testProps.setOutputProperty( "latencymean", StatPack.mean( nsLate ) );
            testProps.setOutputProperty( "latencystddev", StatPack.stddev( nsLate ) );
            testProps.setOutputProperty( "latencyvariance", StatPack.variance( nsLate ) );

            Range mode = StatPack.mode( nsLate, TestProps.histBins );
            testProps.setOutputProperty( "latencymode.lobound", (Number)mode.getLoBound() );
            testProps.setOutputProperty( "latencymode.hibound", (Number)mode.getHiBound() );
            testProps.setOutputProperty( "latencymode.count",   mode.getCount() );
            testProps.setOutputProperty( "latencymode.percent", (double)mode.getCount() * 100.0 / ((double)TestProps.iterations - 1) );

            // report on what we're going to use.
            testProps.close();            ;
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /** Initializes the elements to be used in the test.*/
    class TestInit extends TestInitializer {

		public TestInit() {
		}

		/** All activities here occur at the memory type specified for the primary thread */
		public void run() {

			event = new AsyncEvent();

			if (TestProps.handlerType.equals(BOUND)) {
				
				/** The bound handler will inherit the properties of this thread.
				 * Note that the priority of the handler is set directly.
				 */
				MemoryArea memory = null;
				if (TestProps.handlerNonHeap) {
					memory = ImmortalMemory.instance();
				}
				handler = new BoundHandler(TestProps.handlerPriority,
						TestProps.handlerNonHeap, memory);
			
			} else if (TestProps.handlerType.equals(UNBOUND)) {
			
				handler = new UnBoundHandler(TestProps.handlerPriority,
						TestProps.handlerNonHeap);
			
			} else {
				System.out.println("Unknown handler type: "
						+ TestProps.handlerType);
			}

			// Allocate timestamp arrays to record data
			triggerTime = new long[TestProps.iterations + 1];
			handlerStartTime = new long[TestProps.iterations + 1];

			/** Add the handler to the event. Note that use of event.setHandler()
			 * caused problems with both Mackinac and TimeSys as of Aug 2004.
			 */
			event.addHandler(handler);
			setTestRunner(new Test(event));
		}
	}

    /** Fairly simple class for the primary thread. When activated, it simply
     * records the time at which it fires an event.
     */
    class Test implements Runnable {
		AsyncEvent myEvent;

		public Test(AsyncEvent event) {
			myEvent = event;
		}

		public void run() {

			// This thread drives the counter used by itself and the event handler.
			// Set counter here to avoid any timing issues with this thread continuing
			// to run or handler pre-emption.
			counter++;
			
			// debug
			//System.out.println("Firing event " + counter);

			// Take timestamp
			triggerTime[counter] = Time.getCycles();

			// Fire the event. This thread WILL continue to run.
			myEvent.fire();

			// If set, waste the specified amount of nanoseconds.
			if (TestProps.waste > 0) {
				Waster.nsWaste(TestProps.waste);
			}
		}
	}

    /** Simple background thread class to generate garbage.
     */
    class BGThread implements Runnable {
        int[] garbage;

        public void run() {
            if ( null == garbage ) {
                garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
            }
            else {
                garbage = null;
            }
        }
	}
}

/**
 * Bound handler for the events fired by the primary thread. Inherits properties
 * from the TestInit thread. Records the time at which the handler is activated.
 */
class BoundHandler extends BoundAsyncEventHandler {

	/** Assigns the priority to the handler */
	public BoundHandler(int priority, boolean nonheap, MemoryArea memory) {
		/** Note that this construct actually requires a seventh parameter, a Runnable.
		 * This is not in the RTSJ book I'm using, so it'll get null for now.
		 */
		super(new PriorityParameters(priority), null,
				new MemoryParameters(-1,-1), memory, null, nonheap, null);
		//this.setSchedulingParameters(new PriorityParameters(priority));
		// debug
		System.out.println("Using Bound handler at priority "
				+ ((PriorityParameters) this.getSchedulingParameters())
						.getPriority());
	}
	
	public void handleAsyncEvent() {
		
		// Record start time of the handler. counter is set by driver thread.
		AsynchronousEvent.handlerStartTime[AsynchronousEvent.counter] = Time.getCycles();
		
		// debug
		//System.out.println("BoundHandler.handleAsyncEvent");
		
	}
}

/** Unbound handler. Records the time at which the handler is activated.
 */
class UnBoundHandler extends AsyncEventHandler {
	
	/** Assigns the priority to the handler */
	public UnBoundHandler(int priority, boolean nonheap) {
		super(nonheap);
		this.setSchedulingParameters(new PriorityParameters(priority));
		// debug
		System.out.println("Using Unbound handler at priority " +
                ((PriorityParameters)this.getSchedulingParameters()).getPriority() );		
	}
	
	public void handleAsyncEvent() {
		
		// Record start time of the handler. counter is set by driver thread.
		AsynchronousEvent.handlerStartTime[AsynchronousEvent.counter] = Time.getCycles();
		
		// debug
		//System.out.println("UnBoundHandler.handleAsyncEvent");
		
	}
}