/*
 * Created on Jun 8, 2004
 *
 * ClassLoading.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

/**
 * 
 * <p>
 * <B><I>PURPOSE: </I> </B> Measure the amount of time it takes to load a class.
 * <p>
 * <B><I>DESIGN: </I> </B> Over a number of samples, measure times before and after
 * loading a class.
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> 
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"100"<td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"200"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgthreads"</td>
 * <td>"1"</td>
 * <td></td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode statistical
 * measurements on the collected data.
 * <p>
 * <B><I>EXAMPLES: </I> </B> None.
 * <p>
 * <B><I>ORIGINS: </I> </B> This test is based upon ideas from the AFRL and the rtpresto suites.
 * <p>
 * 
 * @see suramadu.tests.jitter.JavaJitterTest
 *      <p>
 * @author Mark Indictor, Dan Allard
 */


import java.io.IOException;
import java.util.Iterator;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * @author mbi, dan
 * Jun 8, 2004
 * 
 * $Id: ClassLoading.java,v 1.1 2004/09/23 15:34:04 mbi Exp $
 *
 * ClassLoading
 */

public class ClassLoading {
	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
			"memtype",    "IMMORTAL",
			"threadtype", "NHRT",
			"iterations", "1",
			"priority",   "250",
			"period",     "250",
			"idelay",     "100",
			"bgthreads",  "0",
			"bgpriority", "250",
			"bgperiod",   "100",
			"bgidelay",   "400",
			"histbins",   "50",
			"numSamples", "10",
			"loadClass",  LOAD_CLASS_NAME
	    };
	    
		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	       return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

		protected void loadArguments() throws PropertyValidationException {
			// Load all of the thread params
			iterations = getIntInputProperty("iterations");
			priority = getIntInputProperty("priority");
			period = getIntInputProperty("period");
			idelay = getIntInputProperty("idelay");
			bgThreads = getIntInputProperty("bgthreads");
			bgPriority = getIntInputProperty("bgpriority");
			bgPeriod = getIntInputProperty("bgperiod");
			bgIdelay = getIntInputProperty("bgidelay");
			histBins = getIntInputProperty("histbins");
			priority = getIntInputProperty("priority");

			// Load the other test params
			String threadType = getInputProperty("threadtype");
			String memType = getInputProperty("memtype");
			numSamples = getIntInputProperty("numSamples");
			loadClass = getInputProperty("loadClass");

			// Convert thread and memory types to actual types
			tt = ThreadType.getThreadType(threadType);
			mt = MemoryType.getMemoryType(memType);

		}

		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void validateArguments() throws PropertyValidationException {
			boolean ok = true;

			if (0 > iterations) {
				ok = false;
				System.out.println(
					"iterations property must be >= 0. Using default: "
						+ iterations);
			}
			if (0 > priority) {
				ok = false;
				System.out.println(
					"priority property must be >= 0. Using default: "
						+ priority);
			}
			if (0 > iterations) {
				ok = false;
				System.out.println(
					"period property must be >= 0. Using default: " + period);
			}
			if (0 > idelay) {
				ok = false;
				System.out.println(
					"idelay property must be >= 0. Using default: " + idelay);
			}
			if (0 > bgThreads) {
				ok = false;
				System.out.println(
					"bgthreads property must be >= 0. Using default: "
						+ bgThreads);
			}
			if (0 > bgPriority) {
				ok = false;
				System.out.println(
					"bgpriority property must be >= 0. Using default: "
						+ bgPriority);
			}
			if (0 > bgPeriod) {
				ok = false;
				System.out.println(
					"bgperiod property must be >= 0. Using default: "
						+ bgPeriod);
			}
			if (0 > bgIdelay) {
				ok = false;
				System.out.println(
					"bgidelay property must be >= 0. Using default: "
						+ bgIdelay);
			}
			if (!ok)
				throw new PropertyValidationException("One or more input properties are invalid.");
		}

		/** retrieved from default or specified properties */
		public static int iterations;
		public static int priority;
		public static long period;
		public static long idelay;
		public static int bgPriority;
		public static long bgPeriod;
		public static long bgIdelay;
		public static int bgThreads;

		/** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
		public static int histBins;

		/** Thread type to use. */
		public static ThreadType tt;

		/** Memory type to use */
		public static MemoryType mt;

		/** Number of test samples */
		public static int numSamples;
		public static String loadClass = null;
	}

	//	Thread priorities

	static final String LOAD_CLASS_NAME =
		"suramadu.tests.latency.ClassLoadingLoadClass";

    /**
     * 
     * @uml.property name="testProps"
     * @uml.associationEnd 
     * @uml.property name="testProps" multiplicity="(1 1)"
     */
    private TestProps testProps;

	// TODO: Put into properties?
	private final static int THREAD_SLEEP_TIME = 100;

	//private static TimeStampLog loadTimeLog = null;
    private static long[] loadTimeLog = null;
	private static ClassLoadingLoader loader = null;
	private static boolean calculated = false;

	
	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public ClassLoading() throws Exception {
        this( new TestProps( ClassLoading.class ) );
    }

    public ClassLoading( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
    }
    
	/**
	 * Start the test
	 *  
	 */
	public void run() {
		try {
			// Setup the high priority thread using the ThreadLauncher to
			// manage it
		    System.out.println(testProps.toString());

			// TODO: figure out values for period and delay
			ClassLoaderInit cli = new ClassLoaderInit();

			ThreadLauncherData data =
				new ThreadLauncherData(
					TestProps.tt,
					TestProps.mt,
					cli,
					TestProps.priority,
					TestProps.period,
					TestProps.idelay);

			// We only need to run the thread once
			data.setIterations(1);

			ThreadLauncher highPriorityLauncher =
				ThreadLauncherFactory.getLauncher(data);
			highPriorityLauncher.setIterations(1);

			ThreadLauncher[] bgThreadArray = null;
            if ( 0 < TestProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    double randomDelay = Math.random() * TestProps.bgIdelay;
                    bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( ThreadType.PERIODICJAVATHREAD, MemoryType.HEAPMEMORY, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                    bgThreadArray[ i ].setIterations( - 1 );
                    bgThreadArray[ i ].launch();
                }
            }
			Thread bl = highPriorityLauncher.launch();

			// wait for the test to complete
			bl.join();

//			 Kill off background threads.
            if ( null != bgThreadArray ) {
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    bgThreadArray[ i ].kill();
                }
            }
            
			//			Now do some calculations
			calculate();

			// Start the lower priority thread
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	/**
	 * 
	 */
	class ClassLoaderInit extends TestInitializer {

        /**
         * 
         * @uml.property name="numSamples"
         */



		public void run() {
            loader = new ClassLoadingLoader();
			
			ClassLoaderThread thread = new ClassLoaderThread();

			loadTimeLog = new long[TestProps.numSamples];

			// We need to make a new string out of the load class
			// since now it's likely to be on the heap
			String myLoadClass = new String(TestProps.loadClass);
			thread.setLoadClass(myLoadClass);

			setTestRunner(thread);
		}

	}

	class ClassLoaderThread implements Runnable {
			String loadClass = null;

		/**
		 * @param myLoadClass
		 */
		public void setLoadClass(String myLoadClass) {
			loadClass = myLoadClass;

		}

		public void run() {
//	System.out.println("Running... " + TestProps.numSamples);
			// Loop for n samples, loading and unloading the 
			// loaderClass and calculating the load time
			for (int index = 0; index < TestProps.numSamples; index++) {
				// We need to recreate the loader on each loop to 
				// reload the class
				ClassLoader loader = new ClassLoadingLoader();
				long beforeTime = -1;
				long afterTime = -1;

				try {

					beforeTime = Time.getCycles();

					loader.loadClass(loadClass);

					afterTime = Time.getCycles();


				} catch (ClassNotFoundException e) {
					System.out.println(
						"Invalid load class specified: " + loadClass);
					System.out.println("Exiting ClassLoader...");
					return;
				} catch (Exception e) {
				   e.printStackTrace();
				}
				long diff = Math.abs(beforeTime - afterTime);
//				System.out.println("Diff is " + diff);
				loadTimeLog[index] = diff;

			
				// Take timestamp before load and after 
			}

		}
	}

	private void calculate() {
		// Get everything out of the loadTimeLog

		long[] results = loadTimeLog;
		testProps.setOutputProperty("loadtimes", results);
		testProps.setOutputProperty("max", StatPack.max(results));
		testProps.setOutputProperty("min", StatPack.min(results));
		testProps.setOutputProperty("mean", StatPack.mean(results));
		testProps.setOutputProperty("stddev", StatPack.stddev(results));
		testProps.setOutputProperty("variance", StatPack.variance(results));

		try {
			testProps.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
       ThreadLauncherFactory.startTest( ClassLoading.class, TestProps.class );
    }
    
    class BGThread implements Runnable {
       int[] garbage;

       public void run() {
           if ( null == garbage ) {
               garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
           }
           else {
               garbage = null;
           }
       }
   }
}
