/*
 * Created on May 27, 2004 Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. United States Government
 * Sponsorship acknowledged. Any commercial use must be negotiated with the Office of Technology Transfer at the California
 * Institute of Technology. This software may be subject to U.S. export control laws and regulations. By accepting this document,
 * the user agrees to comply with all applicable U.S. export laws and regulations. User has the responsibility to obtain export
 * licenses, or other export authority as may be required before exporting such information to foreign countries or providing access
 * to foreign persons.
 */
package suramadu.tests.latency;

import java.io.IOException;
import java.util.Iterator;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;
import suramadu.util.time.TimeStampLog;
import suramadu.util.time.Waster;

/**
 * @author dan Measure the context switch latency between threads <B><I>INPUTS: </I> </B> <br>
 *         <table>
 *         <tr>
 *         <th align="left">Property</th>
 *         <th align="right">Default Value</th>
 *         <th align="left">Description</th>
 *         </tr>
 *         <tr>
 *         <td>"iterations"</td>
 *         <td>"100"</td>
 *         <td></td>
 *         </tr>
 *         </table> <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode
 *         statistical measurements on the collected data.
 *         <p>
 *         <p>
 *         <B><I>ORIGINS: </I> </B> This test is based upon ideas from the AFRL and the rtpresto suites.
 *         <p>
 *         Notes: Uses the Suramadu time native methods to perform timestamping in lieu of RTSJ Time classes
 */
public class ContextSwitch {
    public static class TestProps extends TestProperties {
        public static final String[] DEFAULT_PROPERTIES = { "memtype", "IMMORTAL", "threadtype", "NHRT", "numSamples", "10", "iterations", "1", "priority", "250", "period", "250", "idelay", "100", "bgthreads", "0",
                                                                "bgpriority", "250", "bgperiod", "100", "bgidelay", "400", "histbins", "50" };

        public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
            return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
        }

        /**
         * @throws PropertyValidationException
         */
        public TestProps() throws PropertyValidationException {
            super();
        }

        /**
         * @param aClass
         * @throws PropertyValidationException
         */
        public TestProps( Class aClass ) throws PropertyValidationException {
            super( aClass, DEFAULT_PROPERTIES );
        }

        protected void loadArguments() throws PropertyValidationException {
            // Load all of the thread params
            iterations = getIntInputProperty( "iterations" );
            numSamples = getIntInputProperty( "numSamples");
            priority = getIntInputProperty( "priority" );
            period = getIntInputProperty( "period" );
            idelay = getIntInputProperty( "idelay" );
            bgThreads = getIntInputProperty( "bgthreads" );
            bgPriority = getIntInputProperty( "bgpriority" );
            bgPeriod = getIntInputProperty( "bgperiod" );
            bgIdelay = getIntInputProperty( "bgidelay" );
            histBins = getIntInputProperty( "histbins" );
            priority = getIntInputProperty( "priority" );
            // Load the other test params
            String threadType = getInputProperty( "threadtype" );
            String memType = getInputProperty( "memtype" );
            loadClass = getInputProperty( "loadClass" );
            // Convert thread and memory types to actual types
            tt = ThreadType.getThreadType( threadType );
            mt = MemoryType.getMemoryType( memType );
        }

        /**
         * Saves argument values in static class members.
         * 
         * @throws NumberFormatException
         *             If a property is misformed.
         */
        protected void validateArguments() throws PropertyValidationException {
            boolean ok = true;
            if ( 0 > iterations ) {
                ok = false;
                System.out.println( "iterations property must be >= 0. Using default: " + iterations );
            }
            if ( 0 > priority ) {
                ok = false;
                System.out.println( "priority property must be >= 0. Using default: " + priority );
            }
            if ( 0 > iterations ) {
                ok = false;
                System.out.println( "period property must be >= 0. Using default: " + period );
            }
            if ( 0 > idelay ) {
                ok = false;
                System.out.println( "idelay property must be >= 0. Using default: " + idelay );
            }
            if ( 0 > bgThreads ) {
                ok = false;
                System.out.println( "bgthreads property must be >= 0. Using default: " + bgThreads );
            }
            if ( 0 > bgPriority ) {
                ok = false;
                System.out.println( "bgpriority property must be >= 0. Using default: " + bgPriority );
            }
            if ( 0 > bgPeriod ) {
                ok = false;
                System.out.println( "bgperiod property must be >= 0. Using default: " + bgPeriod );
            }
            if ( 0 > bgIdelay ) {
                ok = false;
                System.out.println( "bgidelay property must be >= 0. Using default: " + bgIdelay );
            }
            if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
        }

        /** retrieved from default or specified properties */
        public static int        iterations;
        public static int        priority;
        public static long       period;
        public static long       idelay;
        public static int        bgPriority;
        public static long       bgPeriod;
        public static long       bgIdelay;
        public static int        bgThreads;
        public static int        numSamples;
        /**
         * the number of histogram bins to use for mode calculation (evenly divided between min and max)
         */
        public static int        histBins;
        /** Thread type to use. */
        public static ThreadType tt;
        /** Memory type to use */
        public static MemoryType mt;
        public static String     loadClass = null;
    }

    private static TestProps testProps;
    private final static int THREAD_SLEEP_TIME = 100;
    //static TimeStampLog      afterLog          = null;
    //static TimeStampLog      beforeLog         = null;
    static long[] afterLog = null;
    static long[] beforeLog = null;
    private static boolean   calculated        = false;

    /**
     * constructor for test class Instantiates testProperties.
     * 
     * @throws Throwable
     * @throws IllegalAccessException
     * @throws InstantiationException
     */
    public ContextSwitch() throws Exception {
        this( new TestProps( ContextSwitch.class ) );
    }

    public ContextSwitch( TestProps props ) throws Exception {
        testProps = props;
        calculated = false;
        System.out.println( testProps );
    }

    /**
     * Start the test
     */
    public void run() {
        try {
           
           
           ThreadLauncher[] bgThreadArray = null;
           if ( 0 < TestProps.bgThreads ) {
               bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
               for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                   //bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( TestProps.tt, TestProps.mt, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                  bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( ThreadType.PERIODICJAVATHREAD, MemoryType.HEAPMEMORY, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                  
                  
                   bgThreadArray[ i ].setIterations( TestProps.iterations );
                   bgThreadArray[ i ].launch();
               }
           }
            // Setup the high priority thread using the ThreadLauncher to
            // manage it
           
            BeforeInit bi = new BeforeInit();
            ThreadLauncherData bdata = new ThreadLauncherData( TestProps.tt, TestProps.mt, bi, TestProps.priority, TestProps.period, TestProps.idelay );
            ThreadLauncher beforeLauncher = ThreadLauncherFactory.getLauncher( bdata );
            beforeLauncher.setIterations(TestProps.iterations);
            System.out.println( beforeLauncher );
            Thread bl = beforeLauncher.launch();
            SharedInfo shared = bi.getSharedInfo();
            // We only need to run each thread once, so we shouldn't
            // need to set the number of iterations
            // Setup the low priority thread using ThreadLaunche
            ThreadLauncherData adata = new ThreadLauncherData( TestProps.tt, TestProps.mt, new AfterInit( shared ), TestProps.priority, TestProps.period, TestProps.idelay );
            ThreadLauncher afterLauncher = ThreadLauncherFactory.getLauncher( adata );
            afterLauncher.setIterations(TestProps.iterations);
            System.out.println( afterLauncher );
            Thread al = afterLauncher.launch();

            // wait for the test to complete
            bl.join();
            al.join();
            // Kill off background threads
            if ( null != bgThreadArray ) {
               for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                   bgThreadArray[ i ].kill();
               }
           }
            
            calculate();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String[] args ) {
        ThreadLauncherFactory.startTest( ContextSwitch.class, TestProps.class );
    }

    static class BeforeThread implements Runnable {
        private SharedInfo   si     = null;
        private int          escape = 0;
        private TimeStampLog log    = null;

        public BeforeThread() {
        }

        public TimeStampLog getLog() {
            return log;
        }

        public void setSharedInfo( SharedInfo si ) {
            this.si = si;
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Runnable#run()
         */
        public void run() {
            try {
                long timeVal;
 
                // Perform measurements for each iteration
                while ( si.currentSample > 0 ) {
                    si.beforeIsReady = false;
                    // This is in case there is an error and we fall
                    // into an endless loop
                    escape++;
                    if ( escape > 100000 ) {
                        System.out.println( "Before thread timed out waiting for after thread..." );
                        break;
                    }
                    // First we need to loop for a bit waiting for
                    // the other thread to be read
                    while ( !si.afterIsReady ) {
                        escape++;
                        if ( escape > 1000000 ) {
                            System.out.println( "Before thread timed out waiting for after thread..." );
                            break;
                        }
                        else if ( escape > 10000000 ) {
                            System.out.println( "Before thread timed out waiting for after thread..." );
                            return;
                        }
                        try {
                            Thread.sleep( 10 );
                        }
                        catch ( InterruptedException e ) {
                            System.out.println( "Before thread unexpectedly interrupted" );
                            return;
                        }
                    }


                    si.beforeIsReady = true;
                    si.afterIsReady = false;

                    // Read timestamp and immediately yield to other thread
                    timeVal = Time.getCycles();
                    Thread.yield();
                    // If the after time was logged, log the
                    // before time
//                    if ( afterLog.getElementAt( (si.currentSample - 1) ) > -1 ) {
//                       
//                       beforeLog.addTimeStamp( (si.currentSample - 1), timeVal );
//                        si.currentSample--;
//                       // Wait a sec for the other thread to catch up
//                       Waster.msWaste(100);
//                    } 
 
                    if ((afterLog[si.currentSample -1]) > -1 ){
                       beforeLog[si.currentSample - 1] = timeVal;
                       si.currentSample--;
                       Waster.msWaste(100);
                    }
                    if ( si.currentSample == 0 ) {
                        si.testComplete = true;

                    }
                }
            }
            catch ( Throwable t ) {
                t.printStackTrace();
            }
        }
    }

    class BeforeInit extends TestInitializer {
        SharedInfo shared = null;

        public void run() {
            // Calibrate waster
            Waster.msWaste(1);
           
            // This is put here for some reason
            shared = new SharedInfo();
            BeforeThread thread = new BeforeThread();
            thread.setSharedInfo( shared );
            setTestRunner( thread );
            shared.totalSamples = TestProps.numSamples;
            shared.currentSample = TestProps.numSamples;
            //			Initialize the logs
            //afterLog = new TimeStampLog( TestProps.numSamples );
            //beforeLog = new TimeStampLog( TestProps.numSamples );
            afterLog = new long[TestProps.numSamples];
            beforeLog = new long[TestProps.numSamples];
        }

        public SharedInfo getSharedInfo() {
            return shared;
        }
    }

    class AfterInit extends TestInitializer {
        SharedInfo si = null;

        public void run() {
            AfterThread thread = new AfterThread();
            thread.setSharedInfo( si );
            setTestRunner( thread );
        }

        /**
         * @param shared
         */
        public AfterInit( SharedInfo shared ) {
            si = shared;
        }
    }

    static class AfterThread implements Runnable {
        // private int iterations = 0;
        private TimeStampLog log = null;
        private SharedInfo   si  = null;

        public AfterThread() {
        }

        public TimeStampLog getLog() {
            return log;
        }

        public void setSharedInfo( SharedInfo si ) {
            this.si = si;
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Runnable#run()
         */
        public void run() {
            try {
                
                long timeVal = 0;
                int escape = 0;
                while ( si.currentSample > 0 ) {
                    //		Just in case
                    escape++;
                    if ( escape > 10000000 ) {
                        System.out.println( "After thread timed out waiting for before thread..." );
                        break;
                    }
                    else if ( escape > 100000000 ) {
                        System.out.println( "After thread timed out waiting for before thread..." );
                        return;
                    }
                  // Set the shared flag which lets the high
                    // priority thread know we're ready to measure
                    si.afterIsReady = true;
                    Thread.yield();

                    // Read time as soon as the other thread yields
                    timeVal = Time.getCycles();

                    // We have to check first if the other thread flag is set
                    // before logging time
                    if ( si.beforeIsReady == true ) {
                        //afterLog.addTimeStamp( (si.currentSample - 1), timeVal );
                        afterLog[si.currentSample - 1] = timeVal; 
                       si.beforeIsReady = false;
                        // Wait a bit for the other thread to get ready
                        Waster.msWaste(100);
                    }
                }
                if ( si.testComplete == true ) return;
            }
            catch ( Throwable t ) {
                t.printStackTrace();
            }
        }
    }

    private static void calculate() {
        // Compare the results between logs
        //long[] results = afterLog.compare( beforeLog, TestProps.numSamples );
        
        //long[] results = StatPack.
       long[] results = new long[TestProps.numSamples];
       
       for (int index = 0; index < results.length; index++) {
          results[index] = Math.abs(afterLog[index] - beforeLog[index]);
       }
       testProps.setOutputProperty( "contextSwitchTimes", results );
        testProps.setOutputProperty( "max", StatPack.max( results ) );
        testProps.setOutputProperty( "min", StatPack.min( results ) );
        testProps.setOutputProperty( "mean", StatPack.mean( results ) );
        testProps.setOutputProperty( "stddev", StatPack.stddev( results ) );
        testProps.setOutputProperty( "variance", StatPack.variance( results ) );

        try {
            testProps.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
        }
    }

    class BGThread implements Runnable {
       int[] garbage;

       public void run() {
           if ( null == garbage ) {
               garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
           }
           else {
               garbage = null;
           }
       }
   }
    
    static class SharedInfo {
        public boolean testComplete     = false;
        public boolean beforeIsReady    = false;
        public boolean afterIsReady     = false;
        public int     currentSample = -1;
        public int     totalSamples  = -1;
    }
}