/*
 * Created on Jun 8, 2004
 *
 * InterruptLatency.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;
import java.util.Iterator;
import javax.realtime.AsynchronouslyInterruptedException;
import javax.realtime.ImmortalMemory;
import javax.realtime.Interruptible;
import javax.realtime.MemoryArea;
import javax.realtime.PeriodicParameters;
import javax.realtime.RealtimeThread;
import javax.realtime.RelativeTime;
import javax.realtime.Timed;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.Range;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * The InterruptLatency class measures the minimum reaction time of the VM to
 * corrective actions and clean-up when an AsynchronousInterruptedException is
 * activated. The InterruptLatency class measures the latency of the activation
 * of the interruptAction() method of an Interruptible when running as the target
 * of an AsychronouslyInterruptedException.doInterruptible() method.
 * <p>
 * <B><I>PURPOSE: </I> </B> Measure the time it takes to activate the interruptAction()
 * method when an AIE.doInterruptible() is interrupted.
 * <p>
 * <B><I>DESIGN: </I> </B> A Timed AIE is created within the context of an RT
 * or NHRT thread and desired memory type. The start of the Timed is recorded
 * and it polls for the current time until it exceeds its deadline and an
 * asynchronous transfer of control (ATC) occurs and the interruptAction() method
 * is activated. The interruptAction start time is recorded as well as the latest
 * polled time from the Timed.run().
 * <p>
 * A variable number of background threads can be run. Each background thread produces a
 * random amount of garbage (an array of integers with a random length). 
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> None.
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * The following parameters may be specified in a properties file whose name is
 * of the form <i>suramadu.tests.latency.InterruptLatency.properties </i>.
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"memtype"</td>
 * <td>"IMMORTAL"</td>
 * <td>Memory type of the thread. Options are HEAP, SCOPE and IMMORTAL</td>
 * </tr>
 * <tr>
 * <td>"threadtype"</td>
 * <td>"NHRT"</td>
 * <td>Type of thread to use. Valid options are RT and NHRT</td>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"100"</td>
 * <td>Number of samples to take</td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"200"</td>
 * <td>Priority of the thread</td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"200"</td>
 * <td>Period of the thread</td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td>Initial delay of the thread</td>
 * </tr>
 * <tr>
 * <td>"bgthreads"</td>
 * <td>"0"</td>
 * <td>Number of background threads</td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"100"</td>
 * <td>Priority of the background threads</td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"100"</td>
 * <td>Period of the background threads</td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td>Maximum value of random initial delay for background threads</td>
 * </tr>
 * <tr>
 * <td>"histbins"</td>
 * <td>"50"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"costmillis"</td>
 * <td>"10"</td>
 * <td>Cost of the thread, in milliseconds, added to <i>costnanos </i></td>
 * </tr>
 * <tr>
 * <td>"costnanos"</td>
 * <td>"0"</td>
 * <td>Cost of the thread, in nanoseconds, added to <i>constmillis </i></td>
 * </tr>
 * <tr>
 * <td>"deadlinemillis"</td>
 * <td>"10"</td>
 * <td>Deadline of the thread, in milliseconds, added to <i>deadlinenanos </i>
 * </td>
 * </tr>
 * <tr>
 * <td>"deadlinenanos"</td>
 * <td>"0"</td>
 * <td>Deadline of the thread, in nanoseconds, added to <i>deadlinemillis </i>
 * </td>
 * </tr>
 * <tr>
 * <td>"waste"</td>
 * <td>"11000000"</td>
 * <td>Amount of time for the thread to waste each period, in nanoseconds</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The following output files will be produced. The
 * file name prefix follows standard Suramadu output file naming conventions.
 * <table>
 * <tr>
 * <th align="left">File suffix</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>latencyFromStartPlusDeadline</td>
 * <td>Calculated latency: start time of handler - (start time of thread + deadline). This
 * gives an idea how precise the interrupt mechanism is in an absolute (theoretical) sense.</td>
 * </tr>
 * <tr>
 * <td>latencyFromStop</td>
 * <td>Direct latency: start time of handler - last polled time of thread. This is a relative
 * latency, indicating how quickly the process happens, but does not provide an absolute time
 * reference (i.e. was the fault detected in a timely fashion?).</td>
 * </tr>
 * <tr>
 * <td>handlerJitter</td>
 * <td>Relative jitter of handler start times</td>
 * </tr>
 * <tr>
 * <td>primaryStartJitter</td>
 * <td>Relative jitter of the thread start time</td>
 * </tr>
 * <tr>
 * <td>primaryStopJitter</td>
 * <td>Relative jitter of the thread's last polled time</td>
 * </tr>
 * <tr>
 * <td>actualCost</td>
 * <td>The difference between the start time of the thread and the last polled time. The
 * expectation is that this should be equal to the value of the deadline property.
 * With the latencyFromStop this should provide information on the overall deterministic
 * nature of the interrupt mechanism.</td>
 * </tr>
 * <tr>
 * <td>SUMMARY</td>
 * <td>Overall latency statistics</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>EXAMPLES: </I> </B>
 * <p>
 * <B><I>ORIGINS: </I> </B>
 * <p>
 * 
 * @author Joseph Hutcherson
 * 
 * @version $Revision: 1.1 $
 */

public class InterruptLatency implements Runnable {
	
	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
				"memtype", "IMMORTAL", 
				"threadtype", "NHRT", 
				"iterations", "100", 
				"priority", "200", 
				"period", "250", 
				"idelay", "750", 
				"bgthreads", "0", 
				"bgpriority", "100", 
				"bgperiod", "100", 
				"bgidelay", "400", 
				"histbins", "50", 
				"costmillis", "10", 
				"costnanos", "0", 
				"deadlinemillis", "10",
				"deadlinenanos", "0",
                "waste", "11000000", // 11 milliseconds, should blow deadline and cost
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps( Class aClass ) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void loadArguments() throws PropertyValidationException {
			threadType = getInputProperty( "threadtype" );
			memType = getInputProperty( "memtype" );
			iterations = getIntInputProperty( "iterations" );
			priority = getIntInputProperty( "priority" );
			period = getIntInputProperty( "period" );
			idelay = getIntInputProperty( "idelay" );
			bgThreads = getIntInputProperty( "bgthreads" );
			bgPriority = getIntInputProperty( "bgpriority" );
			bgPeriod = getIntInputProperty( "bgperiod" );
			bgIdelay = getIntInputProperty( "bgidelay" );
			histBins = getIntInputProperty( "histbins" );
			costMillis = getIntInputProperty( "costmillis" );
			costNanos = getIntInputProperty( "costnanos" );
			deadlineMillis = getIntInputProperty( "deadlinemillis" );
			deadlineNanos = getIntInputProperty( "deadlinenanos" );
            	waste = getIntInputProperty( "waste" );
		}
		
		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void validateArguments() throws PropertyValidationException {
			boolean ok = true;
			
			if ( 0 > iterations ) {
				ok = false;
				System.out.println( "iterations property must be >= 0. Using default: " + iterations );
			}
			if ( 0 > priority ) {
				ok = false;
				System.out.println( "priority property must be >= 0. Using default: " + priority );
			}
			if ( 0 > iterations ) {
				ok = false;
				System.out.println( "period property must be >= 0. Using default: " + period );
			}
			if ( 0 > idelay ) {
				ok = false;
				System.out.println( "idelay property must be >= 0. Using default: " + idelay );
			}
			if ( 0 > bgThreads ) {
				ok = false;
				System.out.println( "bgthreads property must be >= 0. Using default: " + bgThreads );
			}
			if ( 0 > bgPriority ) {
				ok = false;
				System.out.println( "bgpriority property must be >= 0. Using default: " + bgPriority );
			}
			if ( 0 > bgPeriod ) {
				ok = false;
				System.out.println( "bgperiod property must be >= 0. Using default: " + bgPeriod );
			}
			if ( 0 > bgIdelay ) {
				ok = false;
				System.out.println( "bgidelay property must be >= 0. Using default: " + bgIdelay );
			}
			if ( 0 > costMillis ) {
				ok = false;
				System.out.println( "costmillis property must be >= 0. Using default: " + costMillis );
			}
			if ( 0 > costNanos ) {
				ok = false;
				System.out.println( "costnanos property must be >= 0. Using default: " + costNanos );
			}
			if (costNanos == 0 && costMillis == 0) {
				ok = false;
				System.out.println( "costnanos and costmillis cannot both be 0");
			}
			if ( 0 > deadlineMillis ) {
				ok = false;
				System.out.println( "deadlinemillis property must be >= 0. Using default: " + deadlineMillis );
			}
			if ( 0 > deadlineNanos ) {
				ok = false;
				System.out.println( "deadlinenanos property must be >= 0. Using default: " + deadlineNanos );
			}
			if (deadlineNanos == 0 && deadlineMillis == 0) {
				ok = false;
				System.out.println( "deadlinnanos and deadlinemillis cannot both be 0");
			}
            if ( 0 > waste ) {
                ok = false;
                System.out.println( "waste property must be >= 0. Using default: " + waste );
            }
			if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
		}
		
		/**
		 * actual properties used by this test run. Filled in from default and specified properties at class instantiation.
		 */
		/** retrieved from default or specified properties */
		public static String         threadType;
		
		/** retrieved from default or specified properties */
		public static String         memType;
		
		/** retrieved from default or specified properties. */
		public static int            iterations;
		
		/** retrieved from default or specified properties. */
		public static int            priority;
		
		/** retrieved from default or specified properties. */
		public static long           period;
		
		/** retrieved from default or specified properties. */
		public static long           idelay;
		
		/** retrieved from default or specified properties. */
		public static int            bgPriority;
		
		/** retrieved from default or specified properties. */
		public static long           bgPeriod;
		
		/** retrieved from default or specified properties. */
		public static long           bgIdelay;
		
		/** retrieved from default or specified properties. */
		public static int            bgThreads;
		
		/** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
		public static int            histBins;

		/** the cost of the test interval */
		public static int            costMillis;
		public static int            costNanos;
		
		/** the deadline of the test interval */
		public static int            deadlineMillis;
		public static int            deadlineNanos;

        /** the time to waste each interval */
        public static int            waste;
        
	}

	/**
	 * Local data for test: generated or otherwise calculated. Not from test properties.
	 */
	/** Test properties for this run */
    static public TestProps 	  testProps;

    /**
     * Thread type to use.
     * 
     * @uml.property name="tt"
     * @uml.associationEnd 
     * @uml.property name="tt" multiplicity="(0 1)"
     */
    public ThreadType tt;

    /**
     * Memory type to use
     * 
     * @uml.property name="mt"
     * @uml.associationEnd 
     * @uml.property name="mt" multiplicity="(0 1)"
     */
    public MemoryType mt;

    /** Timestamp arrays */
    static public long	primaryStop[];
    static public long	primaryStart[];

    static public long	handlerStart[];

    /** Flag to politely shutdown the background threads */
    boolean keepRunning = true;

    /** iteration counter */
    static public  int           counter = 0;

    /** Handlers and parameters */
    public static RelativeTime rtCost;
    public static RelativeTime rtDeadline;
    public static RelativeTime rtStart;
    public static RelativeTime rtPeriod;
    public static PeriodicParameters params;
        
    /** Keep this declaration fairly high level, and cast to appropriate class when necessary */
    public static Thread primary;
    
	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public InterruptLatency() throws Exception {
        this( new TestProps( InterruptLatency.class ) );
    }

    public InterruptLatency( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
        counter   = 0;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( InterruptLatency.class, TestProps.class );
    }

    public void run() {
        try {
            /**
             * Since we need iterations + 1 number of measurements to get iterations number of timings, increment iterations before
             * using it.
             */
            TestProps.iterations++;
            
            /** ImmortalInit creates event handlers in immortal memory */
            ImmortalInit tmp = new ImmortalInit();
            tmp.start();
            tmp.join();

            ThreadLauncher[] bgThreadArray = null;
            if ( 0 < TestProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    double randomDelay = Math.random() * TestProps.bgIdelay;
                    bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher(
                    		new ThreadLauncherData(
                    				ThreadType.PERIODICJAVATHREAD,
								MemoryType.HEAPMEMORY,
								new BGThread(),
								TestProps.bgPriority,
								TestProps.bgPeriod,
								TestProps.bgIdelay ) );
                    bgThreadArray[ i ].launch();
                }
            }

            /** Can vary thread types used */
            tt = ThreadType.getThreadType( TestProps.threadType );

            /** Can vary memory types used */
            mt = MemoryType.getMemoryType( TestProps.memType );


            // Create the RT thread launcher for the driver thread 
            ThreadLauncher launcher = ThreadLauncherFactory.getLauncher( 
                    new ThreadLauncherData( tt, mt, new TestInit(), TestProps.priority, TestProps.period, TestProps.idelay) );

            // Set the handlers for the driver thread
            //launcher.setRTCost(rtCost);
            //launcher.setRTDeadline(rtDeadline);
            launcher.setIterations( TestProps.iterations );

            int duration = (int) (((float) TestProps.iterations * TestProps.period) / 1000);
            System.out.println("This will take about " + duration + " seconds to run");            
            
            // Note that this is a realtime thread so it can be restarted by the event handlers
            try {
				primary = launcher.launch();
			} catch (Exception e) {
				System.out.println("Exception launching/casting thread: " + e);
				e.printStackTrace();
			}
            
            // Wait for the test to complete. 
            primary.join();

            // Kill off background threads.
            keepRunning = false;
            if ( null != bgThreadArray ) {
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    bgThreadArray[ i ].kill();
                }
            }

            // The deadline, in nanoseconds
            long deadline = (TestProps.deadlineMillis * 1000000) + TestProps.deadlineNanos;
            
            long[] latency = new long[ TestProps.iterations ];
            long[] directLatency = new long[ TestProps.iterations ];
            
            // Get two different latencies. latency[] is the difference between the
            // theoretical stop time of the thread till the handler starts. directLatency[]
            // is the difference between the handler start time and the thread stop time.
            // Direct latency is relative; it gives no indication if the detection of the
            // fault was deterministic.
            for (int i = 0; i < latency.length; i++) {
            		long nsHandler = (long) Time.nsFromCycles(handlerStart[i]);
            		long nsPrimary = (long) Time.nsFromCycles(primaryStart[i]);
            		//System.out.println(nsHandler + "\t" + nsPrimary + "\t" + deadline);
            		latency[i] = nsHandler - (nsPrimary + deadline);
            		directLatency[i] = nsHandler - (long) Time.nsFromCycles(primaryStop[i]);
            }

            long nsPeriod = TestProps.period * 1000000;

            // Difference from when the thread started to when it stopped. It should
            // match the deadline.
            long[] actualCost = new long[TestProps.iterations];
            for (int i = 0; i < actualCost.length; i++) {
            		actualCost[i] = (long) Time.nsFromCycles(primaryStop[i] - primaryStart[i]);
            }

            long rtPeriod = Time.cyclesFromMs( TestProps.period );

            long[] startJitter = StatPack.diff (primaryStart);
            long[] stopJitter = StatPack.diff (primaryStop);
            long[] handlerJitter = StatPack.diff (handlerStart);

            for ( int i = 0; i < startJitter.length ; i++ ) {
                startJitter[i] = (long) Time.nsFromCycles (startJitter[i] - rtPeriod);
                stopJitter[i] = (long) Time.nsFromCycles (stopJitter[i] - rtPeriod);
                handlerJitter[i] = (long) Time.nsFromCycles (handlerJitter[i] - rtPeriod);
            }

            testProps.setOutputProperty( "latencyFromStartPlusDeadline", latency );
            testProps.setOutputProperty( "latencyFromStartPlusDeadlinemax", StatPack.max( latency ) );
            testProps.setOutputProperty( "latencyFromStartPlusDeadlinemin", StatPack.min( latency ) );
            testProps.setOutputProperty( "latencyFromStartPlusDeadlinemean", StatPack.mean( latency ) );
            testProps.setOutputProperty( "latencyFromStartPlusDeadlinestddev", StatPack.stddev( latency ) );
            testProps.setOutputProperty( "latencyFromStartPlusDeadlinevariance", StatPack.variance( latency ) );

            testProps.setOutputProperty( "latencyFromStop", directLatency );
            testProps.setOutputProperty( "latencyFromStopmax", StatPack.max( directLatency ) );
            testProps.setOutputProperty( "latencyFromStopmin", StatPack.min( directLatency ) );
            testProps.setOutputProperty( "latencyFromStopmean", StatPack.mean( directLatency ) );
            testProps.setOutputProperty( "latencyFromStopstddev", StatPack.stddev( directLatency ) );
            testProps.setOutputProperty( "latencyFromStopvariance", StatPack.variance( directLatency ) );

            testProps.setOutputProperty( "actualCost", actualCost );
            testProps.setOutputProperty( "actualCostmax", StatPack.max( actualCost ) );
            testProps.setOutputProperty( "actualCostmin", StatPack.min( actualCost ) );
            testProps.setOutputProperty( "actualCostmean", StatPack.mean( actualCost ) );
            testProps.setOutputProperty( "actualCoststddev", StatPack.stddev( actualCost ) );
            testProps.setOutputProperty( "actualCostvariance", StatPack.variance( actualCost ) );

            testProps.setOutputProperty( "startJitter", startJitter );
            testProps.setOutputProperty( "startJittermax", StatPack.max( startJitter ) );
            testProps.setOutputProperty( "startJittermin", StatPack.min( startJitter ) );
            testProps.setOutputProperty( "startJittermean", StatPack.mean( startJitter ) );
            testProps.setOutputProperty( "startJitterstddev", StatPack.stddev( startJitter ) );
            testProps.setOutputProperty( "startJittervariance", StatPack.variance( startJitter ) );

            testProps.setOutputProperty( "stopJitter", stopJitter );
            testProps.setOutputProperty( "stopJittermax", StatPack.max( stopJitter ) );
            testProps.setOutputProperty( "stopJittermin", StatPack.min( stopJitter ) );
            testProps.setOutputProperty( "stopJittermean", StatPack.mean( stopJitter ) );
            testProps.setOutputProperty( "stopJitterstddev", StatPack.stddev( stopJitter ) );
            testProps.setOutputProperty( "stopJittervariance", StatPack.variance( stopJitter ) );

            testProps.setOutputProperty( "handlerJitter", handlerJitter );
            testProps.setOutputProperty( "handlerJittermax", StatPack.max( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittermin", StatPack.min( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittermean", StatPack.mean( handlerJitter ) );
            testProps.setOutputProperty( "handlerJitterstddev", StatPack.stddev( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittervariance", StatPack.variance( handlerJitter ) );

            Range mode = StatPack.mode( latency, TestProps.histBins );
            testProps.setOutputProperty( "mode.lobound", (Number)mode.getLoBound() );
            testProps.setOutputProperty( "mode.hibound", (Number)mode.getHiBound() );
            testProps.setOutputProperty( "mode.count", mode.getCount() );
            testProps.setOutputProperty( "mode.percent", (double)mode.getCount() * 100.0 / ((double)TestProps.iterations - 1) );

            // report on what we're going to use.
            testProps.close();
            ;
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    class TestInit extends TestInitializer {
    		public TestInit() {
    		}
        public void run() {
        		System.out.println("Now trying " + (TestProps.iterations - 1) + " iterations:");
             setTestRunner(new Test());
        }
    }

    class ImmortalInit extends RealtimeThread {
		public void run() {

			//System.out.println("Creating handlers in immortal memory");
			MemoryArea immortal = ImmortalMemory.instance();

			immortal.enter(new Runnable() {
				public void run() {

					rtCost = new RelativeTime(TestProps.costMillis,
							TestProps.costNanos);
					rtDeadline = new RelativeTime(TestProps.deadlineMillis,
							TestProps.deadlineNanos);
					rtStart = new RelativeTime(TestProps.idelay, 0);
					rtPeriod = new RelativeTime(TestProps.period, 0);

					params = new PeriodicParameters(rtStart, rtPeriod, rtCost,
							rtDeadline, null, null);

					handlerStart = new long[TestProps.iterations];
					primaryStop = new long[TestProps.iterations];
					primaryStart = new long[TestProps.iterations];
					
				}
			});
		}
	}
    
    /**
	 * Primary thread for the test. In each iteration it will start an Interruptible
	 * that will exceeds its deadline, forcing an AsynchronouslyInterruptedException
	 * to be activated. Timestamps are taken when the Interruptible is started, when
	 * it is interrupted, and when the handler is activated.
	 */
    class Test implements Runnable {

		Timed aie;
		Interruptible interruptMe;
		
		public Test() {
			aie = new Timed(rtDeadline);
			interruptMe = new ToBeInterrupted();
		}

		public void run() {
			try {
				
				//System.out.println("Driver thread, counter: " + counter);

				if (aie.doInterruptible(interruptMe)) {
					//System.out.println("Did doInterruptible");
				} else {
					// Means that VM doesn't support AIE or that the waste value is too short.
					System.out.println("Count: " + counter + " Interrupt not triggered");
				}

			} catch (Exception e) {
				System.out.println("Exception in Test class: " + e);
				e.printStackTrace();
			}
			counter++;
			if (counter == TestProps.iterations) {
				System.out.println("Done.");
			}
		}
	}

    /**
     * This class forces an AsynchronouslyInterruptedException to be thrown by
     * exceeding its deadline. It takes starting timestamps, then polls for the
     * current time in a loop in the run() method until the deadline is exceeded
     * and control is passed to the interruptAction() method. It then records
     * the last polled time and the start time of the handler. 
     */
    class ToBeInterrupted implements Interruptible {

    		long currentTime;
    		
    		public void run(AsynchronouslyInterruptedException e)
				throws AsynchronouslyInterruptedException {

			/** Get the current time as the first thing we do */
			long startTime = Time.getCycles();

			/** Record the timestamp for the starting thread */
			primaryStart[counter] = startTime;

			/** Calculate the delay in cycles */
			long delay = Time.cyclesFromNs(TestProps.waste);

			/**
			 * Poll for current time (in cycles) in a tight loop. When we delay
			 * past the deadline an AIE gets thrown.
			 */
			currentTime = Time.getCycles();
			while (currentTime < startTime + delay) {
				currentTime = Time.getCycles();
			}
		}

		/** Invoked when the run() method exceeds its deadline */
		public void interruptAction(AsynchronouslyInterruptedException e) {

			/** Record the start time of this handler */
			handlerStart[counter] = Time.getCycles();

			/** Record the last polled time from run() */
			primaryStop[counter] = currentTime;

			/** Not needed for this test */
			try {
				if (e.happened(true)) {
					/** This is the statement to enable if you want to verify
					 * that yes, indeed the interruptAction is being invoked and
					 * applies to this instance.
					 */
					//System.out.println("Caught interrupt!");
				}
				if (e.isEnabled()) {
					//System.out.println("It is enabled");
				}
			} catch (ClassCastException cce) {

			}
		}

    }
	/**
	 * Simple background class that will generate garbage with gusto.
	 */
	class BGThread implements Runnable {
        int[] garbage;

        public void run() {
            if ( null == garbage ) {
                garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
            }
            else {
                garbage = null;
            }
        }
	}
}
