/*
 * JNIOverhead.java Created on August 5, 2003, 1:02 PM $Id: JNIOverhead.java,v 1.1 2004/09/23 15:34:04 mbi Exp $
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

import java.util.Iterator;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.Range;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * The JNIOverhead class measures the overhead involved in making a simple JNI call.
 * <p>
 * <B><I>PURPOSE: </I> </B> This test is particularly pertinent in calculating the error of high-resolution time measurement. Since
 * it takes time to measure time, the results of this test can be used to correct for this overhead.
 * <p>
 * <B><I>DESIGN: </I> </B> The JNI native method to retrieve CPU machine cycles is called repeatedly, and the value logged in an
 * array. The differences between the first and all subsequent, consecutive calls is calculated, and statistical calculations made
 * from these values. Since each difference represents the time it takes to return from the JNI call and then call it again, the
 * total value can be considered the entire JNI call overhead even though it was measured in the reverse order as is considered
 * normal.
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> No attempt is made in this test to measure specific JNI callback or data copying methodologies.
 * <p>
 * <B><I>INPUTS: </I> </B> <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode statistical
 * measurements on the collected data.
 * <p>
 * <B><I>EXAMPLES: </I> </B> None.
 * <p>
 * <B><I>ORIGINS: </I> </B> This test was conceived and written by JPL and the Golden Gate Team.
 * <p>
 * 
 * @author Mark Indictor
 *         <p>
 * @version $Id: JNIOverhead.java,v 1.1 2004/09/23 15:34:04 mbi Exp $
 */
public class JNIOverhead implements Runnable {
	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
				"iterations", "100000", 
				"histbins",   "1000",
				"priority",   "250"
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             if property is malformed.
	     */
	    protected void loadArguments() throws PropertyValidationException {
	        iterations = getIntInputProperty( "iterations" );
	        histBins = getIntInputProperty( "histbins" );
	        priority = getIntInputProperty( "priority" );
	    }

	    protected void validateArguments() throws PropertyValidationException {
	        boolean ok = true;

	        if ( 0 > iterations ) {
	            ok = false;
	            System.out.println( "iterations property must be >= 0. Using default: " + iterations );
	        }
	        if ( 0 > histBins ) {
	            ok = false;
	            System.out.println( "histBins property must be >= 0. Using default: " + histBins );
	        }
	        if ( 0 > priority ) {
	            ok = false;
	            System.out.println( "priority property must be >= 0. Using default: " + priority );
	        }
	        if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
	    }

	    /** The number of iterations to perform */
	    private static int        iterations;

	    /** retrieved from default or specified properties. */
	    public static int         priority;

	    /** retrieved from default or specified properties */
	    public static String      threadType;

	    /** retrieved from default or specified properties */
	    public static String      memType;

	    /** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
	    public static int         histBins;
	
	}
	
	
    /** the memory areay in which to run */
    public static MemoryType     mt;
    
    /** the thread type to run in */
    public static ThreadType	 tt;

    /** An array in which to store time measurements as they are taken. */
    static long                  timeArray[];

    /**
     * actual properties used by this test run. Filled in from default and specified properties at class instantiation.
     */
    public static TestProps testProps;

	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public JNIOverhead() throws Exception {
        this( new TestProps( JNIOverhead.class ) );
    }

    public JNIOverhead( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( JNIOverhead.class, TestProps.class );
    }

    /**
     * The run() method does the setup and launcing of the actual test code
     */
    public void run() {
        try {
            // TODO: Import from props
            tt = ThreadType.JAVATHREAD;
            mt = MemoryType.HEAPMEMORY;

            /**
             * Begin the test.
             */
            ThreadLauncher launcher  = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( tt, mt, new Test(), TestProps.priority, 100, 100 ) );
            launcher.setIterations( 1 );
            Thread testThread        = launcher.launch();

            /**
             * Wait for the realtime thread to finish.
             */
            try {
                testThread.join();
            }
            catch ( InterruptedException ie ) {
            }

            // log results
            long[] diff = StatPack.diff( timeArray );
            testProps.setOutputProperty( "rawdata", diff );
            testProps.setOutputProperty( "max", StatPack.max( diff ) );
            testProps.setOutputProperty( "min", StatPack.min( diff ) );
            testProps.setOutputProperty( "mean", StatPack.mean( diff ) );
            testProps.setOutputProperty( "stddev", StatPack.stddev( diff ) );

            Range mode = StatPack.mode( diff, TestProps.histBins );
            testProps.setOutputProperty( "mode.lobound", (Number)mode.getLoBound() );
            testProps.setOutputProperty( "mode.hibound", (Number)mode.getHiBound() );
            testProps.setOutputProperty( "mode.count", mode.getCount() );
            testProps.setOutputProperty( "mode.percent", (double)mode.getCount() * 100.0 / ((double)TestProps.iterations - 1) );

            // report on what we're going to use.
            testProps.close();
            ;
        }
        catch ( NumberFormatException nfe ) {
            nfe.printStackTrace();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Static inner class which implements the Runnable interface. This will be the execution path of the actual test code when run
     * in a thread of the desired type.
     * 
     * @see suramadu.util.Launcher
     */
    private static class Test implements Runnable {
        /** The code that does the work. */
        public void run() {
            System.out.println( "." );
            timeArray = new long[ TestProps.iterations ];
            for ( int i = 0; i < TestProps.iterations; i++ ) {
                timeArray[ i ] = Time.getCycles();
            }
        }
    }
}