/*
 * Created on Jun 8, 2004
 *
 * MissedDeadline.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

import java.util.Iterator;
import javax.realtime.AsyncEventHandler;
import javax.realtime.BoundAsyncEventHandler;
import javax.realtime.ImmortalMemory;
import javax.realtime.MemoryArea;
import javax.realtime.MemoryParameters;
import javax.realtime.PeriodicParameters;
import javax.realtime.PriorityParameters;
import javax.realtime.RealtimeThread;
import javax.realtime.RelativeTime;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.RTThreadLauncher;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.Range;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * The MissedDeadline class measures the latency from the missing of a deadline to the invocation
 * of its handler. This latency is calculated by subtracting the start time of the thread plus
 * its deadline from the start time of the handler.
 * <p>
 * <B><I>PURPOSE: </I> </B> Measure the time it takes to get to the handler when a real-time
 * thread's deadline is missed.
 * <p>
 * <B><I>DESIGN: </I> </B> An RT or NHRT thread with desired memory type will record the time
 * at which it is run. It will then sample the current time in a loop for a configurable duration.
 * The VM will at some point detect a missed deadline and activate the handler. The handler will
 * record its start time, and restart the thread for the next iteration. The handler will also
 * record the current loop time of the thread. This time is only valid on single CPU workstations
 * because the thread will continue to execute in the presence of the handler, i.e. it will not
 * be pre-empted.
 * <p>
 * A variable number of background threads can be run. Each background thread produces a
 * random amount of garbage (an array of integers with a random length). 
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> The latency definition is rather simplistic as it does not
 * account for possibilities such as the VM triggering the handler too early but then is slow
 * to activate the handler. This is due to the fact that the thread is not pre-empted (and
 * we could obtain the last timestamp taken) and we have no further insight into the workings
 * of the VM.
 * <p>
 * <B><I>INPUTS: </I> </B> <br> The following parameters may be specified in a properties
 * file whose name is of the form <i>suramadu.tests.latency.MissedDeadline.properties</i>.
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"memtype"</td>
 * <td>"IMMORTAL"</td>
 * <td>Memory type of the thread. Options are HEAP, SCOPE and IMMORTAL</td>
 * </tr>
 * <tr>
 * <td>"threadtype"</td>
 * <td>"NHRT"</td>
 * <td>Type of thread to use. Valid options are RT and NHRT</td>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"100"</td>
 * <td>Number of samples to take</td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"200"</td>
 * <td>Priority of the thread</td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"200"</td>
 * <td>Period of the thread</td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td>Initial delay of the thread</td>
 * </tr>
 * <tr>
 * <td>"bgthreads"</td>
 * <td>"0"</td>
 * <td>Number of background threads</td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"100"</td>
 * <td>Priority of the background threads</td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"100"</td>
 * <td>Period of the background threads</td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td>Maximum value of random initial delay for background threads</td>
 * </tr>
 * <tr>
 * <td>"histbins"</td>
 * <td>"50"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"costmillis"</td>
 * <td>"10"</td>
 * <td>Cost of the thread, in milliseconds, added to <i>costnanos</i></td>
 * </tr>
 * <tr>
 * <td>"costnanos"</td>
 * <td>"0"</td>
 * <td>Cost of the thread, in nanoseconds, added to <i>constmillis</i></td>
 * </tr>
 * <tr>
 * <td>"deadlinemillis"</td>
 * <td>"10"</td>
 * <td>Deadline of the thread, in milliseconds, added to <i>deadlinenanos</i></td>
 * </tr>
 * <tr>
 * <td>"deadlinenanos"</td>
 * <td>"0"</td>
 * <td>Deadline of the thread, in nanoseconds, added to <i>deadlinemillis</i></td>
 * </tr>
 * <tr>
 * <td>"waste"</td>
 * <td>"11000000"</td>
 * <td>Amount of time for the thread to waste each period, in nanoseconds</td>
 * </tr>
 * <tr>
 * <td>"handlerpriority"</td>
 * <td>"250"</td>
 * <td>Priority of the missed deadline handler</td>
 * </tr>
 * <tr>
 * <td>"handlertype"</td>
 * <td>"BOUND"</td>
 * <td>Type of handler, either BOUND or UNBOUND</td>
 * </tr>
 * <tr>
 * <td>"handlernonheap"</td>
 * <td>"TRUE"</td>
 * <td>If the handler should have NHRT properties. Other possible value is 'FALSE'. If
 * <i>handlertype</i> is UNBOUND this value should be FALSE.</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The following output files will be produced. The file name
 * prefix follows standard Suramadu output file naming conventions.
 * <table>
 * <tr>
 * <th align="left">File suffix</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>latency</td>
 * <td>Start time of handler - (start time of thread + deadline)</td>
 * </tr>
 * <tr>
 * <td>handlerJitter</td>
 * <td>Relative jitter of handler start times</td>
 * </tr>
 * <tr>
 * <td>primaryJitter</td>
 * <td>Relative jitter of the thread start time</td>
 * </tr>
 * <tr>
 * <td>preempted</td>
 * <td>Time at which the thread was preempted by the handler. Only valid for single CPU
 * workstations</td>
 * </tr>
 * <tr>
 * <td>SUMMARY</td>
 * <td>Overall latency statistics</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>EXAMPLES: </I> </B>
 * <p>
 * <B><I>ORIGINS: </I> </B>
 * <p>
 * 
 * @author Joseph Hutcherson
 * 
 * @version $Revision: 1.1 $
 */

public class MissedDeadline implements Runnable {
	
	private static final String BOUND = "BOUND";
	private static final String UNBOUND = "UNBOUND";
	
	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
				"memtype", "SCOPE", 
				"threadtype", "NHRT", 
				"iterations", "100", 
				"priority", "200", 
				"period", "250", 
				"idelay", "750", 
				"bgthreads", "0", 
				"bgpriority", "100", 
				"bgperiod", "100", 
				"bgidelay", "400", 
				"histbins", "50", 
				"costmillis", "10", 
				"costnanos", "0", 
				"deadlinemillis", "10",
				"deadlinenanos", "0",
                "waste", "11000000", // 11 milliseconds, should blow deadline and cost
				"handlerpriority", "250",
				"handlertype", BOUND,
				"handlernonheap", "TRUE",
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void loadArguments() throws PropertyValidationException {
			threadType = getInputProperty( "threadtype" );
			memType = getInputProperty( "memtype" );
			iterations = getIntInputProperty( "iterations" );
			priority = getIntInputProperty( "priority" );
			period = getIntInputProperty( "period" );
			idelay = getIntInputProperty( "idelay" );
			bgThreads = getIntInputProperty( "bgthreads" );
			bgPriority = getIntInputProperty( "bgpriority" );
			bgPeriod = getIntInputProperty( "bgperiod" );
			bgIdelay = getIntInputProperty( "bgidelay" );
			histBins = getIntInputProperty( "histbins" );
			costMillis = getIntInputProperty( "costmillis" );
			costNanos = getIntInputProperty( "costnanos" );
			deadlineMillis = getIntInputProperty( "deadlinemillis" );
			deadlineNanos = getIntInputProperty( "deadlinenanos" );
        	waste = getIntInputProperty( "waste" );
        	handlerPriority = getIntInputProperty( "handlerpriority" );
        	handlerType = getInputProperty( "handlertype" );
        	handlerNonHeap = getBooleanInputProperty( "handlernonheap" );
		}
		
		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void validateArguments() throws PropertyValidationException {
			boolean ok = true;
			
			if ( 0 > iterations ) {
				ok = false;
				System.out.println( "iterations property must be >= 0. Using default: " + iterations );
			}
			if ( 0 > priority ) {
				ok = false;
				System.out.println( "priority property must be >= 0. Using default: " + priority );
			}
			if ( 0 > iterations ) {
				ok = false;
				System.out.println( "period property must be >= 0. Using default: " + period );
			}
			if ( 0 > idelay ) {
				ok = false;
				System.out.println( "idelay property must be >= 0. Using default: " + idelay );
			}
			if ( 0 > bgThreads ) {
				ok = false;
				System.out.println( "bgthreads property must be >= 0. Using default: " + bgThreads );
			}
			if ( 0 > bgPriority ) {
				ok = false;
				System.out.println( "bgpriority property must be >= 0. Using default: " + bgPriority );
			}
			if ( 0 > bgPeriod ) {
				ok = false;
				System.out.println( "bgperiod property must be >= 0. Using default: " + bgPeriod );
			}
			if ( 0 > bgIdelay ) {
				ok = false;
				System.out.println( "bgidelay property must be >= 0. Using default: " + bgIdelay );
			}
			if ( 0 > costMillis ) {
				ok = false;
				System.out.println( "costmillis property must be >= 0. Using default: " + costMillis );
			}
			if ( 0 > costNanos ) {
				ok = false;
				System.out.println( "costnanos property must be >= 0. Using default: " + costNanos );
			}
			if (costNanos == 0 && costMillis == 0) {
				ok = false;
				System.out.println( "costnanos and costmillis cannot both be 0");
			}
			if ( 0 > deadlineMillis ) {
				ok = false;
				System.out.println( "deadlinemillis property must be >= 0. Using default: " + deadlineMillis );
			}
			if ( 0 > deadlineNanos ) {
				ok = false;
				System.out.println( "deadlinenanos property must be >= 0. Using default: " + deadlineNanos );
			}
			if (deadlineNanos == 0 && deadlineMillis == 0) {
				ok = false;
				System.out.println( "deadlinnanos and deadlinemillis cannot both be 0");
			}
            if ( 0 > waste ) {
                ok = false;
                System.out.println( "waste property must be >= 0. Using default: " + waste );
            }
            if ( 0 > handlerPriority ) {
                ok = false;
                System.out.println( "handlerPriority property must be >= 0. Using default: " + handlerPriority );
            }
	        if ( !handlerType.equals(BOUND) && !handlerType.equals(UNBOUND) ) {
	            ok = false;
	            System.out.println( "handlertype property must either '" + BOUND + "' or '" +
	            		UNBOUND + "'. Using default: " + handlerType );
	        }
			if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
		}
		
		/**
		 * actual properties used by this test run. Filled in from default and specified properties at class instantiation.
		 */
		/** retrieved from default or specified properties */
		public String         threadType;
		
		/** retrieved from default or specified properties */
		public String         memType;
		
		/** retrieved from default or specified properties. */
		public int            iterations;
		
		/** retrieved from default or specified properties. */
		public int            priority;
		
		/** retrieved from default or specified properties. */
		public long           period;
		
		/** retrieved from default or specified properties. */
		public long           idelay;
		
		/** retrieved from default or specified properties. */
		public int            bgPriority;
		
		/** retrieved from default or specified properties. */
		public long           bgPeriod;
		
		/** retrieved from default or specified properties. */
		public long           bgIdelay;
		
		/** retrieved from default or specified properties. */
		public int            bgThreads;
		
		/** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
		public int            histBins;

		/** the cost of the test interval */
		public int            costMillis;
		public int            costNanos;
		
		/** the deadline of the test interval */
		public int            deadlineMillis;
		public int            deadlineNanos;

        /** the time to waste each interval */
        public int            waste;
        
        /** priority of the handlers - should be greater than the thread's */
        public int			handlerPriority;
        
        /** type of handler */
        public String			handlerType;

        /** nonheap flag to handler */
        public boolean		handlerNonHeap;
	}

	/**
	 * Local data for test: generated or otherwise calculated. Not from test properties.
	 */
	/** Test properties for this run */
    static public TestProps 	  testProps;

    /**
     * Thread type to use.
     * 
     * @uml.property name="tt"
     * @uml.associationEnd 
     * @uml.property name="tt" multiplicity="(0 1)"
     */
    public ThreadType tt;

    /**
     * Memory type to use
     * 
     * @uml.property name="mt"
     * @uml.associationEnd 
     * @uml.property name="mt" multiplicity="(0 1)"
     */
    public MemoryType mt;

    /** Timestamp arrays */
    static public long	primaryStop[];
    static public long	primaryStart[];

    static public long	handlerStart[];
    
    /** Globals used by the primary thread and handler */
    static public long	currentTime = 0;
    static public boolean active = false;
    
    boolean keepRunning = true;

    /** iteration counter */
    static public  int           counter = 0;

    /** Handlers and parameters */
    public static AsyncEventHandler rtMissHandler;
    public static RelativeTime rtCost;
    public static RelativeTime rtDeadline;
    public static RelativeTime rtStart;
    public static RelativeTime rtPeriod;
    public static PeriodicParameters params;
        
    /** Keep this declaration fairly high level, and cast to appropriate class when necessary */
    public static RealtimeThread primary;
    
	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public MissedDeadline() throws Exception {
        this( new TestProps( MissedDeadline.class ) );
    }

    public MissedDeadline( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
        counter   = 0;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( MissedDeadline.class, TestProps.class );
    }

    public void run() {
        try {
            /**
             * Since we need iterations + 1 number of measurements to get iterations number of timings, increment iterations before
             * using it.
             */
            testProps.iterations++;
            
            /** ImmortalInit creates event handlers in immortal memory */
            ImmortalInit tmp = new ImmortalInit();
            tmp.start();
            tmp.join();

            ThreadLauncher[] bgThreadArray = null;
            if ( 0 < testProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ testProps.bgThreads ];
                for ( int i = 0; i < testProps.bgThreads; i++ ) {
                    double randomDelay = Math.random() * testProps.bgIdelay;
                    bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher(
                    		new ThreadLauncherData(
                    				ThreadType.PERIODICJAVATHREAD,
								MemoryType.HEAPMEMORY,
								new BGThread(),
								testProps.bgPriority,
								testProps.bgPeriod,
								testProps.bgIdelay ) );
                    bgThreadArray[ i ].launch();
                }
            }

            /** Can vary thread types used */
            tt = ThreadType.getThreadType( testProps.threadType );

            /** Can vary memory types used */
            mt = MemoryType.getMemoryType( testProps.memType );


            // Create the RT thread launcher for the driver thread 
            RTThreadLauncher launcher = (RTThreadLauncher) ThreadLauncherFactory.getLauncher( 
                    new ThreadLauncherData( tt, mt, new TestInit(), testProps.priority, testProps.period, testProps.idelay) );

            // Set the handlers for the driver thread
            launcher.setMissHandler(rtMissHandler);
            launcher.setRTCost(rtCost);
            launcher.setRTDeadline(rtDeadline);
            launcher.setIterations( testProps.iterations );

            int duration = (int) (((float) testProps.iterations * testProps.period) / 1000);
            System.out.println("This will take about " + duration + " seconds to run");            
            
            // Note that this is a realtime thread so it can be restarted by the event handlers
            try {
				primary = (RealtimeThread) launcher.launch();
			} catch (Exception e) {
				System.out.println("Exception launching/casting thread: " + e);
				e.printStackTrace();
			}
            
            // Wait for the test to complete. 
            primary.join();

            // Kill off background threads.
            keepRunning = false;
            if ( null != bgThreadArray ) {
                for ( int i = 0; i < testProps.bgThreads; i++ ) {
                    bgThreadArray[ i ].kill();
                }
            }

            // The deadline, in nanoseconds
            long deadline = (testProps.deadlineMillis * 1000000) + testProps.deadlineNanos;
            
            long[] latency = new long[ testProps.iterations ];
            
            // Define latency as: difference between the handler start and the
            // time at which the primary thread started plus the deadline (when it
            // should have finished)
            for (int i = 0; i < latency.length; i++) {
            		long nsHandler = (long) Time.nsFromCycles(handlerStart[i]);
            		long nsPrimary = (long) Time.nsFromCycles(primaryStart[i]);
            		//System.out.println(nsHandler + "\t" + nsPrimary + "\t" + deadline);
            		latency[i] = nsHandler - (nsPrimary + deadline);
            }

            long nsPeriod = testProps.period * 1000000;

            // If on a single CPU machine, this value should be how long the
            // primary thread ran until it was pre-empted by the handler. On
            // multi-CPU workstations this value is bogus.
            long[] preempted = new long[testProps.iterations];
            for (int i = 0; i < preempted.length; i++) {
            		preempted[i] = (long) Time.nsFromCycles(primaryStop[i] - primaryStart[i]);
            }

            long rtPeriod = Time.cyclesFromMs( testProps.period );

            long[] primaryJitter = StatPack.diff (primaryStart);
            long[] handlerJitter = StatPack.diff (handlerStart);

            for ( int i = 0; i < primaryJitter.length ; i++ ) {
                primaryJitter[i] = (long) Time.nsFromCycles (primaryJitter[i] - rtPeriod);
                handlerJitter[i] = (long) Time.nsFromCycles (handlerJitter[i] - rtPeriod);
            }

            testProps.setOutputProperty( "latency", latency );
            testProps.setOutputProperty( "latencymax", StatPack.max( latency ) );
            testProps.setOutputProperty( "latencymin", StatPack.min( latency ) );
            testProps.setOutputProperty( "latencymean", StatPack.mean( latency ) );
            testProps.setOutputProperty( "latencystddev", StatPack.stddev( latency ) );
            testProps.setOutputProperty( "latencyvariance", StatPack.variance( latency ) );

            testProps.setOutputProperty( "preempted", preempted );
            testProps.setOutputProperty( "preemptedmax", StatPack.max( preempted ) );
            testProps.setOutputProperty( "preemptedmin", StatPack.min( preempted ) );
            testProps.setOutputProperty( "preemptedmean", StatPack.mean( preempted ) );
            testProps.setOutputProperty( "preemptedstddev", StatPack.stddev( preempted ) );
            testProps.setOutputProperty( "preemptedvariance", StatPack.variance( preempted ) );

            testProps.setOutputProperty( "primaryJitter", primaryJitter );
            testProps.setOutputProperty( "primaryJittermax", StatPack.max( primaryJitter ) );
            testProps.setOutputProperty( "primaryJittermin", StatPack.min( primaryJitter ) );
            testProps.setOutputProperty( "primaryJittermean", StatPack.mean( primaryJitter ) );
            testProps.setOutputProperty( "primaryJitterstddev", StatPack.stddev( primaryJitter ) );
            testProps.setOutputProperty( "primaryJittervariance", StatPack.variance( primaryJitter ) );

            testProps.setOutputProperty( "handlerJitter", handlerJitter );
            testProps.setOutputProperty( "handlerJittermax", StatPack.max( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittermin", StatPack.min( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittermean", StatPack.mean( handlerJitter ) );
            testProps.setOutputProperty( "handlerJitterstddev", StatPack.stddev( handlerJitter ) );
            testProps.setOutputProperty( "handlerJittervariance", StatPack.variance( handlerJitter ) );

            Range mode = StatPack.mode( latency, testProps.histBins );
            testProps.setOutputProperty( "mode.lobound", (Number)mode.getLoBound() );
            testProps.setOutputProperty( "mode.hibound", (Number)mode.getHiBound() );
            testProps.setOutputProperty( "mode.count", mode.getCount() );
            testProps.setOutputProperty( "mode.percent", (double)mode.getCount() * 100.0 / ((double)testProps.iterations - 1) );

            // report on what we're going to use.
            testProps.close();
            ;
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    class TestInit extends TestInitializer {
        public void run() {
        		//System.out.println("TestInit run method");
             
             /** Need access to this thread from the event handler */
             setTestRunner(new Test(testProps.waste));
        }
    }

    class ImmortalInit extends RealtimeThread {
		public void run() {

			//System.out.println("Creating handlers in immortal memory");
			MemoryArea immortal = ImmortalMemory.instance();

			immortal.enter(new Runnable() {
				public void run() {

					if (testProps.handlerType.equals(BOUND)) {
						
						/** The bound handler will inherit the properties of this thread.
						 * Note that the priority of the handler is set directly.
						 */
						/** The bound handler will inherit the properties of this thread.
						 * Note that the priority of the handler is set directly.
						 */
						MemoryArea memory = null;
						if (testProps.handlerNonHeap) {
							memory = ImmortalMemory.instance();
						}
						rtMissHandler = new BoundDeadlineHandler(testProps.handlerPriority,
								testProps.handlerNonHeap, memory);
						//rtOverrunHandler = new CostOverrunHandler(
						//	testProps.handlerPriority);

					
					} else if (testProps.handlerType.equals(UNBOUND)) {
					
						rtMissHandler = new UnboundDeadlineHandler(testProps.handlerPriority,
								testProps.handlerNonHeap);
						//rtOverrunHandler = new CostOverrunHandler(
						//		testProps.handlerPriority);
					
					} else {
						System.out.println("Unknown handler type: "
								+ testProps.handlerType);
					}
					rtCost = new RelativeTime(testProps.costMillis,
							testProps.costNanos);
					rtDeadline = new RelativeTime(testProps.deadlineMillis,
							testProps.deadlineNanos);
					rtStart = new RelativeTime(testProps.idelay, 0);
					rtPeriod = new RelativeTime(testProps.period, 0);
					
					params = new PeriodicParameters(rtStart, rtPeriod, rtCost,
							rtDeadline, null, rtMissHandler);
					
					handlerStart = new long[testProps.iterations];					
					primaryStop = new long[testProps.iterations];
					primaryStart = new long[testProps.iterations];
				}
			});
		}
	}
    
    /**
     * The primary driver for the test. Takes timestamp when started, then proceeds to
     * waste time until a missed deadline or cost overrun occurs. Records current time
     * in the wasting loop, in the hope that the thread will be blocked while the handler
     * runs. Note that blocking does not occur on multiple CPU machines.
     * 
     * Globals are used to record the times of interest due to the use of the RTThreadLauncher.
     * The thread created and returned by RTThreadLauncher is an RTJavaThread type and
     * cannot be cast to any derived type. This is due to the fact that this class is
     * passed as the runnable logic to the thread and is not used as the thread itself.
     * Because of this extended getter/setter methods for timestamps are not accessible.
     * 
     * Also note this class extends RealtimeThread, not Runnable. This allows adjusting
     * of schedule and release parameters, plus use of interruptibles. 
     */
    class Test extends RealtimeThread {
		private int waste = 0;
		
		public Test(int waste) {
			this.waste = waste;
			// This should be getting done through the launcher, but doesn't appear to
			this.setReleaseParameters(params);
			this.setSchedulingParameters(new PriorityParameters(testProps.priority));
		}

		public void run() {

			/**
			 * NOTE: The VM starts (in theory) the deadline timer when run() is
			 * invoked. Take the start time immediately, otherwise the results
			 * will be inconsistent with the VM.
			 */
			long startTime = Time.getCycles();
			
			// debug
			//System.out.println("Test.run counter: " + counter);

			/*
			  System.out.println("Test.run Priority: " +
			  	((PriorityParameters)this.getSchedulingParameters()).getPriority() );
			  System.out.println("Test.run Cost: " +
			  	((ReleaseParameters)this.getReleaseParameters()).getCost() );
			  System.out.println("Test.run Deadline: " +
			  	((ReleaseParameters)this.getReleaseParameters()).getDeadline() );
			*/
			
			primaryStart[counter] = startTime;

			// Calculate the delay in cycles
			long delay = Time.cyclesFromNs(waste);

			// Get the current time
			currentTime = Time.getCycles();

			// Set the active flag after currentTime is initialized
			active = true;

			/**
			 * Poll for the most current time in a tight loop until we've
			 * exceeded the delay. Note that this thread WILL CONTINUE TO RUN
			 * even after a missed deadline or cost overrun. It may or may not
			 * be blocked while the handler is running, depending on the number
			 * of CPUs. This timestamp will be captured by the handler, and will
			 * not be valid for any analysis on multi-CPU workstations.
			 */
			while (currentTime < startTime + delay) {
				//System.out.print(period + ".");
				currentTime = Time.getCycles();
			}
		}
	}

    /**
	 * Simple background class that will generate garbage.
	 */
    class BGThread implements Runnable {
        int[] garbage;

        public void run() {
            if ( null == garbage ) {
                garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
            }
            else {
                garbage = null;
            }
        }
    }
}

/**
 * Bound missed deadline handler. Records the time at which the handler was invoked
 * and increments the counter by which the primary thread start times are recorded.
 * Manages a flag for the primary thread to allow correct pairings of start times.
 */
class BoundDeadlineHandler extends BoundAsyncEventHandler {

	/**
	 * Creates a new handler
	 * @param priority at which this handler should execute
	 */
	public BoundDeadlineHandler(int priority, boolean nonheap, MemoryArea memory) {
		/** Note that this construct actually requires a seventh parameter, a Runnable.
		 * This is not in the RTSJ book I'm using, so it'll get null for now.
		 */
		super(new PriorityParameters(priority), null,
				new MemoryParameters(-1,-1), memory, null, nonheap, null);
		this.setSchedulingParameters(new PriorityParameters(priority));
	}
	
	/**
	 * Records time at which this handler was executed.
	 * 
	 * @see javax.realtime.AsyncEventHandler#handleAsyncEvent()
	 */
	public void handleAsyncEvent() {
		
		/** Get the time at which this handler was activated */
		long myStart = Time.getCycles();
		
		// debug
		//System.out.println("BoundDeadline.handleAsyncEvent triggered. Counter: " +
		//		MissedDeadline.counter);

		/** Clear out pending count to prevent any spurious firings of this handler */
		int fireCount = getAndClearPendingFireCount();

		/**
		 * Check if the primary thread has set the flag indicating it is active
		 * to prevent spurious results. When set, this flag indicates the primary
		 * thread has recorded a start time at the current index. The handler
		 * records its start time at the same index, then resets this flag and
		 * increments the counter for the next iteration.
		 */
		if (MissedDeadline.active) {
			MissedDeadline.handlerStart[MissedDeadline.counter] = myStart;
			
			/**
			 * Note that this time is worthless on multi-CPU workstations because
			 * the primary thread continues to run and update the timestamp.
			 */
			MissedDeadline.primaryStop[MissedDeadline.counter] = MissedDeadline.currentTime;
			
			MissedDeadline.counter++;
			MissedDeadline.active = false;
		}
		
		/**
		 * Restart the primary thread. 
		 */
		MissedDeadline.primary.schedulePeriodic();

	}
}

/**
 * Unbound missed deadline handler. Identical to BoundDeadlineHandler, except that
 * it is derived from AsyncEventHandler.
 */
class UnboundDeadlineHandler extends AsyncEventHandler {

	/**
	 * Creates a new handler
	 * @param priority at which this handler should execute
	 */
	public UnboundDeadlineHandler(int priority, boolean nonheap) {
		super(nonheap);
		this.setSchedulingParameters(new PriorityParameters(priority));
	}
	
	/**
	 * Records time at which this handler was executed.
	 * 
	 * @see javax.realtime.AsyncEventHandler#handleAsyncEvent()
	 */
	public void handleAsyncEvent() {
		
		/** Get the time at which this handler was activated */
		long myStart = Time.getCycles();
		
		// debug
		//System.out.println("UnboundDeadline.handleAsyncEvent triggered. Counter: " +
		//		MissedDeadline.counter);

		/** Clear out pending count to prevent any spurious firings of this handler */
		int fireCount = getAndClearPendingFireCount();

		/**
		 * Check if the primary thread has set the flag indicating it is active
		 * to prevent spurious results. When set, this flag indicates the primary
		 * thread has recorded a start time at the current index. The handler
		 * records its start time at the same index, then resets this flag and
		 * increments the counter for the next iteration.
		 */
		if (MissedDeadline.active) {
			MissedDeadline.handlerStart[MissedDeadline.counter] = myStart;
			
			/**
			 * Note that this time is worthless on multi-CPU workstations because
			 * the primary thread continues to run and update the timestamp.
			 */
			MissedDeadline.primaryStop[MissedDeadline.counter] = MissedDeadline.currentTime;
			
			MissedDeadline.counter++;
			MissedDeadline.active = false;
		}
		
		/**
		 * Restart the primary thread. 
		 */
		MissedDeadline.primary.schedulePeriodic();

	}
}
