/*
 * Created on 

 *  PriorityInheritance.java Created on Jun 8, 2004
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

/**
 * 
 * <p>
 * <B><I>PURPOSE: </I> </B> This test measures the latency of a context switch that results from a thread undergoing
 * priority inheritance. It IS specific to the RTSJ specification and may not be run in regular Java.
 * <p>
 * <B><I>DESIGN: </I> </B> Execute three test threads of high, medium and low priorities; also execute a # of
 * background threads according to NOTE below.  All threads run for a single iteration, taking a number
 * of samples over iteration
 * <p>
 * PI Latency Test runs in this order:
 * <p>
 * Start the low priority thread; this enters a common resource through a synchronized method and waits in a loop
 * for the high priority thread to call the same locked method. 
 * <p>
 * Start the medium priority thread which wastes time for the duration of the sample and puts the low
 * pri thread to sleep.
 * <p>
 * Start the high pri thread; this takes a timestamp (HighBeforeTime) calls the common locked method.  
 * When the  method is called, Priority Inheritance should boost the low priority thread's priority up so that it
 * may continue; as soon as it resumes it takes a timestamp (the LowBoostTime).
 * <p>
 * Measure the difference between the HighBeforeTime and the LowBoostTime; this is the Priority Inheritance
 * Latency. 
 * <p>
 * <B><I>NOTE: </I> <B> Results of this test may not be entirely (TODO:) valid (?) while running on a multi-processor
 * machine, as the low priority thread may have an opportunity to run and exit the locked method without priority 
 * inheritance actually occuring.  To counter this, a number of highest-priority background threads need to be
 * run to (TODO:) "clog" (?) up all processors past the first; e.g. a dual-processor machine will require a single
 * highest-priority background thread running for (TODO:) valid test results.
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> This test uses RTSJ specified RealTimeThreads and NoHeapRealtimeThreads and requires 
 * an RTSJ implementation in order to run.
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"highPriority"</td>
 * <td>"250"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"medPriority"</td>
 * <td>"250"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"lowPriority"</td>
 * <td>"250"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"200"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgthreads"</td>
 * <td>"1"</td>
 * <td></td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode statistical
 * measurements on the collected data.
 * <p>
 * <B><I>EXAMPLES: </I> </B> None.
 * <p>
 * <B><I>ORIGINS: </I> </B> This test is based upon ideas from the AFRL and the rtpresto suites.
 * <p>
 * 
 * @see suramadu.tests.jitter.JavaJitterTest
 *      <p>
 * @author Mark Indictor, Dan Allard
 */

import java.io.IOException;
import java.util.Iterator;

import javax.realtime.PriorityParameters;
import javax.realtime.RealtimeThread;

import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;
import suramadu.util.time.Waster;

public class PriorityInheritance implements Runnable {

public static class TestProps extends TestProperties{
	protected final static String[] DEFAULT_PROPERTIES = {
			"memtype",    "IMMORTAL", 
			"threadtype", "NHRT", 
			"iterations", "1", 
			"numSamples", "10",
			"highPriority",   "300", 
			"medPriority", "200",
			"lowPriority", "100",
			"period",     "1000", 
			"idelay",     "750", 
			"bgthreads",  "1", 
			"bgpriority", "400", 
			"bgperiod",   "1500", 
			"bgidelay",   "0"
		
		};
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

		/**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void loadArguments() throws PropertyValidationException {
	        threadType = getInputProperty( "threadtype" );
	        memType = getInputProperty( "memtype" );
	        iterations = getIntInputProperty( "iterations" );
	        numSamples = getIntInputProperty( "samples");
	        highPriority = getIntInputProperty( "highPriority" );
	        medPriority = getIntInputProperty( "medPriority");
	        lowPriority = getIntInputProperty( "lowPriority");
	        period = getIntInputProperty( "period" );
	        idelay = getIntInputProperty( "idelay" );
	        bgThreads = getIntInputProperty( "bgthreads" );
	        bgPriority = getIntInputProperty( "bgpriority" );
	        bgPeriod = getIntInputProperty( "bgperiod" );
	        bgIdelay = getIntInputProperty( "bgidelay" );
	    }

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void validateArguments() throws PropertyValidationException {
	        boolean ok = true;

	        if ( 0 > highPriority ) {
	            ok = false;
	            System.out.println( "high priority property must be >= 0. Using default: " + highPriority );
	        }
	        if ( 0 > medPriority ) {
	        	ok = false;
	        	System.out.println( "med priority property must be >= 0. Using default: " + medPriority)	;
	        }
	        if ( 0 > lowPriority ) {
	        	ok = false;
	        	System.out.println( "low priority property must be >= 0. Using default: " + medPriority)	;
	        }
	        if ( 0 > iterations ) {
	            ok = false;
	            System.out.println( "iterations property must be >= 0. Using default: " + period );
	        }
	        if ( 0 > numSamples ) {
	            ok = false;
	            System.out.println( "samples property must be >= 0. Using default: " + period );
	        }
	        if ( 0 > idelay ) {
	            ok = false;
	            System.out.println( "idelay property must be >= 0. Using default: " + idelay );
	        }
	        if ( 0 > bgThreads ) {
	            ok = false;
	            System.out.println( "bgthreads property must be >= 0. Using default: " + bgThreads );
	        }
	        if ( 0 > bgPriority ) {
	            ok = false;
	            System.out.println( "bgpriority property must be >= 0. Using default: " + bgPriority );
	        }
	        if ( 0 > bgPeriod ) {
	            ok = false;
	            System.out.println( "bgperiod property must be >= 0. Using default: " + bgPeriod );
	        }
	        if ( 0 > bgIdelay ) {
	            ok = false;
	            System.out.println( "bgidelay property must be >= 0. Using default: " + bgIdelay );
	        }
	        if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
	    }

		
		/** retrieved from default or specified properties */
	    public static String         threadType;
	
	    /** retrieved from default or specified properties */
	    public static String         memType;
	
	    /** retrieved from default or specified properties. */
	    public static int            iterations;
	    public static int 		     numSamples;
	    
	    /** retrieved from default or specified properties. */
	    public static int            highPriority;
	    public static int			 medPriority;
	    public static int            lowPriority;	
	
	    /** retrieved from default or specified properties. */
	    public static long           period;
	
	    /** retrieved from default or specified properties. */
	    public static long           idelay;
	
	    /** retrieved from default or specified properties. */
	    public static int            bgPriority;
	
	    /** retrieved from default or specified properties. */
	    public static long           bgPeriod;
	
	    /** retrieved from default or specified properties. */
	    public static long           bgIdelay;
	
	    /** retrieved from default or specified properties. */
	    public static int            bgThreads;

	}



	/**
	 * @author dan
	 * 
	 * Measures the latency of a context switch that results from a thread undergoing priority inheritance.
    * 
	 * <B><I>INPUTS: </I> </B> <br>
	 * <table>
	 * <tr>
	 * <th align="left">Property</th>
	 * <th align="right">Default Value</th>
	 * <th align="left">Description</th>
	 * </tr>
	 * <tr>
	 * <td>"iterations"</td>
	 * <td>"100"</td>
	 * <td></td>
	 * </tr>
	 * </table>
	 * 
	 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard
	 * deviation and mode statistical measurements on the collected data.
	 * <p>
	 * 
	 * <p>
	 * <B><I>ORIGINS: </I> </B> This test is based upon ideas from the AFRL and the rtpresto suites.
	 * <p>
	 * 
	 * Notes: Uses the Suramadu time native methods to perform timestamping in lieu of RTSJ Time classes
	 */

	// Thread priorities
	static final int HIGH_PRIORITY_THREAD = 1;
	static final int MEDIUM_PRIORITY_THREAD = 2;
	static final int LOW_PRIORITY_THREAD = 3;

	private static TestProps testProps;

	/** Thread type to use. */
	public static ThreadType tt;
	public static MemoryType mt;

	// Some static shared information

	static long[] highBeforeLog = null;
	static long[] highEntryLog = null;
	static long[] mediumEndLog = null;
	static long[] lowBoostLog = null;
	static long[] lowLeavingLog = null;
	static long[] lowBeforeLog = null;
	static long[] lowAfterLog = null;

	private static boolean calculated = false;

/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
	
    public PriorityInheritance() throws Exception {
        this( new TestProps( PriorityInheritance.class ) );
    }

    public PriorityInheritance( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
    }

	/**
	 * Start the test
	 *  
	 */
	public void run() {
		try {
			/** Can vary thread types used */
            tt = ThreadType.getThreadType( TestProps.threadType );

            /** Can vary memory types used */
            mt = MemoryType.getMemoryType( TestProps.memType );
            
			// Setup the high priority thread using the ThreadLauncher to
			// manage it
			HighPriorityInit hpi = new HighPriorityInit();

			ThreadLauncherData hdata =
				new ThreadLauncherData(tt, mt, hpi, TestProps.highPriority,
						TestProps.period, TestProps.idelay);
			
			// Iterations is normally 1 for this test
			hdata.setIterations(TestProps.iterations);

			ThreadLauncher highPriorityLauncher =
				ThreadLauncherFactory.getLauncher(hdata);
			//
			Thread highPriThread = highPriorityLauncher.launch();

			SharedInfo shared = hpi.getSharedInfo();
			
			LockedResource lockedResource = hpi.getLockedResource();

			MediumPriorityInit mpi = new MediumPriorityInit();
			mpi.setSharedInfo(shared);

			// Setup the low priority thread using ThreadLauncher
			ThreadLauncherData mdata =
				new ThreadLauncherData(tt, mt, mpi, 
                                       TestProps.medPriority, 
                                       TestProps.period,
									   TestProps.idelay);
			mdata.setIterations(TestProps.iterations);

			ThreadLauncher mediumPriorityLauncher =
				ThreadLauncherFactory.getLauncher(mdata);


			// Setup the low priority thread using ThreadLauncher

			LowPriorityInit lpi = new LowPriorityInit();

			ThreadLauncherData ldata =
				new ThreadLauncherData(tt, mt, lpi, 
                                       TestProps.lowPriority,
									   TestProps.period,
									   TestProps.idelay);

			ldata.setIterations(TestProps.iterations);

			ThreadLauncher lowPriorityLauncher =
				ThreadLauncherFactory.getLauncher(ldata);

			lpi.setSharedInfo(shared);
			lpi.setLockedResource(lockedResource);

            // Create background threads as specified in test props
			// These are higher priority threads intended to eat
			// up a cpu
			ThreadLauncher[] bgThreadArray = null;
            if ( 0 < TestProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                   // bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( tt, mt, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                   bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( ThreadType.PERIODICJAVATHREAD, MemoryType.HEAPMEMORY, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                   
                    bgThreadArray[ i ].launch();
                }
            }
			
			Thread medPriThread = mediumPriorityLauncher.launch();
			Thread lowPriThread = lowPriorityLauncher.launch();

			// wait for the test to complete
			highPriThread.join();
			medPriThread.join();
			lowPriThread.join();
			shared.testComplete = true;
			
//			 Kill off background threads.
            if ( null != bgThreadArray ) {
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    bgThreadArray[ i ].kill();
                }
            }
				
			calculate();
			// Start the lower priority thread
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public void calculate() throws IOException {

		long intoLockArray[] = new long[TestProps.numSamples];
		long lowBoostArray[] = new long[TestProps.numSamples];
		long lowToHighArray[] = new long[TestProps.numSamples];
		
		for (int index = 0; index < TestProps.numSamples ; index++) {
		   
			long highBefore = highBeforeLog[index];
			//long highEntry = highEntryLog.getElementAt(index);
			//long mediumEnd = mediumEndLog.getElementAt(index);
			long lowBoost = lowBoostLog[index];
			//long lowLeaving = lowLeavingLog.getElementAt(index);
			//long lowAfter = lowAfterLog.getElementAt(index);

			// How long for the high pri thread to get into
			// the lock
			//intoLockArray[index] = Math.abs(highEntry - highBefore);
			
			
			// How long for the high pri thread to boost low pri
			// thread
			lowBoostArray[index] = Math.abs(highBefore - lowBoost);
			
			// How long between low thread end and high thread
			// get into lock
			//lowToHighArray[index] = Math.abs(lowLeaving - highEntry);

		}
/*
		testProps.setOutputProperty("highEnterLockTime", intoLockArray);
		testProps.setOutputProperty("max", StatPack.max(intoLockArray));
		testProps.setOutputProperty("min", StatPack.min(intoLockArray));
		testProps.setOutputProperty("mean", StatPack.mean(intoLockArray));
		testProps.setOutputProperty("stddev", StatPack.stddev(intoLockArray));
		testProps.setOutputProperty("variance", StatPack.variance(intoLockArray));
*/
		testProps.setOutputProperty("highToLowBoostTime", lowBoostArray);
		testProps.setOutputProperty("max", StatPack.max(lowBoostArray));
		testProps.setOutputProperty("min", StatPack.min(lowBoostArray));
		testProps.setOutputProperty("mean", StatPack.mean(lowBoostArray));
		testProps.setOutputProperty("stddev", StatPack.stddev(lowBoostArray));
		testProps.setOutputProperty("variance", StatPack.variance(lowBoostArray));
         
		testProps.close();
		/*
		testProps.setOutputProperty("lowBackToHighTime", lowToHighArray);
		testProps.setOutputProperty("max", StatPack.max(lowToHighArray));
		testProps.setOutputProperty("min", StatPack.min(lowToHighArray));
		testProps.setOutputProperty("mean", StatPack.mean(lowToHighArray));
		testProps.setOutputProperty("stddev", StatPack.stddev(lowToHighArray));
		testProps.setOutputProperty("variance", StatPack.variance(lowToHighArray));
*/
	}

	class HighPriorityInit extends TestInitializer {
		SharedInfo shared = null;
		LockedResource pilr = null;

		public void run() {
		   Waster.msWaste(1);
			shared = new SharedInfo();

			// Created the locked resource
			pilr = new LockedResource();

			// Create the high priority runnable
			HighPriorityThread hpt = new HighPriorityThread();
			hpt.setSharedInfo(shared);
			hpt.setLockedResource(pilr);

			setTestRunner(hpt);

			// Create logs
			highBeforeLog = new long[TestProps.numSamples];
			highEntryLog = new long[TestProps.numSamples];
			mediumEndLog = new long[TestProps.numSamples];
			lowBoostLog = new long[TestProps.numSamples];
			lowLeavingLog = new long[TestProps.numSamples];
			lowBeforeLog = new long[TestProps.numSamples];
			lowAfterLog = new long[TestProps.numSamples];

		}

		public LockedResource getLockedResource() {
			return pilr;
		}
		public SharedInfo getSharedInfo() {
			return shared;
		}
	}

	class MediumPriorityInit extends TestInitializer {
		SharedInfo info = null;

		public void setSharedInfo(SharedInfo si) {
			info = si;
		}

		public void run() {
			MediumPriorityThread mpt = new MediumPriorityThread();
			mpt.setSharedInfo(info);
			setTestRunner(mpt);
		}
	}
	class LowPriorityInit extends TestInitializer {
		SharedInfo info = null;
		LockedResource lockedResource = null;

		public void setSharedInfo(SharedInfo si) {
			info = si;
		}
		/**
		 * @param lockedResource
		 */
		public void setLockedResource(LockedResource lockedResource) {
			this.lockedResource = lockedResource;

		}
		public void run() {

			LowPriorityThread lpt = new LowPriorityThread();
			lpt.setLockedResource(lockedResource);
			lpt.setSharedInfo(info);
			setTestRunner(lpt);

		}
	}

	class SharedInfo {
		public boolean testComplete = false;

		public boolean highReady = false;

		public boolean mediumReady = false;

		public boolean lowReady = false;

		public boolean highStarted = false;
		public int currentIndex = 0;

	}

	class HighPriorityThread implements Runnable {
		SharedInfo shared = null;
		LockedResource pilr = null;

		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			for (int index = 0; index < TestProps.numSamples; index++) {
				shared.highReady = true;
				
				// Wait until all threads are ready
				while (!(shared.lowReady
					&& shared.mediumReady
					&& shared.highReady)) {
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						System.out.println("High priority thread interrupted");
					}
				}

				// This thread needs to start last
				try {
					Thread.sleep(600);
				} catch (InterruptedException ie) {
					ie.printStackTrace();
				}
				shared.highStarted = true;
				shared.highReady = false;
				shared.currentIndex = index;
                //System.out.println("Hi pri start sample " + index);
				// The locked resource is waiting for this flag to reset
				

				long highBefore = Time.getCycles();
			//System.out.println("About to lock...");	
				pilr.lock(HIGH_PRIORITY_THREAD, shared);
		    //System.out.println("After lock...");
				// We need to reset the started parameter used by
				// the lock 
				shared.highStarted = false;

				long he = pilr.getHighEntryTimestampVal();
				long lb = pilr.getLowBoostTimestampVal();
				long ll = pilr.getLowLeavingTimestampVal();

				highBeforeLog[index] = highBefore;
				highEntryLog[index] = he;
				lowBoostLog[index]= lb;
				lowLeavingLog[index] = ll;
				
				// lowBoostPriorities[index] = pilr.getLowBoostPriority();
			}
            System.out.println("Hi pri test end");
		}

		/**
		 * @param pilr
		 */
		public void setLockedResource(LockedResource pilr) {
			this.pilr = pilr;
		}

		/**
		 * @param shared
		 */
		public void setSharedInfo(SharedInfo shared) {
			this.shared = shared;
		}
	}

	class MediumPriorityThread implements Runnable {
		static final long LOOP_TIME = 1000000000;

		SharedInfo shared = null;

		public void run() {
		   //System.out.println("In med thread");
			for (int index = 0; index < TestProps.numSamples; index++) {
				shared.mediumReady = true;
				
				// Wait until all threads are ready
				while (!(shared.lowReady
					&& shared.mediumReady
					&& shared.highReady)) {
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						System.out.println("Medium priority thread interrupted");
					}
				}
//				if (index > shared.currentIndex)
//				   index = shared.currentIndex;
//				System.out.println("Medium: Starting sample " + index);

				// Wait a bit more
				try {
					Thread.sleep(100);
				} catch (InterruptedException ie) {
					ie.printStackTrace();
				}
				shared.mediumReady = false;
				try
            {
               Thread.sleep(400);
            } catch (InterruptedException e)
            {
               // TODO Auto-generated catch block
               e.printStackTrace();
            }
				
				// Loop for a while
				long elapsedTime = 0;
				long presentTime = 0;

				long initialTime = Time.getCycles();
                //long initialTime = getFakeTime();
				
				//System.out.println("Medium: Starting Loop");

				Waster.msWaste(TestProps.period + 500);
				presentTime = Time.getCycles();
				elapsedTime = presentTime - initialTime;
				//System.out.println("Medium: done looping: " + elapsedTime);

				mediumEndLog[index] = elapsedTime;

			}
            System.out.println("Med pri test end");
		}

		/**
				   * @param shared
				   */
		public void setSharedInfo(SharedInfo shared) {
			this.shared = shared;

		}
	}

	class LowPriorityThread implements Runnable {
		SharedInfo shared = null;
		LockedResource lockedResource = null;

		public void run() {
		   //System.out.println("In lo thread");
			// Take a measurement for each sample
			for (int index = 0; index < TestProps.numSamples; index++) {

				shared.lowReady = true;
				
				// Wait until all threads are ready
				while (!(shared.lowReady
					&& shared.mediumReady
					&& shared.highReady)) {
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						System.out.println("Low priority thread interrupted");
					}
				}
				//System.out.println("Low: Starting n Sample " + index);
				try {
					Thread.sleep(100);
				} catch (InterruptedException ie) {
					ie.printStackTrace();
				}

				shared.lowReady = false;

				// Take the time just before grabbing the lock
				long beforeTimestamp = Time.getCycles();

				lockedResource.lock(LOW_PRIORITY_THREAD, shared);

				long afterTimestamp = Time.getCycles();
				shared.lowReady = false;
				
				lowBeforeLog[index] = beforeTimestamp;
				lowAfterLog[index] = afterTimestamp;
			}
			System.out.println("Low pri test end");
		}
		
        /**
		 * @param pilr
		 */
		public void setLockedResource(LockedResource lockedResource) {
			this.lockedResource = lockedResource;
		}

		/**
		 * @param shared
		 */
		public void setSharedInfo(SharedInfo shared) {
			this.shared = shared;
		}
	}

	class LockedResource {
		private long entryTimestampVal = 0;
		private long lowBoostTimestampVal = 0;
		private long lowLeavingTimestampVal = 0;
		private long highEntryTimestampVal = 0;
		private long currentTimestampVal = 0;
		private int lowBoostPriority = 0;

		public synchronized void lock(int threadType, SharedInfo shared) {
			// Take the first timestamp upon entering, it
			// belongs to somebody
			entryTimestampVal = Time.getCycles();

			// If this is the high priority thread, save the
			// timestamp
			if (threadType == HIGH_PRIORITY_THREAD) {
				highEntryTimestampVal = entryTimestampVal;
			
			} else {
				// System.out.println("Low thread in lock");
			}

			int ind = 0;
			while (shared.highStarted == false) {
				
			}

			// Immediately take a timestamp
			currentTimestampVal = Time.getCycles();
			
			// If this is the low priority thread, save the
			// timestamp for later retrieval

			if (threadType == LOW_PRIORITY_THREAD) {
				lowBoostTimestampVal = currentTimestampVal;

				RealtimeThread thread = RealtimeThread.currentRealtimeThread();
				lowBoostPriority = 
	              ((PriorityParameters)thread.getSchedulingParameters()).getPriority();
				
				// Finally save the departure time
				lowLeavingTimestampVal = Time.getCycles();
			} else if (threadType == HIGH_PRIORITY_THREAD) {
				// System.out.println("High priority thread finished with lock");
			}
		}

		public long getLowBoostTimestampVal() {
			return lowBoostTimestampVal;
		}

		public long getHighEntryTimestampVal() {
			return highEntryTimestampVal;
		}

		public long getLowLeavingTimestampVal() {
			return lowLeavingTimestampVal;
		}

		/**
		 * @return Returns the lowBoostPriority.
		 */
		public int getLowBoostPriority() {
			return lowBoostPriority;
		}
		/**
		 * @param lowBoostPriority The lowBoostPriority to set.
		 */
		public void setLowBoostPriority(int lowBoostPriority) {
			this.lowBoostPriority = lowBoostPriority;
		}
	}
	public static void main(String[] args) {
	   ThreadLauncherFactory.startTest( PriorityInheritance.class, TestProps.class );
	
	}
	class BGThread implements Runnable {
	   int[] garbage;

       public void run() {
           if ( null == garbage ) {
               garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
           }
           else {
               garbage = null;
           }
       }
   }
}
