/*
 * Created on Jun 8, 2004
 *
 * Synchronization.java
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */
package suramadu.tests.latency;

/**
 * 
 * <p>
 * <B><I>PURPOSE: </I> </B> Measure the time it takes to enter and leave a synchronized method
 * <p>
 * <B><I>DESIGN: </I> </B> Over a number of samples, measure time to enter and leave a synchronized 
 * method
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B> 
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"100"<td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"750"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgpriority"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgperiod"</td>
 * <td>"200"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgidelay"</td>
 * <td>"400"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"bgthreads"</td>
 * <td>"1"</td>
 * <td></td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode statistical
 * measurements on the collected data.
 * <p>
 * <B><I>EXAMPLES: </I> </B> None.
 * <p>
 * <B><I>ORIGINS: </I> </B> This test is based upon ideas from the AFRL and the rtpresto suites.
 * <p>
 * 
 * @see suramadu.tests.jitter.JavaJitterTest
 *      <p>
 * @author Mark Indictor, Dan Allard
 */

import java.io.IOException;
import java.util.Iterator;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.Range;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * @author mbi
 * Jun 8, 2004
 * 
 * $Id: Synchronization.java,v 1.1 2004/09/23 15:34:04 mbi Exp $
 *
 * Synchronization
 */
public class Synchronization {
	static final String LOAD_CLASS_NAME = "suramadu.tests.latency.SynchronizationLoadClass";

	public static class TestProps extends TestProperties {
	    public static final String[] DEFAULT_PROPERTIES = {
				"memtype", "IMMORTAL",
				"threadtype", "NHRT",
				"iterations", "1",
				"priority", "250",
				"period", "250",
				"idelay", "100",
				"bgthreads", "0",
				"bgpriority", "250",
				"bgperiod", "100",
				"bgidelay", "400",
				"histbins", "50",
				"numSamples", "10"
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

		protected void loadArguments() throws PropertyValidationException {
			// Load all of the thread params
			iterations = getIntInputProperty("iterations");
			priority = getIntInputProperty("priority");
			period = getIntInputProperty("period");
			idelay = getIntInputProperty("idelay");
			bgThreads = getIntInputProperty("bgthreads");
			bgPriority = getIntInputProperty("bgpriority");
			bgPeriod = getIntInputProperty("bgperiod");
			bgIdelay = getIntInputProperty("bgidelay");
			histBins = getIntInputProperty("histbins");
			priority = getIntInputProperty("priority");

			// Load the other test params
			String threadType = getInputProperty("threadtype");
			String memType = getInputProperty("memtype");
			numSamples = getIntInputProperty("numSamples");

			// Convert thread and memory types to actual types
			tt = ThreadType.getThreadType(threadType);
			mt = MemoryType.getMemoryType(memType);
		}

		/**
		 * Saves argument values in static class members.
		 * 
		 * @throws NumberFormatException
		 *             If a property is misformed.
		 */
		protected void validateArguments()throws PropertyValidationException {
			boolean ok = true;

			if (0 > iterations) {
				ok = false;
				System.out.println(
					"iterations property must be >= 0. Using default: "
						+ iterations);
			}
			if (0 > priority) {
				ok = false;
				System.out.println(
					"priority property must be >= 0. Using default: "
						+ priority);
			}
			if (0 > iterations) {
				ok = false;
				System.out.println(
					"period property must be >= 0. Using default: " + period);
			}
			if (0 > idelay) {
				ok = false;
				System.out.println(
					"idelay property must be >= 0. Using default: " + idelay);
			}
			if (0 > bgThreads) {
				ok = false;
				System.out.println(
					"bgthreads property must be >= 0. Using default: "
						+ bgThreads);
			}
			if (0 > bgPriority) {
				ok = false;
				System.out.println(
					"bgpriority property must be >= 0. Using default: "
						+ bgPriority);
			}
			if (0 > bgPeriod) {
				ok = false;
				System.out.println(
					"bgperiod property must be >= 0. Using default: "
						+ bgPeriod);
			}
			if (0 > bgIdelay) {
				ok = false;
				System.out.println(
					"bgidelay property must be >= 0. Using default: "
						+ bgIdelay);
			}
			if (!ok)
				throw new PropertyValidationException("One or more input properties are invalid.");
		}

		/** retrieved from default or specified properties */
		public static int iterations;
		public static int priority;
		public static long period;
		public static long idelay;
		public static int bgPriority;
		public static long bgPeriod;
		public static long bgIdelay;
		public static int bgThreads;

		/** the number of histogram bins to use for mode calculation (evenly divided between min and max) */
		public static int histBins;

		/** Thread type to use. */
		public static ThreadType tt;

		/** Memory type to use */
		public static MemoryType mt;

		/** Number of test samples */
		public static int numSamples;
	}

	/**
	 * 
	 */
	private static TestProps testProps;

	private static boolean calculated = false;

	//private static TimeStampLog enterTimeLog = null;
	//private static TimeStampLog leaveTimeLog = null;
	private static long[] enterTimeLog = null;
	private static long[] leaveTimeLog = null;

	/**
	 * Start the test
	 *  
	 */
	public void run() {
		try {
			// Setup the high priority thread using the ThreadLauncher to
			// manage it

			SynchronizationInit si = new SynchronizationInit();
			
			ThreadLauncherData tldata =
				new ThreadLauncherData(
					TestProps.tt,
					TestProps.mt,
					si,
					TestProps.priority,
					TestProps.period,
					TestProps.idelay);

			// For this test iterations is normally 1
			tldata.setIterations(TestProps.iterations);

			ThreadLauncher highPriorityLauncher =
				ThreadLauncherFactory.getLauncher(tldata);
			highPriorityLauncher.setIterations(TestProps.iterations);

			ThreadLauncher[] bgThreadArray = null;
            if ( 0 < TestProps.bgThreads ) {
                bgThreadArray = new ThreadLauncher[ TestProps.bgThreads ];
                for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                    double randomDelay = Math.random() * TestProps.bgIdelay;
                    bgThreadArray[ i ] = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( ThreadType.PERIODICJAVATHREAD, MemoryType.HEAPMEMORY, new BGThread(), TestProps.bgPriority, TestProps.bgPeriod, TestProps.bgIdelay ) );
                    bgThreadArray[ i ].setIterations( - 1 );
                    bgThreadArray[ i ].launch();
                }
            }
			Thread bl = highPriorityLauncher.launch();

			// wait for the test to complete
			bl.join();
            // Kill off background threads
			if ( null != bgThreadArray ) {
               for ( int i = 0; i < TestProps.bgThreads; i++ ) {
                   bgThreadArray[ i ].kill();
               }
           }
			calculate();
			// Start the lower priority thread
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * 
	 */
	private void calculate() {
		//		Get everything out of the loadTimeLog

		//long[] results = enterTimeLog.toArray(TestProps.numSamples);
	    long[] results = enterTimeLog;
		testProps.setOutputProperty("enterMethodTimes", results);
		testProps.setOutputProperty("max", StatPack.max(results));
		testProps.setOutputProperty("min", StatPack.min(results));
		testProps.setOutputProperty("mean", StatPack.mean(results));
		testProps.setOutputProperty("stddev", StatPack.stddev(results));
		testProps.setOutputProperty("variance", StatPack.variance(results));

		Range mode = StatPack.mode(results, TestProps.histBins);
		testProps.setOutputProperty("mode.lobound", (Number) mode.getLoBound());
		testProps.setOutputProperty("mode.hibound", (Number) mode.getHiBound());
		testProps.setOutputProperty("mode.count", mode.getCount());
		testProps.setOutputProperty(
			"mode.percent",
			(double) mode.getCount()
				* 100.0
				/ ((double) TestProps.iterations - 1));
		
		// Now calculate the leave time		
		//results = leaveTimeLog.toArray(TestProps.numSamples);
		results = leaveTimeLog;
		testProps.setOutputProperty("leaveMethodTimes", results);
		testProps.setOutputProperty("max", StatPack.max(results));
		testProps.setOutputProperty("min", StatPack.min(results));
		testProps.setOutputProperty("mean", StatPack.mean(results));
		testProps.setOutputProperty("stddev", StatPack.stddev(results));
		testProps.setOutputProperty("variance", StatPack.variance(results));

		Range lmode = StatPack.mode(results, TestProps.histBins);
		testProps.setOutputProperty("mode.lobound", (Number) lmode.getLoBound());
		testProps.setOutputProperty("mode.hibound", (Number) lmode.getHiBound());
		testProps.setOutputProperty("mode.count", lmode.getCount());
		testProps.setOutputProperty(
			"mode.percent",
			(double) lmode.getCount()
				* 100.0
				/ ((double) TestProps.iterations - 1));
				
		try {
			testProps.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	/**
	 * 
	 */
	class SynchronizationInit extends TestInitializer {

        /**
         * 
         * @uml.property name="numSamples"
         */


		public void run() {
			SynchronizationThread thread = new SynchronizationThread();
		    //enterTimeLog = new TimeStampLog(numSamples);
			//leaveTimeLog = new TimeStampLog(numSamples);
			enterTimeLog = new long[TestProps.numSamples];
			leaveTimeLog = new long[TestProps.numSamples];	
			setTestRunner(thread);
		}



	}

	class SynchronizationThread implements Runnable {

		public void run() {
			// Loop for n samples, loading and unloading the 
			// loaderClass and calculating the load time
			for (int index = 0; index < TestProps.numSamples; index++) {
				// First unload the class to make sure it
				// is not already available

				// unload(LOAD_CLASS_NAME);

				// We need to recreate this on each loop

				long beforeTime = -1;
				long afterTime = -1;

				beforeTime = Time.getCycles();
				// beforeTime = new Date().getTime();
				SynchronizationResource.synchMethod();
				afterTime = Time.getCycles();
				// afterTime = new Date().getTime();

				//Calculate time to enter method
				long enterTime =
					SynchronizationResource.getEnterTime() - beforeTime;

				// Calculate time to leave method
				long leaveTime =
					afterTime - SynchronizationResource.getLeaveTime();

				long diff = Math.abs(beforeTime - afterTime);
				
				// Take timestamp before load and after 

				beforeTime = -1;
				afterTime = -1;

				beforeTime = Time.getCycles();
				// beforeTime = new Date().getTime();
				SynchronizationResource.normalMethod();
				afterTime = Time.getCycles();
				// afterTime = new Date().getTime();

				// Calculate time to enter method
				enterTime = SynchronizationResource.getEnterTime() - beforeTime;

				// Calculate time to leave method
				leaveTime = afterTime - SynchronizationResource.getLeaveTime();

				//enterTimeLog.addTimeStamp(index, enterTime);
				//leaveTimeLog.addTimeStamp(index, leaveTime);
				enterTimeLog[index] = enterTime;
				leaveTimeLog[index] = leaveTime;
			}

		}
	}
	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public Synchronization() throws Exception {
        this( new TestProps( Synchronization.class ) );
    }

    public Synchronization( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
	public static void main(String[] args) {
	   ThreadLauncherFactory.startTest( Synchronization.class, TestProps.class );
	}
	class BGThread implements Runnable {
       int[] garbage;

       public void run() {
           if ( null == garbage ) {
               garbage = new int[ (int)Math.round( Math.random() * 1024.0 ) ];
           }
           else {
               garbage = null;
           }
       }
   }
}
