/*
 * WaitFreeWrite.java Created on July 12, 2004, 10:19 AM
 * 
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED. 
 *  United States Government Sponsorship acknowledged. Any commercial use must be 
 *  negotiated with the Office of Technology Transfer at the California Institute 
 *  of Technology. 
 *
 *  This software may be subject to U.S. export control laws and regulations.  
 *  By accepting this document, the user agrees to comply with all applicable U.S. 
 *  export laws and regulations.  User has the responsibility to obtain export 
 *  licenses, or other export authority as may be required before exporting such 
 *  information to foreign countries or providing access to foreign persons. 
 */

package suramadu.tests.latency;

import java.util.Iterator;
import java.util.Random;
import javax.realtime.WaitFreeWriteQueue;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * This test measures the time it takes for data written to a WaitFreeWriteQueue to become available by a
 * <p>
 * <B><I>PURPOSE: </I> </B>To determine the latency of data passing through a WaitFreeQueue.
 * <p>
 * <B><I>DESIGN: </I> </B> 
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B>
 * <p>
 * <B><I>INPUTS: </I> </B> <br>
 * <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"priority"</td>
 * <td>"250"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"period"</td>
 * <td>"10"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"idelay"</td>
 * <td>"200"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"qsize"</td>
 * <td>"30"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"readermemtype"</td>
 * <td>"SCOPE"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"readerthreadtype"</td>
 * <td>"NHRT"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"writermemtype"</td>
 * <td>"SCOPE"</td>
 * <td></td>
 * </tr>
 * <tr>
 * <td>"writerthreadtype"</td>
 * <td>"NHRT"</td>
 * <td></td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B> The output contains the maximum, minimum, median, mean, standard deviation and mode statistical
 * measurements on the collected data.
 * <p>
 * <B><I>EXAMPLES: </I> </B> None.
 * <p>
 * <B><I>ORIGINS: </I> </B>
 * <p>
 * 
 * @author Mark Indictor
 */
public class WaitFreeWrite implements Runnable {
	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
				"iterations", "1000", 
				"priority", "250", 
				"period", "10",
				"idelay", "200",
				"qsize", "30",
				"readermemtype", "SCOPE", 
				"readerthreadtype", "NHRT", 
				"writermemtype", "SCOPE", 
				"writerthreadtype", "NHRT", 
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void loadArguments() throws PropertyValidationException {
	        iterations = getIntInputProperty( "iterations" );
	        priority = getIntInputProperty( "priority" );
	        period = getIntInputProperty( "period" );
	        idelay = getIntInputProperty( "idelay" );
	        qsize = getIntInputProperty( "qsize" );
	        
	        writertt = ThreadType.getThreadType( getInputProperty( "writerthreadtype" ) );
	        writermt = MemoryType.getMemoryType( getInputProperty( "writermemtype" ) );
	        readertt = ThreadType.getThreadType( getInputProperty( "readerthreadtype" ) );
	        readermt = MemoryType.getMemoryType( getInputProperty( "readermemtype" ) );
	    }

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void validateArguments() throws PropertyValidationException {
	        boolean ok = true;

	        if ( 0 > iterations ) {
	            ok = false;
	            System.out.println( "iterations property must be >= 0. Using default: " + iterations );
	        }
	        if ( 0 >= priority ) {
	            ok = false;
	            System.out.println( "priority property must be >= 0. Using default: " + priority );
	        }
	        if ( 0 > period ) {
	            ok = false;
	            System.out.println( "period property must be >= 0. Using default: " + period );
	        }
	        if ( 0 > idelay ) {
	            ok = false;
	            System.out.println( "idelay property must be >= 0. Using default: " + idelay );
	        }
	        if ( 0 > qsize ) {
	            ok = false;
	            System.out.println( "qsize property must be >= 0. Using default: " + qsize );
	        }
	        if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
	    }

	    /** retrieved from default or specified properties. */
	    public static int            iterations;
	    
	    /** Size of ReadFreeWriteQueue */
	    public static int            priority;

	    /** retrieved from default or specified properties. */
	    public static long           period;

	    /** retrieved from default or specified properties. */
	    public static int            idelay;

	    /** Size of WaitFreeWriteQueue */
	    public static int            qsize;

		/** retrieved from default or specified properties */
	    public static ThreadType     readertt;

	    /** retrieved from default or specified properties */
	    public static MemoryType     readermt;

		/** retrieved from default or specified properties */
	    public static ThreadType     writertt;

	    /** retrieved from default or specified properties */
	    public static MemoryType     writermt;
	}


    /**
     * actual properties used by this test run. Filled in from default and specified properties at class instantiation.
     */
    public static TestProps      testProps;

     /** array for collecting timestamps during test */
    public static long           timeArray[];

    /** iteration counter */
    public static int            counter;

    /** the queues */
    public static WaitFreeWriteQueue wfwq;
    public static long[] readTimeArray;
    public static long[] writeTimeArray;

	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public WaitFreeWrite() throws Exception {
        this( new TestProps( WaitFreeWrite.class ) );
    }

    public WaitFreeWrite( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
        counter   = 0;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( WaitFreeWrite.class, TestProps.class );
    }

    public void run() {
        try {
            // Start up the reader.
            ThreadLauncher readerLauncher = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( TestProps.readertt, TestProps.readermt, new ReaderInit(), TestProps.priority + 1, TestProps.period, TestProps.idelay ) );
            readerLauncher.setIterations( TestProps.iterations );
            Thread readerThread = readerLauncher.launch();	// start reader first because it will block.

            // Start up the writer
            ThreadLauncher writerLauncher = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( TestProps.writertt, TestProps.writermt, new WriterInit(), TestProps.priority, TestProps.period, TestProps.idelay ) );
            writerLauncher.setIterations( TestProps.iterations );
            Thread writerThread = writerLauncher.launch();	// start writer

            // wait for the test to complete
            writerThread.join();
            readerThread.join();
            
            // log results
            long[] times = new long[ TestProps.iterations ];
            for ( int i = 0; i < times.length; i++ ) {
                times[ i ] = (long)Time.nsFromCycles( readTimeArray[ i ] - writeTimeArray[ i ] );
            }

            testProps.setOutputProperty( "wfwtimes", times );
            testProps.setOutputProperty( "wfwtimesmax", StatPack.max( times ) );
            testProps.setOutputProperty( "wfwtimesmin", StatPack.min( times ) );
            testProps.setOutputProperty( "wfwtimesmean", StatPack.mean( times ) );
            testProps.setOutputProperty( "wfwtimesstddev", StatPack.stddev( times ) );
            testProps.setOutputProperty( "wfwtimesvariance", StatPack.variance( times ) );

            // report on what we're going to use.
            testProps.close();            ;
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    class ReaderInit extends TestInitializer {
        public void run() {
            setTestRunner( new Reader() );
        }
    }

    static class Reader implements Runnable {
        public Object obj;
        
        public Reader() {
            wfwq           = new WaitFreeWriteQueue( TestProps.qsize );
            readTimeArray  = new long[ TestProps.iterations ];
            writeTimeArray = new long[ TestProps.iterations ];
        }
        
        public void run() {
            obj = wfwq.read();
            readTimeArray[ counter++ ] = Time.getCycles();
            //System.out.println( "Read: " + obj );
        }
    }

    class WriterInit extends TestInitializer {
        public void run() {
            setTestRunner( new Writer( new Random() ) );
        }
    }

    class Writer implements Runnable {
        private class PayLoad {
            int value;
            
            public String toString() {
                return Integer.toString( value );
            }
        }
        
        Random r;
        PayLoad obj;
        
        public Writer( Random r ) {
            this.r = r;
            obj = new PayLoad();
        }
        
        public void run() {
            try {
	            obj.value = r.nextInt( Integer.MAX_VALUE );
	            //System.out.print( "Writing: " + obj + "..." );
	            writeTimeArray[ counter ] = Time.getCycles();
	            wfwq.write( obj );
	            //System.out.println( "Done." );
           }
            catch ( Exception e ) {
                e.printStackTrace();
            }
        }
    }
}