/*
 *  Copyright 2004, by the California Institute of Technology. ALL RIGHTS RESERVED.
 *  United States Government Sponsorship acknowledged. Any commercial use must be
 *  negotiated with the Office of Technology Transfer at the California Institute
 *  of Technology.
 *
 *  This software may be subject to U.S. export control laws and regulations.
 *  By accepting this document, the user agrees to comply with all applicable U.S.
 *  export laws and regulations.  User has the responsibility to obtain export
 *  licenses, or other export authority as may be required before exporting such
 *  information to foreign countries or providing access to foreign persons.
 */
 
package suramadu.tests.throughput;

import java.util.Iterator;
import suramadu.util.exec.MemoryType;
import suramadu.util.exec.PropertyValidationException;
import suramadu.util.exec.TestInitializer;
import suramadu.util.exec.TestProperties;
import suramadu.util.exec.ThreadLauncher;
import suramadu.util.exec.ThreadLauncherData;
import suramadu.util.exec.ThreadLauncherFactory;
import suramadu.util.exec.ThreadType;
import suramadu.util.stats.StatPack;
import suramadu.util.time.Time;

/**
 * This test exercises Java's arithmetic capabilities to measure throughput during computationally intensive operations with
 * floating point numbers.
 * <p>
 * <B><I>PURPOSE: </I> </B> To measure the Java and RTSJ computational throughput with floating point arithmatic.
 * <p>
 * <B><I>DESIGN: </I> </B> This class does floating point operations - 30 operations in one call to doOps. The arguments to that
 * function are simply to make sure that the compiler can't optimize away any operations. x3 must not be zero.
 * <p>
 * <B><I>ASSUMPTIONS: </I> </B>
 * <p>
 * <B><I>INPUTS: </I> </B> <table>
 * <tr>
 * <th align="left">Property</th>
 * <th align="right">Default Value</th>
 * <th align="left">Description</th>
 * </tr>
 * <tr>
 * <td>"iterations"</td>
 * <td>"1000000"</td>
 * <td>The number of test iterations to perform.</td>
 * </tr>
 * <tr>
 * <td>"printTimeDeltas"</td>
 * <td>"No"</td>
 * <td>Indicates whether or not to print all of the time deltas.</td>
 * </tr>
 * </table>
 * <p>
 * <B><I>OUTPUTS: </I> </B>
 * <p>
 * <B><I>EXAMPLES: </I> </B>*
 * <p>
 * <B><I>ORIGINS: </I> </B>
 * <p>
 * 
 * @author Alex Murray
 */
public class FloatingPoint implements Runnable {
    /** Default number of iterations */
    public static final int   DEFAULT_ITERATIONS = 1;

	public static class TestProps extends TestProperties{
	    public static final String[] DEFAULT_PROPERTIES = {
				"memtype", "HEAP", 
				"threadtype", "JAVA", 
				"iterations", Integer.toString( DEFAULT_ITERATIONS ), 
				"priority", "250",
				"period", "100", 
				"idelay", "100"
	    };
	    
	    public static Iterator factory( Class aClass ) throws PropertyValidationException, InstantiationException, IllegalAccessException {
	        return factory( TestProps.class, aClass, DEFAULT_PROPERTIES );
	    }

		/**
		 * @throws PropertyValidationException
		 */
		public TestProps() throws PropertyValidationException {
			super();
		}

		/**
		 * @param aClass
		 * @throws PropertyValidationException
		 */
		public TestProps(Class aClass) throws PropertyValidationException {
			super( aClass, DEFAULT_PROPERTIES );
		}

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void loadArguments() throws PropertyValidationException {
	        priority = getIntInputProperty( "priority" );
	        period = getIntInputProperty( "period" );
	        idelay = getIntInputProperty( "idelay" );
	        iterations = getIntInputProperty( "iterations" );
	        memType = getInputProperty( "memtype" );
	        threadType = getInputProperty( "threadtype" );
	    }

	    /**
	     * Saves argument values in static class members.
	     * 
	     * @throws NumberFormatException
	     *             If a property is misformed.
	     */
	    protected void validateArguments() throws PropertyValidationException {
	        boolean ok = true;
	        System.out.println( threadType + "," + memType );

	        if ( 0 > priority ) {
	            ok = false;
	            System.out.println( "priority property must be >= 0. Using default: " + priority );
	        }
	        if ( 0 > period ) {
	            ok = false;
	            System.out.println( "period property must be >= 0. Using default: " + period );
	        }
	        if ( -1 > iterations ) {
	            ok = false;
	            System.out.println( "iterations property must be >= 0. Using default: " + iterations );
	        }
	        if ( 0 > idelay ) {
	            ok = false;
	            System.out.println( "idelay property must be >= 0. Using default: " + idelay );
	        }
	        if ( !ok ) throw new PropertyValidationException( "One or more input properties are invalid." );
	    }
	    
	    /** The number of iterations to perform */
	    public static int         iterations;

	    public static boolean     printTimeDeltas;

	    /** retrieved from default or specified properties. */
	    public static int         priority;

	    /** retrieved from default or specified properties. */
	    public static long        period;

	    /** retrieved from default or specified properties. */
	    public static long        idelay;

	    /** retrieved from default or specified properties */
	    public static String      threadType;

	    /** retrieved from default or specified properties */
	    public static String      memType;
	}

    /**
     * The properties object, for printing
     * 
     * @uml.property name="testProps"
     * @uml.associationEnd 
     * @uml.property name="testProps" multiplicity="(1 1)"
     */
    private TestProps testProps;

    /** Thread type to use. */
    public static ThreadType  tt;

    /** Memory type to use */
    public static MemoryType  mt;

	/**
     * constructor for test class Instantiates testProperties.
	 * @throws Throwable
	 * @throws IllegalAccessException
	 * @throws InstantiationException
     */
    public FloatingPoint() throws Exception {
        this( new TestProps( FloatingPoint.class ) );
    }

    public FloatingPoint( TestProps props ) throws Exception {
        System.out.println( props.getProperty( "propertiesfile" ) );
        testProps = props;
    }
    
    /**
     * @param args
     *            the command line arguments
     */
    public static void main( String args[] ) {
        ThreadLauncherFactory.startTest( FloatingPoint.class, TestProps.class );
    }

    public void run() {
        try {
            /** Can vary thread types used */
            tt = ThreadType.getThreadType( TestProps.threadType );

            /** Can vary memory types used */
            mt = MemoryType.getMemoryType( TestProps.memType );
            
            TestInit init = new TestInit();

            ThreadLauncher launcher = ThreadLauncherFactory.getLauncher( new ThreadLauncherData( tt, mt, init, TestProps.priority, TestProps.period, TestProps.idelay ) );
            launcher.setIterations( 1 );
            Thread t = launcher.launch();

            // wait for the test to complete (based on a calculation of how long it will take)
            t.join();

            // output results.
            long[] diff = StatPack.diff( init.tester.execTimes, 1 );

            for ( int i = 0; i < diff.length; i++ ) {
                diff[ i ] = (long)Time.nsFromCycles( diff[ i ] );
            }

            testProps.setOutputProperty( "exectimes", diff );
            testProps.setOutputProperty( "max", StatPack.max( diff ) );
            testProps.setOutputProperty( "min", StatPack.min( diff ) );
            testProps.setOutputProperty( "mean", StatPack.mean( diff ) );
            testProps.setOutputProperty( "stddev", StatPack.stddev( diff ) );
            testProps.setOutputProperty( "variance", StatPack.variance( diff ) );
            testProps.close();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    class TestInit extends TestInitializer {
        public Test tester;
        
        public void run() {
            tester = new Test( TestProps.iterations );
            setTestRunner( tester );
        }
    }

    class Test implements Runnable {
        public long[] execTimes;
        
        int iterations;
        double x1 = - 300943.45 ;
        double x2 = 3945.34 ;
        double x3 = - 102094.89349 ;
        double result;

        public Test( int iterations ) {
            this.iterations = iterations;
            execTimes = new long[ iterations + 1 ];
        }
        
        public void run() {
            for ( int i = 0; i < iterations + 1; i++ ) {
	            execTimes[ i ] = Time.getCycles();
	            
	            result = x1 * x2;
	            result /= x3;
	            result += x2;
	            result *= -35467.80796875;
	            result *= .009857467;
	            result -= x2;
	            result /= .009857467;
	            result += x1;
	            result -= 94082885.7467;
	            x1 = result /= -3.504060702;
	
	            result += x1;
	            result /= x3;
	            result += x1;
	            result *= 3467.80796875;
	            result *= -0.009857467;
	            result -= x1;
	            result /= -.009857467;
	            result += x2;
	            result -= -82885.7467;
	            x2 = result /= 3.500702;
	
	            result *= x1;
	            result /= x3;
	            result += x3;
	            result *= -467.80796875;
	            result *= 98.0857467;
	            result -= x3;
	            result /= 75.007467;
	            result += x3;
	            result -= -85.7467;
	            x3 = result /= 3045.500702;
            }
        }
    }
}